#include "behavior_analyzer.h"
#include "../include/CtApBehaviorAnalysis_/behaviorNodeGlobal.h"
#include "glog/logging.h"
#include "stdlib.h"
// #include <vector>
namespace ADC
{
	namespace planning
	{
		void BehaviorAnalysis::BehaviorAnalyzerStartRun(LocalView &local_view)
		{
			UpdateRMBaseInfo(*local_view.road_model);
			UpdateSMBaseInfo(*local_view.system_management);
			UpdateVSEBaseInfo(*local_view.vehicle_state);
			UpdateEHRBaseInfo(*local_view.control_waypoints);

			this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (this->input_BA.SM_DrvrSetSpd + 0.35f)) / c_BA_ms2kmh / 1.05f; // calc set speed by empirical formula
			this->msg_BA2TRPLongt.ba_trp_t_tm_gap_actu = GetTimeGapInfo(this->input_BA.SM_DrvrSetDstLvl); // always send driver time gap
			this->msg_BA2SM.ba_takeover_req = uint8_t(0); // 20220328: reset takeover request here
			bool isRMCylCntNrml = MonitorRMCycleCounter();
			if (!this->input_BA.RM_RoadModel.lane_list.empty() && isRMCylCntNrml == true)
			{
				RunOnce();
				// cout << "k_BA_Dst2FrtNrstOffsetThr01: " << k_BA_Dst2FrtNrstOffsetThr01 << std::endl;
				// cout << "k_BA_MinCoolDur4LCRmpMrgCARd: " << k_BA_MinCoolDur4LCRmpMrgCARd << std::endl;
				// cout << "k_BA_MinLaneEndLCPntDst01: " << k_BA_MinLaneEndLCPntDst01 << std::endl;
				// cout << "k_BA_LnEndTHThr: " << k_BA_LnEndTHThr << std::endl;
			}
			else
			{
				// if there is no normal lane list info from RM, then reset TRP state
				this->msg_BA2TRPLat.ba_trp_lat_sts = BA::TRPLatSts_None;
				this->msg_BA2TRPLat.ba_opt_action = BA::BA_ActionTyp_None;
				this->msg_BA2TRPLongt.ba_trp_indx_trp_lgt_sts = BA::TRPLongtSts_None;
				this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = false;
				this->msg_BA2TRPLongt.ba_trp_indx_spdlim_typ = uint8_t(0);
				this->msg_BA2TRPLongt.ba_trp_is_esaa = false;
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_None;
				this->msg_BA2SM.ba_drcn_ind_req = BA::DircnIndReqSts_BothOff;
				this->BA_DircnIndReq = BA::DircnIndReqSts_BothOff;
				this->BA_DircnIndReq_Deb = BA::DircnIndReqSts_BothOff;
				this->BA_AvgFlowDenseHigh_Deb = false;
				this->BA_FuncOnDurTm = 0.0f;
				this->BA_LaneChgAblFlg4TRPLat = false;
				this->BA_TmpTrgtRdTypMem = BA::TargetRoadTyp_NotAvlbl;
				// 20220906: ba debug
				this->msg_BA2Obs.m_ba_optimal_corridor_id = this->input_BA.RM_RoadModel.rm_header.cyclecounter;
				this->msg_BA2Obs.m_ba_optimal_target_obj_id = this->input_BA.EHR_EHRInfo.msg_header.cycle_counter;
				this->msg_BA2Obs.m_ba_optimal_action = this->input_BA.SM_LatActvFunc;
				this->msg_BA2Obs.m_ba_behavior_node_update_done = isRMCylCntNrml;
				// exit(0); // added 2022/09/12: actively exit to let node be restarted
				// assert(isRMCylCntNrml == true);
				// added 20220914 
				// when received rm counter is stuck, throw a error on purpose
				// if (isRMCylCntNrml == false)
				// {
				// 	uint32_t wp_id = this->input_BA.EHR_EHRInfo.control_waypoints_list.at(2).control_waypoint_id;
				// }
			}
		}

		void BehaviorAnalysis::UpdateSMBaseInfo(const dpc::SM4Decision &sm_msg)
		{
			this->input_BA.SM_DriveMode = BA::DriveMd(sm_msg.sm_drvmd);
			this->input_BA.SM_DrvrSetSpd = sm_msg.sm_drvrsetspd;
			this->input_BA.SM_DrvrSetDstLvl = BA::SeldDistLvl(sm_msg.sm_seltdistlvl);
			this->input_BA.SM_LFComfortAvlblBehavior = BA::LFCmfrtAvlblBhvr(sm_msg.sm_lfcmfrtavlbhvr);
			this->input_BA.SM_LCAvlblBehavior = BA::LCAvlblBhvr(sm_msg.sm_lcavlbhvr);
			this->input_BA.SM_LCPAvlblBehavior = BA::LCPAvlblBhvr(sm_msg.sm_lcpavlbhvr);
			this->input_BA.SM_LongtAvlblBehavior = BA::LongtAvlblBhvr(sm_msg.sm_lgtavlbhvr);
			this->input_BA.SM_DirectionIndicatorSts = BA::DircnIndSts(sm_msg.sm_drvr_indctdrct);
			this->input_BA.SM_LatActvFunc = BA::LatActvFuncSts(sm_msg.sm_latactvfunc);
			this->input_BA.SM_AutoLnChgAvlbl = BA::AutoLCAvlblSts(sm_msg.sm_nrp_lcavl);
			this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtID = sm_msg.sm_spdlmtid01;
			this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtVal = sm_msg.sm_speedlimit01 / c_BA_ms2kmh;
			this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtDist = sm_msg.sm_controlwaypoint_offset01;
			this->input_BA.SM_SignSpeedLimit02.m_SignSpdLmtID = sm_msg.sm_spdlmtid02;
			this->input_BA.SM_SignSpeedLimit02.m_SignSpdLmtVal = sm_msg.sm_speedlimit02 / c_BA_ms2kmh;
			this->input_BA.SM_SignSpeedLimit02.m_SignSpdLmtDist = sm_msg.sm_controlwaypoint_offset02;
			this->input_BA.SM_AutoLnChgAvlbl = BA::AutoLCAvlblSts_BothAvl;
		}

		void BehaviorAnalysis::UpdateVSEBaseInfo(const VED4DPC &vse_msg)
		{
			this->input_BA.VSE_LongitudinalVelocity = vse_msg.LongitudinalVelocity;
			this->input_BA.VSE_LatAcc = vse_msg.LateralAcceleration;
			this->input_BA.VSE_LongtAcc = vse_msg.LongitudinalAcceleration;
			this->input_BA.VSE_Yawrate = vse_msg.Yawrate; // deg/s
		}

		// void BehaviorAnalysis::Callback4SMInfo(const mpc_drv_msgs::msg::MsgSM2BA::SharedPtr sm_msg)
		// {
		// 	std::lock_guard<std::mutex> lock(data_mutex_sm);
		// 	UpdateSMBaseInfo(*sm_msg);
		// }

		// void BehaviorAnalysis::Callback4VSEInfo(const mpc_drv_msgs::msg::VehicleStateOutput::SharedPtr vse_msg)
		// {
		// 	std::lock_guard<std::mutex> lock(data_mutex_vse);
		// 	UpdateVSEBaseInfo(*vse_msg);
		// }

		void BehaviorAnalysis::UpdateRMBaseInfo(const RoadInfo &rm_msg)
		{
			this->input_BA.RM_LaneLineCrossed = rm_msg.car_position.lane_line_crossed; // NOT defined in RM
			this->input_BA.RM_RoadModel = rm_msg;
		}

		void BehaviorAnalysis::UpdateEHRBaseInfo(const ControlWaypoints &ehr_msg)
		{
			this->input_BA.EHR_EHRInfo = ehr_msg;
		}

		// void BehaviorAnalysis::Callback4RMInfo(const mpc_drv_msgs::msg::RoadInfo::SharedPtr rm_msg)
		// {
		// 	std::lock_guard<std::mutex> lock(data_mutex_rm);
		// 	UpdateRMBaseInfo(*rm_msg);
		// }

		// void BehaviorAnalysis::Callback4EHRInfo(const mpc_drv_msgs::msg::ControlWaypoints::SharedPtr ehr_msg)
		// {
		// 	std::lock_guard<std::mutex> lock(data_mutex_ehr);
		// 	UpdateEHRBaseInfo(*ehr_msg);
		// }

		void BehaviorAnalysis::RunOnce()
		{
			// bool autoLCSpdCondStsfy = JudgeAutoLCSpdCond(); // judge whether auto lane change spd condition is satisfied or not
			bool autoLCCrvCondStsfy = JudgeAutoLCCrvCond(); // 20220920: judge whether auto lane change curve condition is satisfied or not
//			DLOG(INFO) << "remain distance in major path " << this->input_BA.EHR_EHRInfo.remain_distance_major_path << " m" << std::endl;
			GlobalInfoProcess globalInfoProc;
			// init globalInfoProc.globalInfoContainer
			globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP = BA::TargetRoadTyp_NotAvlbl;
			globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP = BA::TargetRoadTyp_NotAvlbl;
			globalInfoProc.globalInfoContainer.m_WayPointID = uint8_t(0);
			globalInfoProc.globalInfoContainer.m_GlobalInfoSrc = BA::GlobalInfoSrc_NoVldWayPoint;

			bool gotoUpdtGlbInfo = false;
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto || this->input_BA.SM_LatActvFunc == BA::LatActvFuncSts_NRP)
			{
				gotoUpdtGlbInfo = true;
			}
			else
			{
				gotoUpdtGlbInfo = false;
			}
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto || 
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto ||
			    this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto || 
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto || 
				this->input_BA.SM_LatActvFunc == BA::LatActvFuncSts_NRP)
			{
				this->msg_BA2TRPLongt.ba_trp_is_esaa = true;
			}
			else
			{
				this->msg_BA2TRPLongt.ba_trp_is_esaa = false;
			}

			bool gotoUpdtGlbCost = UpdateGlobalInfo(gotoUpdtGlbInfo, globalInfoProc);
			DLOG(INFO) << "gotoUpdtGlbInfo " << int(gotoUpdtGlbInfo) << std::endl;
			DLOG(INFO) << "gotoUpdtGlbCost " << int(gotoUpdtGlbCost) << std::endl;

			DLOG(INFO) << "Distance to waypoint = " << globalInfoProc.globalInfoContainer.m_DistToWayPoint << " m"<< std::endl;
			DLOG(INFO) << "m_GlobalInfoSrc = " << int(globalInfoProc.globalInfoContainer.m_GlobalInfoSrc) << std::endl;
			DLOG(INFO) << "m_TgtRoadTyp_CWP = " << int(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP) << std::endl;
			DLOG(INFO) << "m_TgtRoadTyp_RWP = " << int(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP) << std::endl;
			DLOG(INFO) << "m_RouteRltvPos = " << int(globalInfoProc.globalInfoContainer.m_RouteRltvPos) << std::endl;
			DLOG(INFO) << "m_LftCrnrLaneID = " << int(globalInfoProc.globalInfoContainer.m_LftCrnrLaneID) << std::endl;
			DLOG(INFO) << "m_RghtCrnrLaneID = " << int(globalInfoProc.globalInfoContainer.m_RghtCrnrLaneID) << std::endl;
			DLOG(INFO) << "m_GlobalLaneChgNum = " << int(globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum) << std::endl;
			DLOG(INFO) << "m_GlobalLaneChgDrct = " << int(globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct) << std::endl;

			/* for (auto tgtLnID : this->input_BA.EHR_EHRInfo.control_waypoints_list[0].control_waypoint_lane_id_list)
			{
				DLOG(INFO) << "raw ctrl target lane id: " << int(tgtLnID) << std::endl;
			} */

			#ifdef GLOG
			for (auto tgtLnID : globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP)
			{
				DLOG(INFO) << "ctrl target lane id: " << int(tgtLnID) << std::endl;
			}
			for (auto tgtLnID : globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP)
			{
				DLOG(INFO) << "rmd target lane id: " << int(tgtLnID) << std::endl;
			}
			DLOG(INFO) << "wpExtend lane info showed below" << std::endl;
			DLOG(INFO) << "m_WpExtendLaneLst size = " << globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.size() << std::endl;
			int wpExtLnCnt = 0;
			for (auto wpExtLn : globalInfoProc.globalInfoContainer.m_WpExtendLaneLst)
			{
				wpExtLnCnt++;
				DLOG(INFO) << wpExtLnCnt << " id = " << int(wpExtLn.lane_id) << std::endl;
				DLOG(INFO) << wpExtLnCnt << " is it emergency lane = " << int(wpExtLn.is_emergency_lane) << std::endl;
				int rightExtLnCnt = 0;
				for (auto rightExtLn : wpExtLn.right_extend_lane)
				{
					rightExtLnCnt++;
					DLOG(INFO) << "  " << rightExtLnCnt << " right extend lane id = " << int(rightExtLn.adjacent_extendlane_id) << std::endl;
				}
				int leftExtLnCnt = 0;
				for (auto leftExtLn : wpExtLn.left_extend_lane)
				{
					leftExtLnCnt++;
					DLOG(INFO) << "  " << leftExtLnCnt << " left extend lane id = " << int(leftExtLn.adjacent_extendlane_id) << std::endl;
				}
			}
			DLOG(INFO) << "**Transit lane info showed below: " << std::endl;
			int transitLnCnt = 0;
			for (auto transitLn : this->input_BA.EHR_EHRInfo.transit_lanes_list)
			{
				transitLnCnt++;
				DLOG(INFO) << transitLnCnt << " transit lane id = " << int(transitLn.transit_lane_id) << std::endl;
				DLOG(INFO) << transitLnCnt << " transit connected lane id = " << int(transitLn.transit_connected_lane_id) << std::endl;
				DLOG(INFO) << transitLnCnt << " transit lane offset = " << transitLn.transit_lane_offset << std::endl;
				DLOG(INFO) << transitLnCnt << " transit lane type = " << int(transitLn.transit_type) << std::endl;
				DLOG(INFO) << "---------------" << std::endl;
			}

			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			auto it_Lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
							                [egoLnID](const Lane &lane_)
											{
												return (egoLnID == lane_.start_id);
											});
			if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
			{
				if (!it_Lane->right_neighbor_lane.empty())
				{
					int rghtNbrLnCnt = 0;
					for (auto rghtNbrLn : it_Lane->right_neighbor_lane)
					{
						rghtNbrLnCnt++;
						DLOG(INFO) << rghtNbrLnCnt << " right neighbor lane id = " << int(rghtNbrLn.lane_id) << std::endl;
						DLOG(INFO) << rghtNbrLnCnt << " right neighbor lane start range = " << rghtNbrLn.start_range << std::endl;
						DLOG(INFO) << rghtNbrLnCnt << " right neighbor lane end range = " << rghtNbrLn.end_range << std::endl; 
					}
				}
				else
				{
					DLOG(INFO) << "right neighbor lane is empty" << std::endl;
				}
				if (!it_Lane->left_neighbor_lane.empty())
				{
					int lftNbrLnCnt = 0;
					for (auto lftNbrLn : it_Lane->left_neighbor_lane)
					{
						lftNbrLnCnt++;
						DLOG(INFO) << lftNbrLnCnt << " left neighbor lane id = " << int(lftNbrLn.lane_id) << std::endl;
						DLOG(INFO) << lftNbrLnCnt << " left neighbor lane start range = " << lftNbrLn.start_range << std::endl;
						DLOG(INFO) << lftNbrLnCnt << " left neighbor lane end range = " << lftNbrLn.end_range << std::endl; 
					}
				}
				else
				{
					DLOG(INFO) << "left neighbor lane is empty" << std::endl;
				}
			}

			auto it_ExtLane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										   [egoLnID](const WPExtendLaneStruct &wp_ln)
										   {
											   return (wp_ln.lane_id == egoLnID);
										   });
			if (it_ExtLane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
			{
				// ego lane sma narrow lane info
				DLOG(INFO) << "ego lane sma info : " << std::endl;
				int extLnSMAnum = 0;
				for (auto sma : it_ExtLane->sma_info)
				{
					DLOG(INFO) << extLnSMAnum << " ego lane sma : " << std::endl;
					DLOG(INFO) << "  lane id : " << int(sma.lane_id) << std::endl;
					DLOG(INFO) << "  sma type : " << int(sma.type) << std::endl;
					DLOG(INFO) << "  sma start offset : " << int(sma.start_offset) << std::endl;
					extLnSMAnum++;
				}
				// left neighbor lane narrow lane info
				extLnSMAnum = 0;
				if (!it_ExtLane->left_extend_lane.empty())
				{
					DLOG(INFO) << "left lane sma info : " << std::endl;
					for (auto extLn : it_ExtLane->left_extend_lane)
					{
						uint8_t extLnID = extLn.adjacent_extendlane_id;
						auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										   			[extLnID](const WPExtendLaneStruct &wp_ln)
										   			{
											   			return (wp_ln.lane_id == extLnID);
										   			});
						if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
						{
							for (auto sma : it_extLn->sma_info)
							{
								DLOG(INFO) << extLnSMAnum << " left lane sma : " << std::endl;
								DLOG(INFO) << "  lane id : " << int(sma.lane_id) << std::endl;
								DLOG(INFO) << "  sma type : " << int(sma.type) << std::endl;
								DLOG(INFO) << "  sma start offset : " << int(sma.start_offset) << std::endl;
								extLnSMAnum++;
							}
						}
					}
				}
				extLnSMAnum = 0;
				if (!it_ExtLane->right_extend_lane.empty())
				{
					DLOG(INFO) << "right lane sma info : " << std::endl;
					for (auto extLn : it_ExtLane->right_extend_lane)
					{
						uint8_t extLnID = extLn.adjacent_extendlane_id;
						auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										   			[extLnID](const WPExtendLaneStruct &wp_ln)
										   			{
											   			return (wp_ln.lane_id == extLnID);
										   			});
						if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
						{
							for (auto sma : it_extLn->sma_info)
							{
								DLOG(INFO) << extLnSMAnum << " right lane sma : " << std::endl;
								DLOG(INFO) << "  lane id : " << int(sma.lane_id) << std::endl;
								DLOG(INFO) << "  sma type : " << int(sma.type) << std::endl;
								DLOG(INFO) << "  sma start offset : " << int(sma.start_offset) << std::endl;
								extLnSMAnum++;
							}
						}
					}
				}
			}

			#endif

			static BA::TargetRoadTyp lastTargetRoadTyp = BA::TargetRoadTyp_NotAvlbl;
			bool allwLCAntiDrctFlg = false;
			// judge if certain actions should be suppressed
			if (gotoUpdtGlbInfo == true && 
				this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_LeftLaneChange_Auto && 
				this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_RightLaneChange_Auto)
			{
				CalcIntrSpprAction(globalInfoProc, autoLCCrvCondStsfy, lastTargetRoadTyp, allwLCAntiDrctFlg);
			}
			DLOG(INFO) << "suppressed action size: " << this->BA_SupprActLst.size() << std::endl;
			if (!this->BA_SupprActLst.empty())
			{
				for (int c = 0; c < this->BA_SupprActLst.size(); c++)
				{
					DLOG(INFO) << "Suppressed action: " << c << " is " << int(this->BA_SupprActLst.at(c)) << std::endl;
				}
			}
			// judge whether to filter speed limit info
			if (gotoUpdtGlbInfo == true)
			{
				FilterSpdLmt(globalInfoProc);
			}
			else
			{
				this->BA_SpdLmtInfoFltred = false;
			}
			if (this->BA_SpdLmtInfoFltred == true)
			{
				this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtID = uint8_t(0);
				this->input_BA.SM_SignSpeedLimit02.m_SignSpdLmtID = uint8_t(0);
			}
			// 20220803: modify drive mode when necessary
			bool upgradeDriveMd4SpltRdCond = false;
			if ((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto /* || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto */) &&
				(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd) &&
				 globalInfoProc.globalInfoContainer.m_DistToWayPoint <= 2500.0f)
			{
				upgradeDriveMd4SpltRdCond = true;
				this->input_BA.SM_DriveMode = BA::DriveMd_Sport;
			}
			// 20221024: modify drive mode when LCR
			// if lane type is normal lane, then set drive mode to comfort, for purpose of sending a more stable lane change remind
			// if lane type is entry or acceleration lane, then keep original drive mode, for purpose of sending a more executable lane change remind
			// 20221025: 
			// add BA_TmpDriveMode for RMU to evaluate lane change space
			bool degradeDriveMd = false; 
			uint8_t egoLaneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;	
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto &&
				globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd)
			{
				auto itr_egoLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
										   		[egoLaneID](const Lane &lane_)
										   		{
										     		return(egoLaneID == lane_.start_id);
										   		});
				if (itr_egoLane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					if (itr_egoLane->lane_type != uint8_t(9) && itr_egoLane->lane_type != uint8_t(16))
					{
						degradeDriveMd = true;
					}
				}
			}
			
			// 20221027:
			// if lcavlblbhvr is lane change auto or left lane change auto or right lane change auto
			// and ego lane is ending in 200m
			// and current action is not lane change
			// then modify drive mode to sport
			bool upgradeDriveMd4mrgCARdCond = false;
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto ||
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto ||
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto)
			{
				auto iter_egoLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
										   		 [egoLaneID](const Lane &lane_)
										   		 {
										     	    return(egoLaneID == lane_.start_id);
										   		 });
				if (iter_egoLane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					if (iter_egoLane->end_range < ADC::common::k_BA_UpgradeDrvMdLnEndRngThr && iter_egoLane->is_end_out_of_range == false)
					{
						upgradeDriveMd4mrgCARdCond = true;
					}
				}
			}
			static bool upgradeDriveMd4mrgCARdSts = false;
			if (upgradeDriveMd4mrgCARdCond == true &&
				this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange &&
				this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange)
			{
				upgradeDriveMd4mrgCARdSts = true;
			}
			if (upgradeDriveMd4mrgCARdCond == false &&
				this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange &&
				this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange)
			{
				upgradeDriveMd4mrgCARdSts = false;
			}
			if (upgradeDriveMd4mrgCARdSts == true)
			{
				this->input_BA.SM_DriveMode = BA::DriveMd_Sport;
			}
			// DLOG(INFO) << "upgradeDriveMd4mrgCARdCond = " << int(upgradeDriveMd4mrgCARdCond) << std::endl;
			// DLOG(INFO) << "upgradeDriveMd4mrgCARdSts = " << int(upgradeDriveMd4mrgCARdSts) << std::endl;
			// 20221104:
			// if lane change intention has been generated for a long period of time while lane change has not been executed
			// then set drivemode to sport secretly
			if (ADC::common::k_BA_ChgDrvMdWaitLongEnough == true &&
				globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd &&
			    this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto &&
				this->BA_Wait4LCDurTm > ADC::common::k_BA_ChgDrvMdWaitDurTmThr)
			{
				this->input_BA.SM_DriveMode = BA::DriveMd_Sport;
			}

			// assign degraded drive mode for RMU
			if (degradeDriveMd == true)
			{
				if (this->input_BA.SM_DriveMode == BA::DriveMd_Sport)
				{
					this->BA_TmpDriveMd = BA::DriveMd_Normal;
				}
				else if (this->input_BA.SM_DriveMode == BA::DriveMd_Normal)
				{
					this->BA_TmpDriveMd = BA::DriveMd_Comfortable;
				}
			}
			else
			{
				this->BA_TmpDriveMd = this->input_BA.SM_DriveMode;
			}
			DLOG(INFO) << "Drive Mode = " << int(this->BA_TmpDriveMd) << std::endl;

			// store globalInfoContainer to globalInfo
			this->globalInfo = globalInfoProc.globalInfoContainer;
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_C || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_C || 
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChangeAbolition_C || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChangeAbolition_C ||
				this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto)
			{
				// lane change is commanded by driver, clear BA_SupprActLst
				this->BA_SupprActLst.clear();
			}
			// ---------------------------------------------------------------------------------------------------

			// 20220929: take line type into consideration, especially for LCP 
			uint8_t egoLneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			bool rghtLnNotCrsbl = false;
			bool lftLnNotCrsbl = false;
			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   [egoLneID](const Lane &lane_)
                     		           {
                                		   return (egoLneID == lane_.start_id);
                              		   });
 			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
  			{
    			// get first lane marker id of right side
    			uint32_t currRghtLnMrkID = uint32_t(0);
    			if (!it_lane->right_lane_marker_id_list.empty())
    			{
      				currRghtLnMrkID = it_lane->right_lane_marker_id_list.front();
					uint8_t currRghtLnTyp = uint8_t(0); // unknown
					if (GetLineTyp(currRghtLnMrkID, currRghtLnTyp) == true)
					{
						if (currRghtLnTyp == 1 || currRghtLnTyp == 3 || currRghtLnTyp == 5)
						{
							rghtLnNotCrsbl = true;
						}
					}
    			}
    			else
    			{
					DLOG(INFO) << "current right lane marker is empty!" << std::endl;
    			}
				// get first lane marker id of left side
    			uint32_t currLftLnMrkID = uint32_t(0);
    			if (!it_lane->left_lane_marker_id_list.empty())
    			{
      				currLftLnMrkID = it_lane->left_lane_marker_id_list.front();
					uint8_t currLftLnTyp = uint8_t(0); // unknown
					if (GetLineTyp(currLftLnMrkID, currLftLnTyp) == true)
					{
						if (currLftLnTyp == 1 || currLftLnTyp == 4 || currLftLnTyp == 5)
						{
							lftLnNotCrsbl = true;
						}
					}
    			}
    			else
    			{
					DLOG(INFO) << "current left lane marker is empty!" << std::endl;
    			}
			}
			// 20220517: LCP is always permitted in the future, add right lane change auto and left lane change auto
			if (ADC::common::k_BA_IsLCPAllw == true)
			{
				if ((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_C || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto) &&
					rghtLnNotCrsbl == false)
				{
					this->input_BA.SM_LCPAvlblBehavior = BA::LCPAvlblBhvr_RightLaneChangePrepare;
				}
				else if ((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_C || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto) &&
						  lftLnNotCrsbl == false)
				{
					this->input_BA.SM_LCPAvlblBehavior = BA::LCPAvlblBhvr_LeftLaneChangePrepare;
				}
				else if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto)
				{
					this->input_BA.SM_LCPAvlblBehavior = BA::LCPAvlblBhvr_LaneChangePrepare_Auto;
				}
			}
			// ---------------------------------------------------------------------------------------------------

			// 20221109: make basic scenario recognition
			if (upgradeDriveMd4mrgCARdCond == true)
			{
				this->BA_CrrtScenarioTyp = BA::ScenarioTyp_MergeIN;
			}
			else if (upgradeDriveMd4SpltRdCond == true)
			{
				this->BA_CrrtScenarioTyp = BA::ScenarioTyp_RampOut;
			}
			else
			{
				this->BA_CrrtScenarioTyp = BA::ScenarioTyp_Normal;
			}

			UpdateBaseInfo();
			DLOG(INFO) << "ego lane id: " << int(this->input_BA.RM_RoadModel.car_position.ego_lane_id) << std::endl;
			// this->input_BA.SM_LFComfortAvlblBehavior = BA::LFCmfrtAvlblBhvr_LongtLat_HOInclILC;
			BA::BA_ActionTyp BA_ActionProc;
			CurrentBehaviorJudgement(this->input_BA.RM_RoadModel,
									 this->BA_TgtLaneStID,
									 this->BA_TgtLaneEndID,
									 this->input_BA.BA_OptCorridorID_LastCycle,
									 this->input_BA.BA_OptAction_LastCycle,
									 this->BA_VehInLnCenter,
									 this->BA_LaneChangeAbolitionAvlbl,
									 BA_ActionProc,
									 this->BA_VehClsToLnCntr,
									 this->BA_ISLCLnIDnotChg,
									 this->input_BA.BA_OptLaneID_LastCycle);

			AvailableBehaviorDetermine(this->input_BA.SM_LFComfortAvlblBehavior,
									   this->input_BA.SM_LCAvlblBehavior,
									   this->input_BA.SM_LCPAvlblBehavior,
									   this->input_BA.SM_LongtAvlblBehavior,
									   this->BA_LaneChangeAbolitionAvlbl,
									   this->BA_LatAvlblBehavior,
									   this->BA_LongtAvlblBehavior);

			HistoryBehaviorStack(this->input_BA.BA_OptAction_LastCycle, this->BA_BehaviorStackVector);

			DirectionIndicatorSignalProcess(this->input_BA.SM_LCAvlblBehavior, this->BA_DirectionIndicatorInfo);

//			bool BA_IsBehaviorNotAvlbl = this->BA_LatAvlblBehavior.empty();
			bool BA_IsBehaviorNotAvlbl = true;
			auto it_res = std::find(this->BA_LatAvlblBehavior.begin(), this->BA_LatAvlblBehavior.end(), BA::BA_ActionTyp_LaneFollow);
			if (it_res != this->BA_LatAvlblBehavior.end())
			{
				BA_IsBehaviorNotAvlbl = false;
			}

			DLOG(INFO) << "Is BA_LatAvlblBehavior empty: " << int(this->BA_LatAvlblBehavior.empty()) << std::endl;
			std::cout << "Is BA_LatAvlblBehavior empty: " << int(this->BA_LatAvlblBehavior.empty()) << std::endl;
			// init locally stored key global info
			// static BA::GlobalInfoSrc lastGlobalInfoSrc = BA::GlobalInfoSrc_NoVldWayPoint;
			// static uint32_t lastWayPointID = uint32_t(0);
			static uint8_t lastSpdLmt = 40; // default value
			uint8_t spdLimActVal = 40;
			float spdLimNxtDst = 1000.0;

			if (!BA_IsBehaviorNotAvlbl)
			{
				this->BA_TORTyp = uint8_t(0); // 20220511: reset BA_TORTyp here
				this->BA_FuncOnDurTm += BEHAVIORANALYZERCYCLE / 1000.0f;
				if (this->BA_FuncOnDurTm > ADC::common::k_BA_MaxFuncOnDurTmThr)
				{
					this->BA_FuncOnDurTm = 0.0f;
				}
				bool isNoLineNoCar = JudgeNoLineNoCarCase();
				if (isNoLineNoCar == true)
				{
					this->msg_BA2TRPLongt.ba_trp_indx_trp_lgt_sts = BA::TRPLongtSts_Comfort;
					this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = 1000;
					this->msg_BA2TRPLongt.ba_trp_indx_tgt_id_iv = 0;
					this->msg_BA2TRPLongt.ba_trp_v_tgt_longt_spd_iv = 0;
					this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = false;
					this->msg_BA2TRPLongt.ba_trp_is_esaa = false;

					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
					this->msg_BA2SM.ba_opt_corridor_id = uint32_t(500);
					this->msg_BA2SM.ba_opt_target_obj_id = uint32_t(0);
					this->BA_LastActTyp2SM = BA::BA_ActionTyp_LaneFollow;

					this->msg_BA2TRPLat.ba_trp_lat_sts = BA::TRPLatSts_None;
					this->msg_BA2TRPLat.ba_opt_corridor_id = uint32_t(0);
					this->msg_BA2TRPLat.ba_opt_action = BA::BA_ActionTyp_None;
					this->msg_BA2TRPLat.ba_lat_actv_func_sts = BA::LatActvFuncSts_None;
					this->BA_LaneChgAblFlg4TRPLat = false;
					this->BA_TRPLatSts = BA::TRPLatSts_Comfort;
					this->BA_Wait4LCDurTm = 0.0f;
				}
				else
				{
					// store internal observation variable, BA_PassageListObs. clear and assign new value
					this->BA_PassageListObs.clear();
					this->BA_PassageListObs.resize(100);
					std::cout << "before expanding" << std::endl;
					ExpandBehaviorSequence(this->BA_LatAvlblBehavior);
					std::cout << "after expanding" << std::endl;
					this->BA_StepNum = -1;
					this->BA_BehaviorNodeUpdateDone = false;
					// update ba global cost

					if (UpdateGlobalCost(gotoUpdtGlbCost, globalInfoProc, allwLCAntiDrctFlg) == true)
					{
						// assign speed limit related info for TRPLongt
						if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd)
						{
							spdLimActVal = lastSpdLmt;
						}
						else
						{
							spdLimActVal = globalInfoProc.globalInfoContainer.m_SpeedLimit;
						}
						spdLimNxtDst = globalInfoProc.globalInfoContainer.m_DistToWayPoint;

						// update locally stored key global info
						// lastGlobalInfoSrc = globalInfoProc.globalInfoContainer.m_GlobalInfoSrc;
						// lastWayPointID = globalInfoProc.globalInfoContainer.m_WayPointID;
						/* if (lastGlobalInfoSrc == BA::GlobalInfoSrc_OnlyActCtrlWayPoint || lastGlobalInfoSrc == BA::GlobalInfoSrc_OnlyVirtCtrlWayPoint)
						{
							lastTargetRoadTyp = globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP;
						}
						else if (lastGlobalInfoSrc == BA::GlobalInfoSrc_OnlyRmdWayPoint)
						{
							lastTargetRoadTyp = globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP;
						} */
						// 20220723: last CWP road type determination 
						if (this->BA_TmpTrgtRdTypMem != globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP)
						{
							lastTargetRoadTyp = this->BA_TmpTrgtRdTypMem;
						}
						this->BA_TmpTrgtRdTypMem = globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP;
						DLOG(INFO) << "last target road type = " << int(lastTargetRoadTyp) << std::endl;
						lastSpdLmt = globalInfoProc.globalInfoContainer.m_SpeedLimit;
					}
					else
					{
						this->BA_Wait4LCDurTm = 0.0f;
					}
					if (gotoUpdtGlbInfo == true)
					{
						// adjust lane change interval dynamically
						DynChgLCInterval(globalInfoProc, lastTargetRoadTyp);
					}
					else
					{
						this->BA_DynChgLCintervalLvl = uint8_t(0);
					}
					DLOG(INFO) << "++ BA_DynChgLCintervalLvl = " << int(this->BA_DynChgLCintervalLvl) << std::endl;

					// sort BA_TerminatedBehaviorNodesVector by actionTyp
					std::sort(this->BA_TerminatedBehaviorNodesVector->begin(), this->BA_TerminatedBehaviorNodesVector->end(), Sort_by_actionTyp);
					// 20220403: when current behavior is LF, if the first terminated behavior node is not started off with LF, empty BA_TerminatedBehaviorNodesVector
					if (!this->BA_TerminatedBehaviorNodesVector->empty())
					{
						if (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LaneFollow)
						{
							if (this->BA_TerminatedBehaviorNodesVector->front().firstActionTyp() != BA::BA_ActionTyp_LaneFollow)
							{
								this->BA_TerminatedBehaviorNodesVector->clear();
							}
						}
					}
					DetermineOptimalCorridor(BA_IsBehaviorNotAvlbl, this->BA_OptimalCorridorID, this->BA_OptimalAction, this->BA_OptimalTargetObjID, this->BA_OptimalPassageID, this->BA_TgtLaneStID, this->BA_TgtLaneEndID);
					// judge whether direction indicator is turned off
					bool drcnIndTurnOff = JudgeDrcnIndTurnOff();

					if (gotoUpdtGlbInfo == true && 
						this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_LeftLaneChange_Auto && 
						this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_RightLaneChange_Auto)
					{
						ManageDircnIndReq(globalInfoProc, drcnIndTurnOff, lastTargetRoadTyp);
					}
					else if (this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange &&
					         this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange)
					{
						this->BA_DircnIndReq = BA::DircnIndReqSts_BothOff;
						this->BA_DircnIndReq_Deb = BA::DircnIndReqSts_BothOff;
					}

					DebounceDrcnIndReq();
					DebounceOptRes();
					// 20221031
					// determine whether to adjust trplatsts and debounce it
					if (ADC::common::k_BA_ChgTRPLatStsAllw == true)
					{
						AdjTrpLatSts(globalInfoProc, gotoUpdtGlbInfo);
					}
					else
					{
						this->BA_TRPLatSts = BA::TRPLatSts_Comfort;
					}
					DebounceTRPLatSts();
					DLOG(INFO) << "TRPLatSts = " << int(this->BA_TRPLatSts) << std::endl;
					DLOG(INFO) << "TRPLatSts_Deb = " << int(this->BA_TRPLatSts_Deb) << std::endl;
					if (GetOptLaneID())
					{
						DLOG(INFO) << "Optimal lane has been found" << std::endl;
					}
					else
					{
						DLOG(INFO) << "Optimal lane has not been found" << std::endl;
					}
					DLOG(INFO) << "optimal corridor id: " << int(this->BA_OptimalCorridorID) << std::endl;
					DLOG(INFO) << "optimal action: " << int(this->BA_OptimalAction) << std::endl;
					DLOG(INFO) << "optimal target id: " << int(this->BA_OptimalTargetObjID) << std::endl;
					DLOG(INFO) << "-------------------" << std::endl;
					DLOG(INFO) << "deb optimal lane id: " << int(this->BA_OptLaneID_Deb) << std::endl;
					DLOG(INFO) << "deb optimal action: " << int(this->BA_OptAction_Deb) << std::endl;
					DLOG(INFO) << "deb optimal corridor id: " << int(this->BA_OptCorridorID_Deb) << std::endl;
					// update this->BA_OrgLaneStID and this->BA_OrgLaneEndID
					if (this->BA_OptAction_Deb != BA::BA_ActionTyp_None)
					{
						UpdateOrgLaneInfo();
						// 20220728: delay the time to release lane change reference line in specific case
						static float tm = 2.0f;
						static bool stDlyRlsRefLine = false;
						BA::BA_ActionTyp procAction_lat = this->BA_OptAction_Deb;
						uint32_t procCorridorID_lat = this->BA_OptCorridorID_Deb;
						uint8_t egoLaneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
						if (ADC::common::k_BA_DlyLnChgInOptionalCase == true &&
						    (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd ||
							 (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
							  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
							  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd ||
							  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd ||
							  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd) && 
							globalInfoProc.globalInfoContainer.m_DistToWayPoint > ADC::common::k_BA_MinDrctIndOnInAdvDst) &&
							this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto)
						{
							if (this->input_BA.SM_DriveMode == BA::DriveMd_Normal || 
								this->input_BA.SM_DriveMode == BA::DriveMd_Abnormal ||
								this->input_BA.SM_DriveMode == BA::DriveMd_Sport ||
								this->input_BA.SM_DriveMode == BA::DriveMd_Comfortable)
							{
								uint8_t ego_LnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
								bool egoLnIsEnding = false;
								auto iter = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
														[ego_LnID](const Lane &lane_)
														{
															return (ego_LnID == lane_.start_id);
														});
								if (iter != this->input_BA.RM_RoadModel.lane_list.end())
								{
									DLOG(INFO) << "is_end_out_of_range = " << int(iter->is_end_out_of_range) << std::endl;
									DLOG(INFO) << "end_range = " << int(iter->end_range) << std::endl;
									if (iter->is_end_out_of_range == false && iter->end_range <= 300.0f)
									{
										egoLnIsEnding = true;
									}
								}
								if (egoLnIsEnding == false)
								{
									if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange) &&
										 this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LaneFollow)
									{
										stDlyRlsRefLine = true;
										tm = 0.0f;
									}
								}
								else
								{
									stDlyRlsRefLine = false;
								}
							}
							else
							{
								stDlyRlsRefLine = false;
							}
							if (stDlyRlsRefLine == true)
							{
								tm = tm + BEHAVIORANALYZERCYCLE / 1000.0f;
								tm = std::min(tm, 2.0f);
							}
							else
							{
								tm = 2.0f;
							}
							DLOG(INFO) << "tm = " << tm << std::endl;
							float dircIndOnInAdvTm = ADC::common::k_BA_DrctIndOnInAdvTm;
							if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_VirtFixDstRd)
							{
								dircIndOnInAdvTm = k_BA_DrctIndOnInAdvTm02;
							}
							if (tm <= dircIndOnInAdvTm)
							{
								// need to delay reference line releasing
								auto it_egoLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
															 [egoLaneID](const Lane &lane_)
															 {
																 return (egoLaneID == lane_.start_id);
															 });
								BA::FeatureTag trgtFeatureTag = BA::FeatureTag_LineFllw;
								auto it_crrd = std::find_if(it_egoLn->corridor_list.begin(), it_egoLn->corridor_list.end(),
															[trgtFeatureTag](const Corridor &corridor_)
															{
																return (trgtFeatureTag == corridor_.feature_tag);
															});
								if (it_crrd != it_egoLn->corridor_list.end())
								{
									procCorridorID_lat = it_crrd->corridor_id;
									procAction_lat = BA::BA_ActionTyp_LaneFollow;
								}
							}							
						}
						// 20220831: 
						// if SM_LCAvlblBehavior equals to lane change semi-auto and BA_OptAction_Deb is lane change or lane change prepare
						// then set procAction_lat to LaneFollow
						if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto)
						{
							if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange ||
							    this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange ||
								this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangePrepare ||
								this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangePrepare)
							{
								auto it_egoLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
															 [egoLaneID](const Lane &lane_)
															 {
																 return (egoLaneID == lane_.start_id);
															 });
								BA::FeatureTag trgtFeatureTag = BA::FeatureTag_LineFllw;
								auto it_crrd = std::find_if(it_egoLn->corridor_list.begin(), it_egoLn->corridor_list.end(),
															[trgtFeatureTag](const Corridor &corridor_)
															{
																return (trgtFeatureTag == corridor_.feature_tag);
															});
								if (it_crrd != it_egoLn->corridor_list.end())
								{
									procCorridorID_lat = it_crrd->corridor_id;
									procAction_lat = BA::BA_ActionTyp_LaneFollow;
								}
								else
								{
									/* TBD */
								}
							}
						}
						DLOG(INFO) << "procAction_lat = " << int(procAction_lat) << std::endl;
						DLOG(INFO) << "procCorridorID_lat = " << int(procCorridorID_lat) << std::endl;

						CorridorGeneration corridor_generation_(this->input_BA.RM_RoadModel, this->BA_OptCorridorID_Deb, this->BA_OptAction_Deb, this->input_BA.VSE_LongitudinalVelocity,
																this->BA_TgtLaneStID, this->BA_TgtLaneEndID, this->BA_IsLCAbltAvlbl, this->BA_OptPassageID_Deb, 
																this->input_BA.BA_OptAction_LastCycle, *this->BA_FirstPassageList.get(), procAction_lat, procCorridorID_lat);
						corridor_generation_.GenerateLatCrrdr(corridor_generation_.latBlueCrrdr, corridor_generation_.latYllwBrdr, corridor_generation_.latRedBrdr);
						corridor_generation_.GenerateLgtCrrdr(corridor_generation_.lgtBlueCrrdr, corridor_generation_.lgtYllwBrdr, corridor_generation_.lgtRedBrdr, corridor_generation_.lgtSpdLmt);
						// 20220705 added
						DLOG(INFO) << "temp blue corridor c0 = " << corridor_generation_.latBlueCrrdr.m_reference_line.c0 << std::endl;
						this->msg_BA2Obs.m_ba_lat_avlbl_behavior.clear();
						if (!WrapUpMsgOutput(corridor_generation_.lgtBlueCrrdr,
											 corridor_generation_.lgtYllwBrdr,
											 corridor_generation_.lgtRedBrdr,
											 corridor_generation_.lgtSpdLmt,
											 corridor_generation_.latBlueCrrdr,
											 corridor_generation_.latYllwBrdr,
											 corridor_generation_.latRedBrdr,
											 spdLimActVal,
											 spdLimNxtDst,
											 procAction_lat,
											 procCorridorID_lat,
											 globalInfoProc))
						{
							DLOG(WARNING) << "BA Msg output is abnormal!" << std::endl;
						}

						// judge if lower set speed actively
						if (gotoUpdtGlbInfo == true /* && globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd */)
						{
							AdjSetSpd(globalInfoProc);
						}
					    DLOG(INFO) << "set speed to trplongt = " << this->msg_BA2TRPLongt.ba_trp_v_sv_setspd << std::endl;
					}
					else
					{
						this->msg_BA2TRPLat.ba_trp_lat_sts = BA::TRPLatSts_None;
						this->msg_BA2TRPLat.ba_lat_actv_func_sts = BA::LatActvFuncSts_None;
						this->msg_BA2TRPLongt.ba_trp_indx_trp_lgt_sts = BA::TRPLongtSts_None;
						this->msg_BA2TRPLongt.ba_trp_indx_opt_action = BA::BA_ActionTyp_None;
						this->BA_LastActTyp2SM = BA::BA_ActionTyp_None; // reset last action type to SM
						//20220628: reset necessary output signals' value
						this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_None;
						this->msg_BA2SM.ba_opt_corridor_id = uint32_t(0);
						this->msg_BA2TRPLat.ba_opt_action = BA::BA_ActionTyp_None;
						this->msg_BA2TRPLat.ba_opt_corridor_id = uint32_t(0);
						this->BA_LaneChgAblFlg4TRPLat = false;
						this->BA_TRPLatSts = BA::TRPLatSts_Comfort;
					}
					this->BA_TerminatedBehaviorNodesVector->clear(); // clear terminatedBehaviorNodesVector
					// 20220328: add takeover request in lane end situation
					// if vehicle is on the ending lane, and the dist btw vehicle and lane end point is less than 5 sec,
					// then send take-over request
					// 20221114: add new element to transit_lane_list in spec case
					bool isEgoLnLaneEndFound = false;
					for (auto transitLn : this->input_BA.EHR_EHRInfo.transit_lanes_list)
					{
						if (transitLn.transit_type == BA::TransitionTyp_LaneEnd)
						{
							if (transitLn.transit_lane_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id)
							{
								isEgoLnLaneEndFound = true;
								break;
							}
						}
					}
					if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd && isEgoLnLaneEndFound == false)
					{
						uint8_t lftNgbrLnID = uint8_t(0);
						uint8_t ego_LneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
						bool egoLnIsEnding = false;
						float egoLnEndRng = 400.0f;
						float lftNgrLaneEndRng = 400.0f;
						uint8_t lftNgrLaneTyp = uint8_t(0);
						auto iter_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
							 					 	 [ego_LneID](const Lane &lane_)
												 	 {
														 return (ego_LneID == lane_.start_id);
												 	 });
						if (iter_lane != this->input_BA.RM_RoadModel.lane_list.end())
						{
							if (iter_lane->is_end_out_of_range == false && iter_lane->end_range <= 300.0f)
							{
								egoLnIsEnding = true;
								egoLnEndRng = iter_lane->end_range;
							}
						}
						if (egoLnIsEnding == true)
						{
							for (auto ngbrLn : it_lane->left_neighbor_lane)
							{
								if (ngbrLn.start_range < 0.0f)
								{
									lftNgbrLnID = ngbrLn.lane_id;
									auto it_lftNgbrLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		        			           					 [lftNgbrLnID](const Lane &lane_)
                     		    			      	 				 {
                                					  	 			    return (lftNgbrLnID == lane_.start_id);
                              		   								 });
									if (it_lftNgbrLn != this->input_BA.RM_RoadModel.lane_list.end())
									{
										lftNgrLaneEndRng = it_lftNgbrLn->end_range;
										lftNgrLaneTyp = it_lftNgbrLn->lane_type;
									}
									break;
								}
							}
							if (egoLnEndRng < lftNgrLaneEndRng && lftNgrLaneTyp != uint8_t(5) && lftNgrLaneTyp != uint8_t(11) && lftNgrLaneTyp != uint8_t(21))
							{
								TransitLaneStruct tmpTransitLnStruct;
								tmpTransitLnStruct.transit_lane_id = ego_LneID;
								tmpTransitLnStruct.transit_connected_lane_id = lftNgbrLnID;
								tmpTransitLnStruct.transit_lane_offset = egoLnEndRng;
								tmpTransitLnStruct.transit_type = BA::TransitionTyp_LaneEnd;
								this->input_BA.EHR_EHRInfo.transit_lanes_list.emplace_back(tmpTransitLnStruct);
							}
						}
					}
					
					if (!this->input_BA.EHR_EHRInfo.transit_lanes_list.empty())
					{
						for (auto transitLn : this->input_BA.EHR_EHRInfo.transit_lanes_list)
						{
							if (transitLn.transit_type == BA::TransitionTyp_LaneEnd)
							{
								// ego lane id is equal to transit lane id
								if ((this->input_BA.RM_RoadModel.car_position.ego_lane_id == transitLn.transit_lane_id))
								{
									// 20220722: add lane width condition
									float dstToEgoLnRghtBndry = 2.0;
									float dstToLftNgbLnRghtBndry = 2.0;
									// float dstToEgoLnLftBndry = 2.0;
									// float dstToRghtNgbLnLftBndry = 2.0;
									float egoLnWidth = dstToEgoLnRghtBndry + dstToLftNgbLnRghtBndry;
									uint8_t egoLneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
									auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   							[egoLneID](const Lane &lane_)
                     		           							{
                                		   							return (egoLneID == lane_.start_id);
                              		   							});
									bool lftLnCrossable = true;
									bool rghtLnCrossable = true;
									if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left &&
										this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange)
									{
										bool getDstToEgoLnRghtBndry = CalcDstToRLnBndry(dstToEgoLnRghtBndry, egoLneID);
										// get left neighbor lane id
										uint8_t lftNgbrLnID = uint8_t(0);
 										if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
										{
											if (!it_lane->left_neighbor_lane.empty())
											{
												for (auto lftLn : it_lane->left_neighbor_lane)
												{
													if (lftLn.start_range < 0.0f)
													{
														lftNgbrLnID = lftLn.lane_id;
													}
												}
											}
											if (lftNgbrLnID != uint8_t(0))
											{
												DLOG(INFO) << "lftNgbrLnID = " << int(lftNgbrLnID) << std::endl;
												bool getDstToLftNgbLnRghtBndr = CalcDstToRLnBndry(dstToLftNgbLnRghtBndry, lftNgbrLnID);
												if (getDstToEgoLnRghtBndry == true && getDstToLftNgbLnRghtBndr == true)
												{
													egoLnWidth = std::abs(dstToEgoLnRghtBndry) + std::abs(dstToLftNgbLnRghtBndry);
												}
											}
										}
    									else
    									{
											DLOG(INFO) << "current left lane marker is empty!" << std::endl;
    									}
									}
									/* else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right &&
											 this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange)
									{} */
									// 20220808:
									if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
  									{
										// get first lane marker id of both sides
    									uint32_t currLftLnMrkID = uint32_t(0);
    									if (!it_lane->left_lane_marker_id_list.empty())
    									{
      										currLftLnMrkID = it_lane->left_lane_marker_id_list.front();
											uint8_t currLftLnTyp = uint8_t(0); // unknown
											if (GetLineTyp(currLftLnMrkID, currLftLnTyp) == true)
											{
												if (!currLftLnTyp == 2 && !currLftLnTyp == 0)
												{
													lftLnCrossable = false;
												}
											}
										}
										uint32_t currRghtLnMrkID = uint32_t(0);
    									if (!it_lane->right_lane_marker_id_list.empty())
    									{
      										currRghtLnMrkID = it_lane->right_lane_marker_id_list.front();
											uint8_t currRghtLnTyp = uint8_t(0); // unknown
											if (GetLineTyp(currRghtLnMrkID, currRghtLnTyp) == true)
											{
												if (!currRghtLnTyp == 2 && !currRghtLnTyp == 0)
												{
													rghtLnCrossable = false;
												}
											}
										}
    								}
									DLOG(INFO) << "ego lane width = " << egoLnWidth << std::endl;
									DLOG(INFO) << "dstToLftNgbLnRghtBndry = " << dstToLftNgbLnRghtBndry << std::endl;
									DLOG(INFO) << "dstToEgoLnRghtBndry = " << dstToEgoLnRghtBndry << std::endl;
									DLOG(INFO) << "left line is crossable = " << int(lftLnCrossable) << std::endl;
									if (this->BA_TORTyp == uint8_t(0) &&
										(lftLnCrossable == true || rghtLnCrossable == true) &&
										((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto &&
										  this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
										  this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange) ||
										 (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto) ||
										 (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_None && this->input_BA.SM_LatActvFunc == BA::LatActvFuncSts_NRP) ) &&
									    ( transitLn.transit_lane_offset <= std::max(this->input_BA.VSE_LongitudinalVelocity * ADC::common::k_BA_LnEndTHThr, ADC::common::k_BA_MinLaneEndLCPntDst) || 
										  ((egoLnWidth < ADC::common::k_BA_MinEgoLnWidthThr || std::abs(dstToLftNgbLnRghtBndry) < ADC::common::k_BA_MinEgoLnWidthThr * 0.45f)) && transitLn.transit_lane_offset < ADC::common::k_BA_MinLaneEndLCPntDst02))
									{
										this->BA_TORTyp = uint8_t(2);
										if (this->msg_BA2TRPLongt.ba_trp_v_sv_setspd > std::max(0.0f, (ADC::common::k_BA_LastLnEndSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f)
										{
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_LastLnEndSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
										}
										DLOG(INFO) << "   Takeover request: lane is ending!" << std::endl;
										DLOG(INFO) << "   TOR lane id = " << int(transitLn.transit_lane_id) << std::endl;
										DLOG(INFO) << "   TOR lane offset = " << transitLn.transit_lane_offset << std::endl;
										break;
									}
									// 20220920: add warning takeover request ahead of function exit takeover request
									if (this->BA_TORTyp == uint8_t(0) &&
										(lftLnCrossable == true || rghtLnCrossable == true) &&
										((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto &&
										  this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
										  this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange) ||
										 (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto) ||
										 (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_None && this->input_BA.SM_LatActvFunc == BA::LatActvFuncSts_NRP) ) &&
									     transitLn.transit_lane_offset <= std::max(this->input_BA.VSE_LongitudinalVelocity * ADC::common::k_BA_LnEndTHThr01, ADC::common::k_BA_MinLaneEndLCPntDst01) &&
										 ADC::common::k_BA_LaneEndWrnInFlg == true)
									{
										this->BA_TORTyp = uint8_t(5);
										DLOG(INFO) << "   Takeover request: lane is ending 01 for warning!" << std::endl;
										DLOG(INFO) << "   TOR lane id = " << int(transitLn.transit_lane_id) << std::endl;
										DLOG(INFO) << "   TOR lane offset = " << transitLn.transit_lane_offset << std::endl;
										break;
									}
								}
							}
						}
					}
					// 20220328: add takeover request when missing last lane change point
					// if control way point is JCTRd, ICRd or RmpDvdRd, check ego lane's last lane change point
					// 20220702: add SpltRd
					// 20220728: add subSpltRd
					BA::TargetRoadTyp trgtRdTyp = globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP;
					if ((trgtRdTyp == BA::TargetRoadTyp_ICRd) || (trgtRdTyp == BA::TargetRoadTyp_JCTRd) || (trgtRdTyp == BA::TargetRoadTyp_RmpDvdRd) || (trgtRdTyp == BA::TargetRoadTyp_SpltRd) ||
					    (globalInfoProc.globalInfoContainer.m_GlobalInfoSrc == BA::GlobalInfoSrc_OnlyRmdWayPoint && globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd))
					{
						if (!globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.empty())
						{
							for (auto glbLnInfo : globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr)
							{
								if (glbLnInfo.m_LaneID == this->input_BA.RM_RoadModel.car_position.ego_lane_id)
								{
									if (!glbLnInfo.m_LastLaneChangePoint.empty())
									{
										if (glbLnInfo.m_LastLaneChangePoint.front() > globalInfoProc.globalInfoContainer.m_DistToWayPoint && globalInfoProc.globalInfoContainer.m_DistToWayPoint > 3)
										{
											// 20220720: suppress missing last lane change point takeover request in certain case
											// ego lane is not target lane 
											// ego lane belongs to raw target lane 
											// first control waypoint type is JCTRd OR ICRd
											// second control waypoint type is rmpDvdRd
											// 20220902: if ego lane's endoffset is less than 150m
											bool tmpFlg = false;
											auto it_inTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), glbLnInfo.m_LaneID);
											if (this->input_BA.EHR_EHRInfo.control_waypoints_list.size() == 2)
											{
												auto it_inRawTrgtLn = std::find(this->input_BA.EHR_EHRInfo.control_waypoints_list.front().control_waypoint_raw_lane_id_list.begin(), 
												                                this->input_BA.EHR_EHRInfo.control_waypoints_list.front().control_waypoint_raw_lane_id_list.end(), 
																				glbLnInfo.m_LaneID);
												if (it_inTrgtLn == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end() &&
											    	it_inRawTrgtLn != this->input_BA.EHR_EHRInfo.control_waypoints_list.front().control_waypoint_raw_lane_id_list.end() &&
													((trgtRdTyp == BA::TargetRoadTyp_ICRd) || (trgtRdTyp == BA::TargetRoadTyp_JCTRd || (trgtRdTyp == BA::TargetRoadTyp_RmpDvdRd))) &&
													 (this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_RmpDvdRd ||
													  this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_RmpMrgCARd ||
													  this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_JCTRd))
											    {
													tmpFlg = true;
													DLOG(INFO) << "@@ last lane change point missing takeover request suppressed!" << std::endl;
												}
											}
											uint8_t egoLaneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
											auto it_lne = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   									[egoLaneID](const Lane &lane_)
                     		           									{
                                		   									return (egoLaneID == lane_.start_id);
                              		   									});
											if (it_lne != this->input_BA.RM_RoadModel.lane_list.end())
											{
												float egoLnEndOffset = it_lne->end_range;
												if (egoLnEndOffset < ADC::common::k_BA_SpprTOR3EndOffsetThr &&
												    globalInfoProc.globalInfoContainer.m_DistToWayPoint > ADC::common::k_BA_SpprTOR3Dist2WpThr)
												{
													tmpFlg = true;
													DLOG(INFO) << "@@ last lane change point missing takeover request suppressed 02!" << std::endl;
												}
											}
											// 20220913: add optimal action and distance to waypoint
											if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange ||
												 this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange ) &&
												globalInfoProc.globalInfoContainer.m_DistToWayPoint > ADC::common::k_BA_SpprTOR3Dist2WpLowThr)
											{
												tmpFlg = true;
												DLOG(INFO) << "@@ last lane change point missing takeover request suppressed 03!" << std::endl;
											}
											if (this->BA_TORTyp == uint8_t(0) && tmpFlg == false)
											{
												this->BA_TORTyp = uint8_t(3);
												DLOG(INFO) << "   Takeover request: last lane change point is missing!" << std::endl;
												DLOG(INFO) << "   TOR lane id = " << int(glbLnInfo.m_LaneID) << std::endl;
												DLOG(INFO) << "   TOR last lane change point = " << glbLnInfo.m_LastLaneChangePoint.front() << std::endl;
												break;
											}
										}
									}
								}
							}
						}
					}

					// 20220822: 
					// if lcavlbl equals to right laneChangeAuto or left laneChangeAuto
					// then according to ego lane end range, determine whether to send lane end takeover request
					if ((this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto) &&
						this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
						this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange &&
					    this->BA_TORTyp == uint8_t(0))
					{
						uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
						auto it_egoLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
						                               [egoLnID](const Lane &ln_)
													   {
														   return (egoLnID == ln_.start_id);
													   });
						if (it_egoLane != this->input_BA.RM_RoadModel.lane_list.end())
						{
							float egoLnEndRng = it_egoLane->end_range;
							if (egoLnEndRng <= std::max(this->input_BA.VSE_LongitudinalVelocity * ADC::common::k_BA_LnEndTHThr, ADC::common::k_BA_MinLaneEndLCPntDst))
							{
								this->BA_TORTyp = uint8_t(2);
								if (this->msg_BA2TRPLongt.ba_trp_v_sv_setspd > std::max(0.0f, (ADC::common::k_BA_LastLnEndSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f)
								{
									this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_LastLnEndSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
								}
								DLOG(INFO) << "Command lane change under NRP and lane is ending" << std::endl;
								DLOG(INFO) << "   TOR lane id = " << int(egoLnID) << std::endl;
								DLOG(INFO) << "   TOR lane offset = " << egoLnEndRng << std::endl;
							}
						}
					}

					// 20221103:
					// ego car is executing lane change and lane change abolition is not available, while PD return a risky target
					// then set BA_TORTyp to 4
					if (this->BA_TORTyp == uint8_t(0) &&
					    this->msg_BA2SM.ba_elk_tgt_id != uint8_t(0))
					{
						this->BA_TORTyp = uint8_t(1);
					}
				}
			}
			else
			{
				this->msg_BA2TRPLat.ba_trp_lat_sts = BA::TRPLatSts_None;
				this->msg_BA2TRPLat.ba_lat_actv_func_sts = BA::LatActvFuncSts_None;
				this->msg_BA2TRPLat.ba_opt_action = BA::BA_ActionTyp_None;
				this->msg_BA2TRPLongt.ba_trp_indx_trp_lgt_sts = BA::TRPLongtSts_None;
				this->msg_BA2TRPLongt.ba_trp_indx_opt_action = BA::BA_ActionTyp_None;
				this->msg_BA2TRPLongt.ba_trp_is_esaa = false;
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_None;
				this->BA_DircnIndReq = BA::DircnIndReqSts_BothOff;
				this->BA_DircnIndReq_Deb = BA::DircnIndReqSts_BothOff;
				this->BA_AvgFlowDenseHigh = false;
				this->BA_LastActTyp2SM = BA::BA_ActionTyp_None;
				this->BA_TORTyp = uint8_t(0);
				this->BA_FuncOnDurTm = 0.0f;
				this->BA_LaneChgAblFlg4TRPLat = false;
				this->BA_TmpTrgtRdTypMem = BA::TargetRoadTyp_NotAvlbl;
				this->BA_Wait4LCDurTm = 0.0f;
				// if there is no available behavior, then find out default object for TRPLongt
				uint32_t dfltObjID = 0;
				float dfltObjDst = 1000;
				float dfltObjVel = 0;
				Corridor dfltObjBlgCrrd;

				bool isLineFllwCrrdFound = false;
				bool flg = GetTrgtObjInfo(dfltObjID, dfltObjDst, dfltObjVel, dfltObjBlgCrrd, isLineFllwCrrdFound);
				if (flg == true)
				{
					this->msg_BA2TRPLongt.ba_trp_indx_tgt_id_iv = dfltObjID;
					this->msg_BA2TRPLongt.ba_trp_v_tgt_longt_spd_iv = dfltObjVel;
					if (dfltObjID != 0)
					{
						this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = true;
						this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = dfltObjDst - ADC::common::k_BA_DistBtwVehFrt2BackAxis;
					}
					else
					{
						this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = false;
						this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = dfltObjDst;
					}
					// other attributes of default object are not extracted
				}
				else
				{
					this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = 1000;
					this->msg_BA2TRPLongt.ba_trp_indx_tgt_id_iv = 0;
					this->msg_BA2TRPLongt.ba_trp_v_tgt_longt_spd_iv = 0;
					this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = false;
				}
				
			}
			DebounceTOR();
			this->msg_BA2SM.ba_takeover_req = this->BA_TORTyp_Deb;
			DLOG(INFO) << "@@@ BA_TORTyp = " << int(this->BA_TORTyp) << std::endl;
			DLOG(INFO) << "@@@ BA_TORTyp_Deb = " << int(this->BA_TORTyp_Deb) << std::endl;


		}

		void BehaviorAnalysis::UpdateBaseInfo()
		{
			this->input_BA.BA_OptAction_LastCycle = BA::BA_ActionTyp(this->BA_OptAction_Deb);
			this->input_BA.BA_OptCorridorID_LastCycle = this->BA_OptCorridorID_Deb;
			this->input_BA.BA_OptTargetObjID_LastCycle = this->BA_OptTargetObjID_Deb;
			this->input_BA.BA_OptLaneID_LastCycle = this->BA_OptLaneID_Deb;
			this->input_BA.BA_TgtLaneStID_LastCycle = this->BA_TgtLaneStID;
			this->input_BA.BA_TgtLaneEndID_LastCycle = this->BA_TgtLaneEndID;
			BA::Index optPssgID;
			optPssgID.m_X = this->BA_OptPassageID_Deb.m_X;
			optPssgID.m_Y = this->BA_OptPassageID_Deb.m_Y;
			auto it_passage = std::find_if(this->BA_FirstPassageList->begin(), this->BA_FirstPassageList->end(),
										   [optPssgID](const Passage &passage_)
										   {
											   return (passage_.passage_id.x == optPssgID.m_X && passage_.passage_id.y == optPssgID.m_Y);
										   });
			if (it_passage != this->BA_FirstPassageList->end())
			{
				this->input_BA.BA_LCCls_LastCycle = it_passage->lc_cls;
				this->input_BA.BA_LCPCls_LastCycle = it_passage->lcp_cls;
			}
			this->input_BA.BA_SpdLimitFlg_LastCycle = this->BA_OptimalSpdLmtFlg;
			this->input_BA.BA_SpdLimitID_LastCycle = this->BA_OptimalSpdLmtID;
			this->input_BA.BA_SpdLimitType_LastCycle = this->BA_OptimalSpdLmtTyp;
		}

		void BehaviorAnalysis::UpdateRespInfo(PD::PDMdOut &pd_MdOut)
		{
			this->response_BA.PD_BehaviorNodeLocalVector = pd_MdOut.pd_behavior_nodelocal_vector;
			this->response_BA.PD_RoadModelVector = pd_MdOut.pd_roadrodel_vector;
		}

		void BehaviorAnalysis::ExpandBehaviorSequence(std::vector<BA::BA_ActionTyp> &latAvlblBhvr)
		{
			// if Behavior node update has not completed, then continue to expand unfinished behavior sequences
			BehaviorNodesAdd behaviorNodesAdd_;
			RoadModelUpdate roadModelUpdate_;
			DM::DirectionIndicatorDuration dircnIndDur_;
			std::vector<dpc::Index> roadModelID2PD_;

			while (!BA_BehaviorNodeUpdateDone)
			{
				behaviorNodesAdd_.Init(this->BA_BehaviorNodeGlobalVector.get(),
									   this->response_BA.PD_BehaviorNodeLocalVector,
									   this->BA_RoadModelVector.get(),
									   this->BA_StepNum,
									   this->input_BA.BA_OptAction_LastCycle,
									   this->input_BA.BA_OptCorridorID_LastCycle,
									   this->input_BA.BA_OptTargetObjID_LastCycle,
									   this->input_BA,
									   latAvlblBhvr);
				// behavior sequence termination judge
				if (!JudgeBehaviorSeqTermination())
				{
					DLOG(WARNING) << "Behavior sequence termination judgement failed." << std::endl;
				}

				// judge whether to go out of the circle
				if (this->BA_BehaviorNodeUpdateDone == true)
				{
					// reset BA_StepNum
					this->BA_StepNum = -1;
					break;
				}

				// action and corridor combination
				// signal preparation inside
				DLOG(INFO) << "SM_LFComfortAvlblBehavior : " << int(this->input_BA.SM_LFComfortAvlblBehavior) << std::endl;
				DLOG(INFO) << "SM_LCAvlblBehavior : " << int(this->input_BA.SM_LCAvlblBehavior) << std::endl;
				DLOG(INFO) << "SM_LCPAvlblBehavior : " << int(this->input_BA.SM_LCPAvlblBehavior) << std::endl;
				DLOG(INFO) << std::endl;
				// 20221025: replace input_BA.SM_DriveMode by BA_TmpDriveMode
				// TmpDriveMode may be different from SM_DriveMode under LCR
				roadModelUpdate_.Init(this->BA_StepNum,
									  BA_RoadModelVector.get(),
									  BA_BehaviorNodeGlobalVector.get(),
									  this->BA_DirectionIndicatorInfo,
									  this->BA_BehaviorStackVector,
									  this->input_BA.SM_LFComfortAvlblBehavior,
									  this->input_BA.SM_LCAvlblBehavior,
									  this->input_BA.SM_LCPAvlblBehavior,
									  this->BA_LaneChangeAbolitionAvlbl,
									  this->BA_VehInLnCenter,
									  this->input_BA.SM_DrvrSetSpd,
									  this->input_BA.VSE_LongitudinalVelocity,
									  this->BA_SupprActLst,
									  this->BA_PassageObLst,
									  this->BA_VehClsToLnCntr,
									  this->BA_LastActTyp2SM,
									  this->input_BA.SM_DrvrSetDstLvl,
									  this->input_BA.SM_AutoLnChgAvlbl,
									  this->input_BA.SM_SignSpeedLimit01,
									  this->input_BA.SM_SignSpeedLimit02,
									  this->input_BA.BA_SpdLimitFlg_LastCycle,
									  this->input_BA.BA_SpdLimitType_LastCycle,
									  this->input_BA.BA_SpdLimitID_LastCycle,
									  this->BA_DynChgLCintervalLvl,
									  this->BA_ISLCLnIDnotChg,
									  this->BA_TmpDriveMd,
									  this->BA_CrrtScenarioTyp);
				// BA_FirstPassageList fill in
				if (this->BA_StepNum == 0)
				{
					this->BA_FirstPassageList->clear();
					if (this->BA_RoadModelVector->empty())
					{
						DLOG(WARNING) << "BA_RoadModelVector is empty!" << std::endl;
					}
					else
					{
						this->BA_FirstPassageList->resize(this->BA_RoadModelVector->front().passage_list.size());
						copy(this->BA_RoadModelVector->front().passage_list.begin(), this->BA_RoadModelVector->front().passage_list.end(), this->BA_FirstPassageList->begin());
					}

					DLOG(INFO) << "***first passage list size: " << this->BA_FirstPassageList->size() << std::endl;
					if (!this->BA_FirstPassageList->empty())
					{
						for (auto firstPsg : *this->BA_FirstPassageList.get())
						{
							DLOG(INFO) << "  passage id = ( " << int(firstPsg.passage_id.x) << ", " << int(firstPsg.passage_id.y) << " )" << std::endl;
							DLOG(INFO) << "  passage action: " << int(firstPsg.action_type) << std::endl;
							DLOG(INFO) << std::endl;
						}
					}
				}
				// BA_PassageListObs fill in
				// currently store first road model inside
				if (this->BA_StepNum == 0)
				{
					this->BA_PassageListObs.clear();
					for (auto roadmodel : *this->BA_RoadModelVector.get())
					{
						if (!roadmodel.passage_list_ob.empty())
						{
//							this->BA_PassageListObs.insert(this->BA_PassageListObs.begin(), roadmodel.passage_list_ob.begin(), roadmodel.passage_list_ob.end());
						}
					}
				}

				// behavior node global vector update
				if (!UpdateBehaviorNodeGlobalVector())
				{
					DLOG(WARNING) << "Behavior node global vector update failed." << std::endl;
				}

				// clear roadModelID2PD_ before assigning new value
				roadModelID2PD_.clear();
				if (!GetRoadModelIDToPD(roadModelID2PD_))
				{
					DLOG(WARNING) << "Road model ID getting failed." << std::endl;
				}

				// wrap up service request to PD
				PD::PDInputBa pd_InputBa;
				std::vector<RoadInfo> roadModelVecTmp(*this->BA_RoadModelVector);
				pd_InputBa.ba_roadmodel_vector.clear();
				pd_InputBa.ba_roadmodel_vector.resize(roadModelVecTmp.size());
				copy(roadModelVecTmp.begin(), roadModelVecTmp.end(), pd_InputBa.ba_roadmodel_vector.begin());
				std::vector<dpc::Index> roadModelIDVecTmp(roadModelID2PD_);
				pd_InputBa.ba_roadmodel_id2pd.clear();
				pd_InputBa.ba_roadmodel_id2pd.resize(roadModelIDVecTmp.size());
				copy(roadModelIDVecTmp.begin(), roadModelIDVecTmp.end(), pd_InputBa.ba_roadmodel_id2pd.begin());
				pd_InputBa.ba_behavior_node_updatedone = false;

				// call PD sub-function
				PD::PDMdOut pd_MdOut;
				PD::PDVSE pd_vse;
				pd_vse.m_vse_heading_ang = 0;
				pd_vse.m_vse_lat_acc = this->input_BA.VSE_LatAcc;
				pd_vse.m_vse_lgt_acc = this->input_BA.VSE_LongtAcc;
				pd_vse.m_vse_lgt_vehspd = this->input_BA.VSE_LongitudinalVelocity;
				pd_vse.m_vse_yawrate = this->input_BA.VSE_Yawrate; //unit: deg/s
				predict::Prediction _prediction;
				PD_SubFunc::PD_FuncRun pd_FuncRun(_prediction);

				#ifdef GLOG
				DLOG(INFO) << "=========================" << std::endl;
				DLOG(INFO) << "BA_StepNum = " << int(this->BA_StepNum) << std::endl;
				DLOG(INFO) << "BA RoadModel Vector size: " << pd_InputBa.ba_roadmodel_vector.size() << std::endl;
				if (pd_InputBa.ba_roadmodel_vector.size() > 0)
				{
					for (auto rm : pd_InputBa.ba_roadmodel_vector)
					{
						DLOG(INFO) << "road model id = ( " << int(rm.road_id.x) << ", " << int(rm.road_id.y) << " )" << std::endl;
						DLOG(INFO) << "  passage number = " << rm.passage_list.size() << std::endl;
						if (rm.passage_list.size() > 0)
						{
							for (auto psg : rm.passage_list)
							{
								DLOG(INFO) << "  passage id = ( " << int(psg.passage_id.x) << ", " << int(psg.passage_id.y) << " )" << std::endl;
								DLOG(INFO) << "  passage action = " << int(psg.action_type) << std::endl;
								DLOG(INFO) << "  passage duration time = " << psg.duration_time << std::endl;
								DLOG(INFO) << "  passage desired dist = " << psg.longt_info.desired_distance << std::endl;
								DLOG(INFO) << "  passage desired spd = " << psg.longt_info.desired_speed << std::endl;
								DLOG(INFO) << "  passage set spd = " << psg.longt_info.set_speed << std::endl;
								DLOG(INFO) << std::endl;
							}
						}
					}
				}
				DLOG(INFO) << "--------------------------" << std::endl;
				#endif

				uint8_t spdLmtTyp_ = this->input_BA.BA_SpdLimitType_LastCycle;
				pd_FuncRun.PD_FuncCall(pd_InputBa,
				                       pd_vse, 
				                       this->input_BA.BA_SpdLimitFlg_LastCycle,
									   this->input_BA.BA_SpdLimitID_LastCycle,
									   spdLmtTyp_);
				DLOG(INFO) << "pd_vse = " << pd_vse.m_vse_lgt_vehspd << std::endl;
				DLOG(INFO) << "PD RoadModel Vector size: " << pd_FuncRun._prediction.res.pd_roadrodel_vector.size() << std::endl;
				/* if (pd_FuncRun._prediction.res.pd_roadrodel_vector.size() > 0)
				{
					for (auto rm : pd_FuncRun._prediction.res.pd_roadrodel_vector)
					{
						DLOG(INFO) << "road model id = ( " << int(rm.road_id.x) << ", " << int(rm.road_id.y) << " )" << std::endl;
						DLOG(INFO) << "  ego lane id = " << int(rm.car_position.ego_lane_id) << std::endl;
						DLOG(INFO) << "  lane list size = " << int(rm.lane_list.size()) << std::endl;
						DLOG(INFO) << "  lane id are showed below: " << std::endl;
						int laneNum = 0;
						for (auto ln : rm.lane_list)
						{
							laneNum++;
							DLOG(INFO) << "   lane " << laneNum << " id = " << int(ln.end_id) << std::endl;
							if (!ln.right_neighbor_lane.empty())
							{
								DLOG(INFO) << "   right neighbor lane id = " << int(ln.right_neighbor_lane.front().lane_id) << std::endl;
								DLOG(INFO) << "   start offset = " << ln.right_neighbor_lane.front().start_range << std::endl;
								DLOG(INFO) << "   end offset = " << ln.right_neighbor_lane.front().end_range << std::endl;
							}
							else
							{
								DLOG(INFO) << "   no right neighbor lane" << std::endl;
							}
							if (!ln.left_neighbor_lane.empty())
							{
								DLOG(INFO) << "   left neighbor lane id = " << int(ln.left_neighbor_lane.front().lane_id) << std::endl;
								DLOG(INFO) << "   start offset = " << ln.left_neighbor_lane.front().start_range << std::endl;
								DLOG(INFO) << "   end offset = " << ln.left_neighbor_lane.front().end_range << std::endl;
							}
							else
							{
								DLOG(INFO) << "   no left neighbor lane" << std::endl;
							}
							DLOG(INFO) << "   -- corridor num = " << int(ln.corridor_list.size()) << std::endl;
							int crrdNum = 0; 
							for (auto crrd : ln.corridor_list)
							{
								crrdNum++;
								DLOG(INFO) << "     corridor " << crrdNum << " id = " << int(crrd.corridor_id) << std::endl;
								DLOG(INFO) << "     corridor feature tag = " << int(crrd.feature_tag) << std::endl;
								DLOG(INFO) << "     object in corridor size = " << int(crrd.object_in_corridor_list.size()) << std::endl;
							}
						}
						DLOG(INFO) << "  object num = " << int(rm.object_list.size()) << std::endl;
					}
				} */
				DLOG(INFO) << "--------------------------" << std::endl;
				DLOG(INFO) << std::endl;

				#ifdef GLOG
				DLOG(INFO) << "PD BehaviorNodeLocal size: " << pd_FuncRun._prediction.res.pd_behavior_nodelocal_vector.size() << std::endl;
				if (pd_FuncRun._prediction.res.pd_roadrodel_vector.size() > 0)
				{
					for (auto bnl : pd_FuncRun._prediction.res.pd_behavior_nodelocal_vector)
					{
						DLOG(INFO) << "bnl id = ( " << int(bnl.m_node_id.x) << ", " << int(bnl.m_node_id.y) << " )" << std::endl;
						DLOG(INFO) << "bnl parent id = ( " << int(bnl.m_parent_node_id.x) << ", " << int(bnl.m_parent_node_id.y) << " )" << std::endl;
						DLOG(INFO) << "bnl clsn sts = " << int(bnl.m_node_clsn_sts) << std::endl;
						if (!bnl.m_hvinfo.empty())
						{
							// DLOG(INFO) << "init s = " << bnl.m_hvinfo.at(0).m_distance_alongs << std::endl;
							// DLOG(INFO) << "init v = " << bnl.m_hvinfo.at(0).m_Velocity << std::endl;
							DLOG(INFO) << "end s = " << bnl.m_hvinfo.back().m_distance_alongs << std::endl;
							DLOG(INFO) << "end v = " << bnl.m_hvinfo.back().m_velocity << std::endl;
						}
						DLOG(INFO) << "bnl spd limit flag = " << int(bnl.spd_limit_flg) << std::endl;
						DLOG(INFO) << "bnl spd limit type = " << int(bnl.spd_limit_type) << std::endl;
					}
				}
				DLOG(INFO) << std::endl;
				#endif
				UpdateRespInfo(pd_FuncRun._prediction.res);
				// 20220707: PD debug msg stored here
				#ifdef PD_Debug
				if (this->BA_StepNum == 0)
				{
					if (!pd_FuncRun._prediction.PD_DebugInfo.empty())
					{
						UpdatePDDbgInfo(pd_FuncRun._prediction.PD_DebugInfo.front());
					}
					else
					{
						DLOG(INFO) << "PD_DebugInfo is empty!" << std::endl;
					}
				}
				#endif
			}
		}

		bool BehaviorAnalysis::LabelOvertimeBehaviorNode()
		{
			if (BA_StepNum == 0)
			{
				BA_RoadModelVector->clear(); // clear BA_RoadModelVector at the beginning of new cycle
				// 20211203: modify condition used to determine whether input_BA.RM_RoadModel is updated
				if (!this->input_BA.RM_RoadModel.lane_list.empty())
				{
					this->BA_RoadModelVector->push_back(this->input_BA.RM_RoadModel);
				}
				else
				{
					DLOG(WARNING) << "BA_RoadModelVector is empty!" << std::endl;
				}
			}
			else
			{
				// clear BA_RoadModelVector, then fill BA_RoadModelVector with RM_RoadModel or PD_RoadModelVector
				if (!BA_RoadModelVector->empty())
				{
					BA_RoadModelVector->clear();
				}
				else
				{
					DLOG(WARNING) << "BA_RoadModelVector is abnormal." << std::endl;
					return false;
				}

				BA_RoadModelVector->insert((*BA_RoadModelVector).end(), response_BA.PD_RoadModelVector.begin(), response_BA.PD_RoadModelVector.end());

				// clear passage list in BA_RoadModelVector
				if (!BA_RoadModelVector->empty())
				{
					for (auto iter = BA_RoadModelVector->begin(); iter != BA_RoadModelVector->end(); ++iter)
					{
						iter->passage_list.clear();
					}
				}
			}

			// foreach behaviorNodeGlobal in BA_BehaviorNodeGlobalVector, check whether total duration time has exceeded preset threshold
			// overtime behaviorNodeGlobal object will be moved to BA_TerminatedBehaviorNodesVector
			if (!BA_BehaviorNodeGlobalVector->empty())
			{
				for (std::vector<BA::behaviorNodeGlobal>::iterator iter = BA_BehaviorNodeGlobalVector->begin();
					 iter != BA_BehaviorNodeGlobalVector->end();)
				{
					if (iter->totalDurationTime() >= k_BA_TotalDurationTimeThreshold)
					{
						// iter->m_TotalCost = 0;
						iter->m_IsNodeTerminated = true;
						BA_TerminatedBehaviorNodesVector->emplace_back(*iter);
						BA_BehaviorNodeGlobalVector->erase(iter);
						if (iter == BA_BehaviorNodeGlobalVector->end())
							break;
					}
					else
					{
						iter++;
					}
				}
			}
			else
			{
				DLOG(WARNING) << "BA_BehaviorNodeGlobalVector is empty!" << std::endl;
			}
			return true;
		}

		void BehaviorAnalysis::JudgeBehaviorNodeTermination()
		{
			if (BA_BehaviorNodeGlobalVector->empty())
			{
				BA_BehaviorNodeUpdateDone = true;
			}
			else
			{
				if (JudgeAllGlobalNodeExp())
					BA_BehaviorNodeUpdateDone = true;
				else
					BA_BehaviorNodeUpdateDone = false;
			}
		}

		bool BehaviorAnalysis::JudgeBehaviorSeqTermination()
		{
			LabelOvertimeBehaviorNode();
			JudgeBehaviorNodeTermination();
			return true;
		}

		bool BehaviorAnalysis::JudgeAllGlobalNodeExp()
		{
			if (!this->BA_BehaviorNodeGlobalVector->empty())
			{
				for (int i = 0; i < this->BA_BehaviorNodeGlobalVector->size(); i++)
				{
					if (this->BA_BehaviorNodeGlobalVector->at(i).m_NodeID.m_X == this->BA_StepNum)
					{
						return false;
					}
				}
				return true;
			}
			else
				return true;
		}

		void BehaviorAnalysis::DetermineOptimalCorridor(bool BA_IsBehaviorNotAvlbl,
														uint32_t &BA_OptCorridorID,
														BA::BA_ActionTyp &BA_OptAction,
														uint32_t &BA_OptTargetObjID,
														BA::Index &BA_OptPassageID,
														uint8_t &BA_TrgtLaneStID,
														uint8_t &BA_TrgtLaneEndID)
		{
			if (!BA_IsBehaviorNotAvlbl)
			{
				if (BA_TerminatedBehaviorNodesVector->empty())
				{
					DLOG(WARNING) << "BA_TerminatedBehaviorNodesVector is abnormal" << std::endl;
					// check whether lane follow is available, and send TOR
					uint32_t dfltObjID = 0;
					float dfltObjDst = 1000;
					float dfltObjVel = 0;
					Corridor dfltObjBlgCrrd;

					bool isLineFllwCrrdFound = false;
					bool flg = GetTrgtObjInfo(dfltObjID, dfltObjDst, dfltObjVel, dfltObjBlgCrrd, isLineFllwCrrdFound);
					// IF
					// flg equals to true
					if (flg == true)
					{
						if (dfltObjBlgCrrd.corridor_id == uint32_t(0))
						{
							// ego lane corridor is actually not found
							BA_OptAction = BA::BA_ActionTyp_None;
						}
						else
						{
							BA_OptCorridorID = dfltObjBlgCrrd.corridor_id;
							BA_OptAction = BA::BA_ActionTyp_LaneFollow;
							BA_OptTargetObjID = dfltObjID;
						}
					}
					// flg equals to false
					else
					{
						BA_OptAction = BA::BA_ActionTyp_None;
					}
					this->BA_TORTyp = uint8_t(1); // send TOR here, terminated behavior node vector is abnormal
					DLOG(INFO) << "   Takeover request: terminated behavior node vector is abnormal!" << std::endl;
					// 20220419
					// if vehicle is arriving ramp and the case condition is satisfied, then send the nearest sign speed limit info when dist is less than 500m
					// else if vehicle is not arriving ramp, then always send the nearest sign speed limit info
					// to be improved
					// 20220601
					// if BA_SpdLmtInfoFltred is true, then no speed limit info will be delivered
					if (this->BA_SpdLmtInfoFltred == false && 
					    this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtID != uint8_t(0) &&
						this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtID != uint8_t(255) &&
						this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtDist < 500.0f)
					{
						this->BA_OptimalSpdLmtTyp = BA::SpdLmtTyp_Sign1st;
						this->BA_OptimalSpdLmtDst = this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtDist;
						this->BA_OptimalSpdLmtVal = this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtVal;
						this->BA_OptimalSpdLmtID = this->input_BA.SM_SignSpeedLimit01.m_SignSpdLmtID;
						this->BA_OptimalSpdLmtFlg = true;
					}
					return;
				}

				if (!BA_TerminatedBehaviorNodesVector->empty())
				{
					DLOG(INFO) << "Terminated behavior nodes vector size = " << BA_TerminatedBehaviorNodesVector->size() << std::endl;
					int tbnNum = 0;
					for (auto tbn : *BA_TerminatedBehaviorNodesVector.get())
					{
						tbnNum++;
						DLOG(INFO) << "tbn order = " << tbnNum << std::endl;
						DLOG(INFO) << "  first action is " << int(tbn.firstActionTyp()) << std::endl;
						DLOG(INFO) << "  total cost is " << tbn.m_TotalCost << std::endl;
						DLOG(INFO) << "  BA calc cost is " << tbn.m_BACalcCost << std::endl;
						DLOG(INFO) << std::endl;
					}
				}

				std::vector<BA::behaviorNodeGlobal>::iterator BA_MinTotalCostIter = BA_TerminatedBehaviorNodesVector->begin();
				float BA_MinTotalCostVal = BA_MinTotalCostIter->totalCost();
				for (std::vector<BA::behaviorNodeGlobal>::iterator iter = BA_TerminatedBehaviorNodesVector->begin();
					 iter != BA_TerminatedBehaviorNodesVector->end(); iter++)
				{
					if (iter->totalCost() < BA_MinTotalCostVal)
					{
						BA_MinTotalCostIter = iter;
						BA_MinTotalCostVal = iter->totalCost();
					}
				}
				BA_OptCorridorID = BA_MinTotalCostIter->firstCorridor();
				BA_OptAction = BA_MinTotalCostIter->firstActionTyp();
				BA_OptTargetObjID = BA_MinTotalCostIter->firstTargetObjID();
				BA_OptPassageID = BA_MinTotalCostIter->firstPassageID();
				BA_OptimalSpdLmtFlg = BA_MinTotalCostIter->getSpdLmtFlg();
				BA_OptimalSpdLmtDst = BA_MinTotalCostIter->getSpdLmtDst();
				BA_OptimalSpdLmtVal = BA_MinTotalCostIter->getSpdLmtVal();
				BA_OptimalSpdLmtTyp = BA_MinTotalCostIter->getSpdLmtTyp();
				BA_OptimalSpdLmtID = BA_MinTotalCostIter->getSpdLmtID();
				// DLOG(INFO) << "Optimal passage id x: " << int(this->optPassageID.m_X) << std::endl;
				// DLOG(INFO) << "Optimal passage id y: " << int(this->optPassageID.m_Y) << std::endl;

				uint8_t lane_num = this->input_BA.RM_RoadModel.lane_list.size();
				for (uint8_t i = 0; i < lane_num; i++)
				{
					Lane tmpLaneObj = this->input_BA.RM_RoadModel.lane_list[i];
					uint8_t corridor_num = tmpLaneObj.corridor_list.size();
					for (uint8_t j = 0; j < corridor_num; j++)
					{
						if (BA_OptCorridorID == tmpLaneObj.corridor_list[j].corridor_id)
						{
							BA_TrgtLaneStID = tmpLaneObj.start_id;
							BA_TrgtLaneEndID = tmpLaneObj.end_id;
							break;
						}
					}
				}
			}
			else
			{
				BA_OptCorridorID = 0;
				BA_OptAction = BA::BA_ActionTyp_None;
				BA_OptTargetObjID = 0;
				BA_TrgtLaneStID = 0;
				BA_TrgtLaneEndID = 0;
				BA_OptPassageID = BA::Index{0, 0};
			}
		}

		bool BehaviorAnalysis::UpdateBehaviorNodeGlobalVector()
		{
			if (!this->BA_BehaviorNodeGlobalVector->empty())
			{
				for (vector<BA::behaviorNodeGlobal>::iterator iter = this->BA_BehaviorNodeGlobalVector->begin(); iter != this->BA_BehaviorNodeGlobalVector->end();)
				{
					if (iter->nodeID().m_X != this->BA_StepNum)
					{
						this->BA_BehaviorNodeGlobalVector->erase(iter);
					}
					else
					{
						// only add up here when no erase operation is needed
						iter++;
					}
					if (iter == this->BA_BehaviorNodeGlobalVector->end())
						break;
				}
				return true;
			}
			else
			{
				DLOG(WARNING) << "BA_BehaviorNodeGlobalVector is empty." << std::endl;
				return false;
			}
		}

		bool BehaviorAnalysis::GetRoadModelIDToPD(std::vector<dpc::Index> &roadModelID2PD)
		{
			if (!this->BA_RoadModelVector->empty())
			{
				for (vector<RoadInfo>::iterator iter = this->BA_RoadModelVector->begin(); iter != this->BA_RoadModelVector->end(); iter++)
				{
					if (iter->road_id.x == this->BA_StepNum)
					{
						roadModelID2PD.emplace_back(iter->road_id);
					}
				}
				return true;
			}
			else
			{
				DLOG(WARNING) << "BA_RoadModelVector is empty." << std::endl;
				return false;
			}
		}

		bool BehaviorAnalysis::FindTgtObj(const RoadInfo &roadModel,
										  const uint8_t &objID,
										  Object &tgtObj)
		{
			for (int i = 0; i < roadModel.object_list.size(); i++)
			{
				if (objID == roadModel.object_list[i].object_id)
				{
					tgtObj = roadModel.object_list[i];
					return true;
				}
			}
			return false;
		}

		bool BehaviorAnalysis::WrapUpMsgOutput(const DM::LongitudinalCorridor &BA_LgtBlueCrrdr,
											   const DM::LongitudinalCorridor &BA_LgtYllwBrdr,
											   const DM::LongitudinalCorridor &BA_LgtRedBrdr,
											   const DM::SpeedLimit &BA_LgtSpdLmt,
											   const DM::LateralBlueCorridor &latBlueCrrdr,
											   const DM::LateralBorder &latYllwBrdr,
											   const DM::LateralBorder &latRedBrdr,
											   const uint8_t &spdLmtActVal,
											   const float &spdLmtNxtDst,
											   const BA::BA_ActionTyp &BA_Action4TRP,
                                               const uint32_t &BA_Corridor4TRP,
											   GlobalInfoProcess &globalInfoProc)
		{
			if (!WrapUpMsgToSM(globalInfoProc))
			{
				DLOG(WARNING) << "SM Msg generation failed!" << std::endl;
				return false;
			}
			if (!WrapUpMsgToTRPLat(latBlueCrrdr, latYllwBrdr, latRedBrdr, BA_Action4TRP, BA_Corridor4TRP))
			{
				DLOG(WARNING) << "TRPLat Msg generation failed!" << std::endl;
				return false;
			}
			if (!WrapUpMsgToTRPLongt(BA_LgtBlueCrrdr, BA_LgtYllwBrdr, BA_LgtRedBrdr, BA_LgtSpdLmt, latBlueCrrdr, spdLmtActVal, spdLmtNxtDst, BA_Action4TRP, BA_Corridor4TRP))
			{
				DLOG(WARNING) << "TRPLongt Msg generation failed!" << std::endl;
				return false;
			}
			#ifdef BA_Debug
			if (!WrapUpMsgToSelf())
			{
				DLOG(WARNING) << "BA SELF Msg generation failed!" << std::endl;
				return false;
			}
			#endif
			return true;
		}

		bool BehaviorAnalysis::WrapUpMsgToSM(GlobalInfoProcess &globalInfoProc)
		{
			this->msg_BA2SM.ba_elk_tgt_id = uint8_t(0); // 20220616: it is reused as ILC target id
			this->msg_BA2SM.ba_ldp_elk_md = BA::LatEmrgFnMd_None; // 20220812: it is resused as cycle counter to SM
			this->msg_BA2SM.ba_elk_tgt_dngr_lvl = uint8_t(0);
			this->msg_BA2SM.ba_lc_end_pnt_x = 0;
			this->msg_BA2SM.ba_lc_end_pnt_y = 0;
			this->msg_BA2SM.ba_alc_abl_tgt_id = uint32_t(0);
			this->msg_BA2SM.ba_alc_abl_tgt_dngr_lvl = BA::TgtDngrLvl_None;
			this->msg_BA2SM.ba_dyn_allw_max_spd = 120;
			this->msg_BA2SM.ba_opt_corridor_id = this->BA_OptCorridorID_Deb;
			this->msg_BA2SM.ba_opt_target_obj_id = this->BA_OptTargetObjID_Deb;
			this->msg_BA2SM.ba_opt_action = this->BA_OptAction_Deb;
			// 20220929:
			// if optimal action is LCP
			// then put near object into target obj id
			if (this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LeftLaneChangePrepare || this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_RightLaneChangePrepare)
			{
				// find the passage with the same action with optimal action
				uint8_t trgtActTyp = this->msg_BA2SM.ba_opt_action;
				auto it_passage = std::find_if(this->BA_FirstPassageList->begin(), this->BA_FirstPassageList->end(),
										   		[trgtActTyp](const Passage &passage_)
										   		{
											  		return (passage_.action_type == trgtActTyp);
										   		});
				if (it_passage != this->BA_FirstPassageList->end())
				{
					if (it_passage->interest_obj_id_list.size() > 2)
					{
						this->msg_BA2SM.ba_opt_target_obj_id = it_passage->interest_obj_id_list.at(2);
					}			
				}
			}
			this->msg_BA2SM.ba_lane_change_abolition_avlbl = this->BA_LaneChangeAbolitionAvlbl;
			DLOG(INFO) << "~~BA_DircnIndReq = " << int(this->BA_DircnIndReq) << std::endl;
			DLOG(INFO) << "~~BA_DircnIndReq_Deb = " << int(this->BA_DircnIndReq_Deb) << std::endl;
			this->msg_BA2SM.ba_drcn_ind_req = this->BA_DircnIndReq_Deb; // 20220728: replace raw value with filtered value
			// 20220516: ILC is also included in LaneFollow action
			if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow)
			{
				BA::ILCDirection BA_DrctOfILC = BA::ILCDirection_Unknown;
				if (GetILCDrct(BA_DrctOfILC) == true)
				{
					DLOG(INFO) << "BA_DrctOfILC = " << BA_DrctOfILC << std::endl;
					// 20220616: add ILC target assignment if ILC corridor comes from vehicle target
					uint32_t ilcTrgtID = uint32_t(0);
					uint8_t optLneID = this->BA_OptLaneID_Deb;
					auto it_OptLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
												 [optLneID](const Lane &lane_)
												 {
													return (lane_.start_id == optLneID); 
												 });
					if (it_OptLn != this->input_BA.RM_RoadModel.lane_list.end())
					{
						// search corridor from optLane
						uint32_t optCrrdID = this->BA_OptCorridorID_Deb;
						auto it_OptCrrd = std::find_if(it_OptLn->corridor_list.begin(), it_OptLn->corridor_list.end(),
						                               [optCrrdID](const Corridor &corridor_)
													   {
														   return (corridor_.corridor_id == optCrrdID);
													   });
						if (it_OptCrrd != it_OptLn->corridor_list.end())
						{
							ilcTrgtID = it_OptCrrd->ilc_object_target;
						}
					}
					// this->msg_BA2SM.ba_elk_tgt_id = ilcTrgtID;
					ShineHighlght(this->BA_ShineHighlightFlg, ilcTrgtID); //20220620: determine whether to shine the highlight when ILC is activated
					if (BA_DrctOfILC == BA::ILCDirection_Right)
					{
						this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_Intrf_ILC_Right;	
					}
					else
					{
						this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_Intrf_ILC_Left;
					}
				}
				else
				{
					this->BA_ShineHighlightFlg = uint8_t(0);
				}
			}

			// keep send lane change abolition until vehicle has been in the lane center
			// whether to set duration threshold
			// as original lane change abolition will be sent no more than one frame, lane change abolition has to be got by special logic
			// tag lane change abolition type:0-noAbolition 1-driverAbolition 2-notDriverAbolition, applied to determine whether to consider display danger object
			// 20220528: change lane change abolition condition from corridor id to lane id 
			uint8_t lnChgAblTyp = uint8_t(0);
			DLOG(INFO) << "BA_OptAction_LastCycle = " << int(this->input_BA.BA_OptAction_LastCycle) << std::endl;
			DLOG(INFO) << "BA_OptLaneID = " << int(this->BA_OptLaneID_Deb) << std::endl;
			if (((this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && this->BA_LaneChangeAbolitionAvlbl == true) \
			 	  || (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition)) &&
			 	lnChgAblTyp == uint8_t(0))
			{
				// DLOG(INFO) << "FLAG00" << std::endl;
				if (this->input_BA.BA_OptLaneID_LastCycle != this->BA_OptLaneID_Deb)
				{
					// DLOG(INFO) << "FLAG01" << std::endl;
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LeftLaneChangeAbolition;
					if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChangeAbolition_C)
					{
						lnChgAblTyp = uint8_t(1);
					}
					else
					{
						lnChgAblTyp = uint8_t(2);
					}
				}
			}
			/* else
			{
				lnChgAblTyp = uint8_t(0);
			} */

			if (((this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && this->BA_LaneChangeAbolitionAvlbl == true) \
			      || (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition)) &&
			    lnChgAblTyp == uint8_t(0))
			{
				if (this->input_BA.BA_OptLaneID_LastCycle != this->BA_OptLaneID_Deb)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_RightLaneChangeAbolition;
					if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChangeAbolition_C)
					{
						lnChgAblTyp = uint8_t(1);
					}
					else
					{
						lnChgAblTyp = uint8_t(2);
					}
				}
			}
			/* else
			{
				lnChgAblTyp = uint8_t(0);
			} */
			// DLOG(INFO) << "lnChgAblTyp = " << int(lnChgAblTyp) << std::endl;

			// 20220817: if lane change duration time is too small, display lane change abolition
			bool lftLCDurTmSmallFlg = false;
			bool lftLCDurTmSmallFlg01 = false;
			bool rghtLCDurTmSmallFlg = false;
			bool rghtLCDurTmSmallFlg01 = false;
			if (this->BA_BehaviorStackVector.size() >= 2 && ADC::common::k_BA_LCDurTmInFlg == true)
			{
				// left side
				if (this->BA_BehaviorStackVector.front().Action == BA::BA_ActionTyp_LeftLaneChange && this->BA_BehaviorStackVector.front().Duration <= 0.8f)
				{
					lftLCDurTmSmallFlg = true;
				}
				if (this->BA_BehaviorStackVector.front().Action == BA::BA_ActionTyp_LeftLaneChangeAbolition && 
					this->BA_BehaviorStackVector.at(1).Action == BA::BA_ActionTyp_LeftLaneChange &&
					this->BA_BehaviorStackVector.at(1).Duration <= 0.8f &&
					(this->BA_BehaviorStackVector.at(1).Duration + this->BA_BehaviorStackVector.at(1).PassTime <= 0.8f))
				{
					lftLCDurTmSmallFlg01 = true;
				}
				// right side
				if (this->BA_BehaviorStackVector.front().Action == BA::BA_ActionTyp_RightLaneChange && this->BA_BehaviorStackVector.front().Duration <= 0.8f)
				{
					rghtLCDurTmSmallFlg = true;
				}
				if (this->BA_BehaviorStackVector.front().Action == BA::BA_ActionTyp_RightLaneChangeAbolition && 
					this->BA_BehaviorStackVector.at(1).Action == BA::BA_ActionTyp_RightLaneChange &&
					this->BA_BehaviorStackVector.at(1).Duration <= 0.8f &&
					(this->BA_BehaviorStackVector.at(1).Duration + this->BA_BehaviorStackVector.at(1).PassTime <= 0.8f))
				{
					rghtLCDurTmSmallFlg01 = true;
				}
			}
			// DLOG(INFO) << "lftLCDurTmSmallFlg = " << int(lftLCDurTmSmallFlg) << std::endl;
			// DLOG(INFO) << "lftLCDurTmSmallFlg01 = " << int(lftLCDurTmSmallFlg01) << std::endl;
			if ((this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && lftLCDurTmSmallFlg == true) ||
			     lftLCDurTmSmallFlg01 == true)
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LeftLaneChangeAbolition;
			}
			if ((this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && rghtLCDurTmSmallFlg == true) ||
			     rghtLCDurTmSmallFlg01 == true)
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_RightLaneChangeAbolition;
			}

			static float lnChgAblNotCmdDur = 0.0f;
			static bool lnChgAblNotCmdFlg = false;
			static uint8_t origTrgtLnID = uint8_t(0);
			static uint32_t origAblCrrdID = uint32_t(0);
			static uint8_t origAblLnID = uint8_t(0);
			static BA::BA_ActionTyp origAction = BA::BA_ActionTyp_None;
			static uint32_t origObjID = uint32_t(0);
			// lock the original corridor id
			if (lnChgAblTyp != uint8_t(0))
			{
				origAblCrrdID = this->BA_OptCorridorID_Deb;
				origAblLnID = this->BA_OptLaneID_Deb;
			}
			if (lnChgAblTyp == uint8_t(2))
			{
				lnChgAblNotCmdFlg = true;
				origTrgtLnID = this->input_BA.BA_OptLaneID_LastCycle;
				if (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange)
				{
					origAction = BA::BA_ActionTyp_LeftLaneChange;
				}
				else if (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange)
				{
					origAction = BA::BA_ActionTyp_RightLaneChange;
				}
			}
			if (lnChgAblNotCmdFlg == true)
			{
				lnChgAblNotCmdDur += BEHAVIORANALYZERCYCLE / 1000.0f;
			}
			// add more conditions to make sure lane change abolition target tagged
			if (lnChgAblNotCmdDur >= ADC::common::k_BA_MaxLnChgAblNotDrvCmdDur || (this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LaneFollow && this->BA_LastActTyp2SM == BA::BA_ActionTyp_LaneFollow))
			{
				lnChgAblNotCmdFlg = false;
				lnChgAblNotCmdDur = 0.0f;
				origTrgtLnID = uint8_t(0);
				origAblCrrdID = uint32_t(0);
				origAblLnID = uint8_t(0);
				origAction = BA::BA_ActionTyp_None;
				origObjID = uint32_t(0);
			}
			// if lnChgAblNotCmdFlg is true, then try to find such a target which is located on the target corridor and is closest to ego vehicle
			// the absolute dist should be less than TH * v_ego
			// 20220528: change target corridor to target lane
			bool ablObjLckedFlg = false;
			DLOG(INFO) << "lnChgAblNotCmdFlg = " << int(lnChgAblNotCmdFlg) << std::endl;
			DLOG(INFO) << "lnChgAblNotCmdDur = " << lnChgAblNotCmdDur << std::endl;
			BA::BA_ActionTyp trgtOrigAction = origAction;
			if (lnChgAblNotCmdFlg == true)
			{
				// 20220927
				// firstly, try to find it in interest object list (if the 4th element of LC passageOb is not 0, then assign the id to ba_alc_abl_tgt_id)
				// secondly, try to find it in target objects array of behaviorNodeGlobal(if the 3rd element of m_target_obj_id_arr is not 0, then assign the id to ba_alc_abl_tgt_id)
				if (!this->BA_PassageObLst.empty())
				{
					auto it_pssgObs = std::find_if(this->BA_PassageObLst.begin(), this->BA_PassageObLst.end(),
												   [trgtOrigAction](const DM::Passageob &pssgob_)
												   {
													   return (trgtOrigAction == pssgob_.action_type);
												   });
					if (it_pssgObs != this->BA_PassageObLst.end())
					{
						if (it_pssgObs->interest_obj_id_list.size() == 5 && it_pssgObs->lc_passagegenflg == false)
						{
							if (it_pssgObs->interest_obj_id_list.back() != uint32_t(0))
							{
								// got it
								origObjID = it_pssgObs->interest_obj_id_list.back();
								DLOG(INFO) << "BA original object id = " << int(origObjID) << std::endl;
							}
						}
					}
				}

				if (origObjID == uint32_t(0))
				{
					if (!this->BA_TerminatedBehaviorNodesVector->empty())
					{
						auto it_bhvrNode = std::find_if(this->BA_TerminatedBehaviorNodesVector->begin(), this->BA_TerminatedBehaviorNodesVector->end(),
														[trgtOrigAction](const BA::behaviorNodeGlobal &bhvrNdGlobal_)
														{
															if (bhvrNdGlobal_.m_ActionTypArr.size() == 2)
															{
																return (trgtOrigAction == bhvrNdGlobal_.m_ActionTypArr.at(1));
															}
															else
															{
																return false;
															}
														});
						if (it_bhvrNode != this->BA_TerminatedBehaviorNodesVector->end())
						{
							if (it_bhvrNode->m_TargetObjIDArr.size() == 3)
							{
								if (it_bhvrNode->m_TargetObjIDArr.back() != uint32_t(0))
								{
									// got it
									origObjID = it_bhvrNode->m_TargetObjIDArr.back();
									DLOG(INFO) << "PD original object id = " << int(origObjID) << std::endl;
								}
							}
						}
					}
				}

				if (origObjID != uint32_t(0))
				{
					ablObjLckedFlg = true;
					this->msg_BA2SM.ba_alc_abl_tgt_id = origObjID;
				}
				// original version is below
				// 20220927: if abolition object id is not locked, then original strategy is applied to get abolition object id
				if (ablObjLckedFlg == false)
				{
					/* uint8_t origTrgtLnID_ = origTrgtLnID;
					auto it_Ln = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
											 [origTrgtLnID_](const Lane &lane_)
											 {
												 return (lane_.start_id == origTrgtLnID_);
											 });
					if (it_Ln != this->input_BA.RM_RoadModel.lane_list.end())
					{
						uint8_t trgtFeatureTag = BA::FeatureTag_LineFllw;
						auto it_crrd = std::find_if(it_Ln->corridor_list.begin(), it_Ln->corridor_list.end(),
									               [trgtFeatureTag](const Corridor &corridor_)
												   {
														return (corridor_.feature_tag == trgtFeatureTag);
												   });
						if (it_crrd != it_Ln->corridor_list.end())
						{
							// last target corridor found
							float minAbsRelDst = 200.0f;
							uint32_t minAbsRelDstObjID = uint32_t(0);
							float absRelDst = minAbsRelDst;
							bool ablObjUpdt = false;
							for (auto objLoc : it_crrd->object_in_corridor_list)
							{
								absRelDst = std::abs(objLoc.relative_center_x);
								if (absRelDst <= ADC::common::k_BA_AblObjDispTH * this->input_BA.VSE_LongitudinalVelocity)
								{
									if (absRelDst < minAbsRelDst)
									{
										minAbsRelDst = absRelDst;
										minAbsRelDstObjID = objLoc.object_id;
										ablObjUpdt = true;
									}
								}
							}
							if (ablObjUpdt == true)
							{
								this->msg_BA2SM.ba_alc_abl_tgt_id = minAbsRelDstObjID;
							}
						}
					} */
				}
			}
			else
			{
				this->msg_BA2SM.ba_alc_abl_tgt_id = uint32_t(0);
			}

			static float lnChgAblLftCnt = 0.0f;
			static float lnChgAblRghtCnt = 0.0f;
			// 20220527: add min waiting time before determining whether to switch lastActTyp2SM 
			if (this->BA_LastActTyp2SM == BA::BA_ActionTyp_LeftLaneChangeAbolition || this->BA_LastActTyp2SM == BA::BA_ActionTyp_RightLaneChangeAbolition)
			{
				if ((lnChgAblLftCnt >= ADC::common::k_BA_MinWaitTmLCAbl2LF || lnChgAblRghtCnt >= ADC::common::k_BA_MinWaitTmLCAbl2LF) && this->BA_VehInLnCenter == true)
				{
					this->BA_LastActTyp2SM = BA::BA_ActionTyp_LaneFollow;
				}
				else
				{
					this->msg_BA2SM.ba_opt_action = this->BA_LastActTyp2SM;
				}
			}
			// limit max lane change abolition duration time for SM
			// 20220507: if lane change abolition duration is less than threshold and optimal corridor id is switched, then set ba_opt_action to LaneFollow
			if (this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LeftLaneChangeAbolition)
			{
				lnChgAblLftCnt += BEHAVIORANALYZERCYCLE / 1000.0f;
			}
			else
			{
				lnChgAblLftCnt = 0.0f;
			}
			if (lnChgAblLftCnt >= ADC::common::k_BA_MaxLnChgAblDurTm)
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
			}
			if (this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_RightLaneChangeAbolition)
			{
				lnChgAblRghtCnt += BEHAVIORANALYZERCYCLE / 1000.0f;
			}
			else
			{
				lnChgAblRghtCnt = 0.0f;
			}
			if (lnChgAblRghtCnt >= ADC::common::k_BA_MaxLnChgAblDurTm)
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
			}
			
			DLOG(INFO) << "origAblCrrdID = " << int(origAblCrrdID) << std::endl;
			DLOG(INFO) << "origAblLnID = " << int(origAblLnID) << std::endl;
			// 20220507: if during lane change abolition happens, the opt corridor id is different from the locked original corridor id
			// then cancel lane change abolition display
			if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition || this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow) \
			    && (origAblLnID != this->BA_OptLaneID_Deb && origAblLnID != uint8_t(0) /* && (origAblCrrdID != this->BA_OptCorridorID_Deb && origAblCrrdID != uint32_t(0) */) \
				/* && (lnChgAblLftCnt > 0.2f || lnChgAblRghtCnt > 0.2f) */)
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
				this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
			}
			// 20220717: if ego lane end range is less than 15m, 
			// then suppress lane change abolition display
			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			auto it_egoLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
										 [egoLnID](const Lane &lane_)
										 {
											 return (egoLnID == lane_.start_id);
										 });
			if (it_egoLn != this->input_BA.RM_RoadModel.lane_list.end())
			{
				uint8_t tmpNgbrLnID = uint8_t(0);
				if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition) &&
					it_egoLn->end_range <= ADC::common::k_BA_MinDstLnEndLCAblCncl)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
					this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
				}
				// 20220811: if lane change is available and neighbor lane is empty,
				// then suppress lane change abolition display
				else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition && this->BA_LaneChangeAbolitionAvlbl == true && lftLCDurTmSmallFlg == false && lftLCDurTmSmallFlg01 == false)
				{
					if (!it_egoLn->left_neighbor_lane.empty())
					{
						for (auto ngbrLn : it_egoLn->left_neighbor_lane)
						{
							if (ngbrLn.start_range < ADC::common::k_BA_MinNgbrLnStRng)
							{
								tmpNgbrLnID = ngbrLn.lane_id;
								break;
							}
						}
						if (tmpNgbrLnID == uint8_t(0))
						{
							this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
							this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
						}
					}
					else
					{
						this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
						this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
					}
				}
				else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition && this->BA_LaneChangeAbolitionAvlbl == true && rghtLCDurTmSmallFlg == false && rghtLCDurTmSmallFlg01 == false)
				{
					if (!it_egoLn->right_neighbor_lane.empty())
					{
						for (auto ngbrLn : it_egoLn->right_neighbor_lane)
						{
							if (ngbrLn.start_range < ADC::common::k_BA_MinNgbrLnStRng)
							{
								tmpNgbrLnID = ngbrLn.lane_id;
								break;
							}
						}
						if (tmpNgbrLnID == uint8_t(0))
						{
							this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
							this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
						}
					}
					else
					{
						this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
						this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
					}
				}
			}
			
			this->BA_LastActTyp2SM = BA::BA_ActionTyp(this->msg_BA2SM.ba_opt_action);

			// 20221103:
			// if current action type is lane change and lane change abolition is not available
			// then try to get interested target id from pd 
			if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange) &&
			     this->BA_LaneChangeAbolitionAvlbl == false)
			{
				BA::BA_ActionTyp trgtAction = this->BA_OptAction_Deb; 
				if (!this->BA_TerminatedBehaviorNodesVector->empty())
				{
					auto it_bhvrNode = std::find_if(this->BA_TerminatedBehaviorNodesVector->begin(), this->BA_TerminatedBehaviorNodesVector->end(),
													[trgtAction](const BA::behaviorNodeGlobal &bhvrNdGlobal_)
													{
														if (bhvrNdGlobal_.m_ActionTypArr.size() == 2)
														{
															return (trgtAction == bhvrNdGlobal_.m_ActionTypArr.at(1));
														}
														else
														{
															return false;
														}
													});
					if (it_bhvrNode != this->BA_TerminatedBehaviorNodesVector->end())
					{
						if (it_bhvrNode->m_TargetObjIDArr.size() == 3)
						{
							if (it_bhvrNode->m_TargetObjIDArr.back() != uint32_t(0))
							{
								// got it
								this->msg_BA2SM.ba_elk_tgt_id = it_bhvrNode->m_TargetObjIDArr.back();
								DLOG(INFO) << "lcNotAvlbl PD original object id = " << int(this->msg_BA2SM.ba_elk_tgt_id) << std::endl;
							}
						}
					}
				}
			}

			// 20220901:
			// if optimal action is left/right lane change
			// then action sent to SM will be set to left/right lane change remind
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto)
			{
				if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LeftLaneChangeRemind;
				}
				else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_RightLaneChangeRemind;
				}
				else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangePrepare)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
				}
				else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangePrepare)
				{
					this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
				}
				// 20220910
				// if distance to split type waypoint is less than 1.5km
				// and lane change number is greater than 0
				// and target lane is drivable
				// then set ba_opt_action to lane change remind
				if (ADC::common::k_BA_AlwysLCR4SpltTypWP == true &&
					globalInfoProc.globalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_DistThr4LCR &&
					globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum > 0 &&
					( globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd ||
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd) )
				{
					uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
					uint8_t lftNbghLnID = uint8_t(0);
					uint8_t rghtNbghLnID = uint8_t(0);
					auto it_Lane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										        [egoLnID](const WPExtendLaneStruct &wplane_)
												{
													return (egoLnID == wplane_.lane_id);
												});
					if (it_Lane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
					{
						if (!it_Lane->left_extend_lane.empty())
						{
							for (auto lftExtLane : it_Lane->left_extend_lane)
							{
								if (lftExtLane.adjacent_extendlane_offset < 1.0f && lftExtLane.adjacent_extendlane_endoffset >= 1.0f)
								{
									lftNbghLnID = lftExtLane.adjacent_extendlane_id;
									break;
								}
							}
						}
						if (!it_Lane->right_extend_lane.empty())
						{
							for (auto rghtExtLane : it_Lane->right_extend_lane)
							{
								if (rghtExtLane.adjacent_extendlane_offset < 1.0f && rghtExtLane.adjacent_extendlane_endoffset >= 1.0f)
								{
									rghtNbghLnID = rghtExtLane.adjacent_extendlane_id;
									break;
								}
							}
						}
					}
					if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
					{
						if (rghtNbghLnID != uint8_t(0))
						{
							auto it_rghtLane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										        		   [rghtNbghLnID](const WPExtendLaneStruct &wplane_)
														   {
															   return (rghtNbghLnID == wplane_.lane_id);
														   });
							if (it_rghtLane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
							{
								if (it_rghtLane->is_emergency_lane == uint8_t(0))
								{
									this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_RightLaneChangeRemind;
								}
							}
						}
					}
					else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
					{
						if (lftNbghLnID != uint8_t(0))
						{
							auto it_lftLane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										        		   [lftNbghLnID](const WPExtendLaneStruct &wplane_)
														   {
															   return (lftNbghLnID == wplane_.lane_id);
														   });
							if (it_lftLane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
							{
								if (it_lftLane->is_emergency_lane == uint8_t(0))
								{
									this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LeftLaneChangeRemind;
								}
							}
						}
					}
				}
			}
			// 20221019:
			// if opt action is contradictory to sm lcavlblbhvr
			// then set ba_opt_action to lf
			if ((this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LeftLaneChange && (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_RightLaneChange_C)) ||
				(this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_RightLaneChange && (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_Auto || this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LeftLaneChange_C)) ||
				((this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_RightLaneChange || this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LeftLaneChange) && this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_None))
			{
				this->msg_BA2SM.ba_opt_action = BA::BA_ActionTyp_LaneFollow;
				this->BA_OptimalAction = BA::BA_ActionTyp_LaneFollow;
				this->BA_OptAction_Deb = BA::BA_ActionTyp_LaneFollow;
			}

			// prepare object 01 to 06
			if (k_BA_PrepareObjToSMFlg == true)
			{
				uint32_t dfltObjID = 0;
				float dfltObjDst = 1000;
				float dfltObjVel = 0;
				Corridor dfltObjBlgCrrd;
				bool isLineFllwCrrdFound = false;
				bool flg = GetTrgtObjInfo(dfltObjID, dfltObjDst, dfltObjVel, dfltObjBlgCrrd, isLineFllwCrrdFound);

				std::vector<ObjectTrim> objListTrim(6);
				// if (GetFrtObjToSM(objListTrim, flg, isLineFllwCrrdFound, dfltObjID) == true)
				// {
				// 	for (int j = 0; j < this->msg_BA2SM.ba_obj_list2sm.size(); j++)
				// 	{
				// 		this->msg_BA2SM.ba_obj_list2sm.at(j) = objListTrim.at(j);
				// 	}
				// }
			}

			return true;
		}

		bool BehaviorAnalysis::WrapUpMsgToTRPLat(const DM::LateralBlueCorridor &latBlueCrrdr,
												 const DM::LateralBorder &latYllwBrdr,
												 const DM::LateralBorder &latRedBrdr,
												 const BA::BA_ActionTyp &BA_Action4TRPLat,
                                                 const uint32_t &BA_Corridor4TRPLat)
		{
			if (this->input_BA.SM_LFComfortAvlblBehavior == BA::LFCmfrtAvlblBhvr_LatStandAlone ||
				this->input_BA.SM_LFComfortAvlblBehavior == BA::LFCmfrtAvlblBhvr_LongtLat_HO ||
				this->input_BA.SM_LFComfortAvlblBehavior == BA::LFCmfrtAvlblBhvr_LongtLat_HOInclILC ||
				this->input_BA.SM_LFComfortAvlblBehavior == BA::LFCmfrtAvlblBhvr_LongtLat_HF)
			{
				this->msg_BA2TRPLat.ba_trp_lat_sts = this->BA_TRPLatSts_Deb;		// assigned for TJA
				this->msg_BA2TRPLat.ba_lat_actv_func_sts = this->input_BA.SM_LatActvFunc; // 20220627
				this->msg_BA2TRPLat.ba_ref_line_c0 = latBlueCrrdr.m_reference_line.c0;
				this->msg_BA2TRPLat.ba_ref_line_c1 = latBlueCrrdr.m_reference_line.c1;
				this->msg_BA2TRPLat.ba_ref_line_c2 = latBlueCrrdr.m_reference_line.c2;
				this->msg_BA2TRPLat.ba_ref_line_c3 = latBlueCrrdr.m_reference_line.c3;
				this->msg_BA2TRPLat.ba_ref_line_rng_sts = latBlueCrrdr.m_reference_line.start_range;
				this->msg_BA2TRPLat.ba_ref_line_rng_ed = latBlueCrrdr.m_reference_line.end_range;
				this->msg_BA2TRPLat.ba_ego_lane_width = latBlueCrrdr.m_width;
				this->msg_BA2TRPLat.ba_opt_corridor_id = this->BA_OptCorridorID_Deb;
				// cartesian coordination
				// for (int i = 0; i < k_BA_SamplePointNumber; i++)
				// {
				// 	this->msg_BA2TRPLat.ba_lft_ylw_brdr_dx[i] = latYllwBrdr.m_left_border[i].s;
				// 	this->msg_BA2TRPLat.ba_lft_ylw_brdr_dy[i] = latYllwBrdr.m_left_border[i].l;
				// 	this->msg_BA2TRPLat.ba_lft_ylw_brdr_prob[i] = 1;
				// 	this->msg_BA2TRPLat.ba_rght_ylw_brdr_dx[i] = latYllwBrdr.m_right_border[i].s;
				// 	this->msg_BA2TRPLat.ba_rght_ylw_brdr_dy[i] = latYllwBrdr.m_right_border[i].l;
				// 	this->msg_BA2TRPLat.ba_rght_ylw_brdr_prob[i] = 1;

				// 	this->msg_BA2TRPLat.ba_lft_red_brdr_dx[i] = latRedBrdr.m_left_border[i].s;
				// 	this->msg_BA2TRPLat.ba_lft_red_brdr_dy[i] = latRedBrdr.m_left_border[i].l;
				// 	this->msg_BA2TRPLat.ba_lft_red_brdr_prob[i] = 1;
				// 	this->msg_BA2TRPLat.ba_rght_red_brdr_dx[i] = latRedBrdr.m_right_border[i].s;
				// 	this->msg_BA2TRPLat.ba_rght_red_brdr_dy[i] = latRedBrdr.m_right_border[i].l;
				// 	this->msg_BA2TRPLat.ba_rght_red_brdr_prob[i] = 1;
				// }
				this->msg_BA2TRPLat.ba_drive_md = this->input_BA.SM_DriveMode; // 20220731: actual drive mode
				this->msg_BA2TRPLat.ba_lat_fd_in_md = BA::FdInMd_None;
				this->msg_BA2TRPLat.ba_lat_fd_out_md = BA::FdOutMd_None;
				this->msg_BA2TRPLat.ba_hd_feel_md = BA::HdFeelMd_None;
				this->msg_BA2TRPLat.ba_opt_action = this->BA_OptAction_Deb;

				// 20220901: remap action and corridor id to TRPLat when lane change comfirmation
				if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto)
				{
					this->msg_BA2TRPLat.ba_opt_action = BA_Action4TRPLat;
					this->msg_BA2TRPLat.ba_opt_corridor_id = BA_Corridor4TRPLat;
				}

				// 20220718: increase lane change abolition duration time for TRPlat
				static BA::BA_ActionTyp actionTyp = BA::BA_ActionTyp_None;
				if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition)
				{
					this->BA_LaneChgAblFlg4TRPLat = true;
					actionTyp = this->BA_OptAction_Deb;
				}
				if (this->BA_LaneChgAblFlg4TRPLat == true)
				{
					if (std::abs(this->msg_BA2TRPLat.ba_ref_line_c0) <= 0.2f)
					{
						this->BA_LaneChgAblFlg4TRPLat = false;
						actionTyp = BA::BA_ActionTyp_None;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange)
					{
						this->BA_LaneChgAblFlg4TRPLat = false;
						actionTyp = BA::BA_ActionTyp_None;
					}
					else if (this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LaneFollow)
					{
						this->BA_LaneChgAblFlg4TRPLat = false;
						actionTyp = BA::BA_ActionTyp_None;
					}
				}
				if (this->BA_LaneChgAblFlg4TRPLat == true)
				{
					this->msg_BA2TRPLat.ba_opt_action = actionTyp;
				}
			}
			else
			{
				this->msg_BA2TRPLat.ba_trp_lat_sts = BA::TRPLatSts_None;			 // lateral comfort is not available, such as LongtStandAlone
				this->msg_BA2TRPLat.ba_lat_actv_func_sts = BA::LatActvFuncSts_None;  // lateral comfort is not available
				this->msg_BA2TRPLat.ba_opt_corridor_id = uint32_t(0);
				this->msg_BA2TRPLat.ba_opt_action = BA::BA_ActionTyp_None;
				this->BA_LaneChgAblFlg4TRPLat = false;
			}

			return true;
		}

		bool BehaviorAnalysis::WrapUpMsgToTRPLongt(const DM::LongitudinalCorridor &lgtBlueCrrdr,
												   const DM::LongitudinalCorridor &lgtYllwBrdr,
												   const DM::LongitudinalCorridor &lgtRedBrdr,
												   const DM::SpeedLimit &lgtSpdLmt,
												   const DM::LateralBlueCorridor &latBlueCrrdr,
												   const uint8_t &spdLmtActVal,
												   const float &spdLmtNxtDst,
												   const BA::BA_ActionTyp &BA_Action4TRPLgt,
                                     			   const uint32_t &BA_Corridor4TRPLgt)
		{
			this->msg_BA2TRPLongt.ba_trp_d_lne_view_rngl = latBlueCrrdr.m_reference_line.end_range;
			this->msg_BA2TRPLongt.ba_trp_d_lne_view_rngr = latBlueCrrdr.m_reference_line.end_range;
			this->msg_BA2TRPLongt.ba_trp_indx_c2_lne_l = latBlueCrrdr.m_reference_line.c2;
			this->msg_BA2TRPLongt.ba_trp_indx_c2_lne_r = latBlueCrrdr.m_reference_line.c2;
			this->msg_BA2TRPLongt.ba_trp_indx_c3_lne_l = latBlueCrrdr.m_reference_line.c3;
			this->msg_BA2TRPLongt.ba_trp_indx_c3_lne_r = latBlueCrrdr.m_reference_line.c3;

			// 20220419 changed
			this->msg_BA2TRPLongt.ba_trp_indx_spdlim_typ = uint8_t(this->BA_OptimalSpdLmtFlg);
			this->msg_BA2TRPLongt.ba_trp_kmh_spdlim_act_val = this->BA_OptimalSpdLmtVal;
			this->msg_BA2TRPLongt.ba_trp_d_spdlim_nxt_dst = this->BA_OptimalSpdLmtDst;
			this->msg_BA2TRPLongt.ba_trp_indx_trp_lgt_sts = BA::TRPLongtSts_Comfort;
			this->msg_BA2TRPLongt.ba_trp_d_blue_crrdr_dist = lgtBlueCrrdr.m_corridor_distance;
			this->msg_BA2TRPLongt.ba_trp_v_blue_crrdr_spd = lgtBlueCrrdr.m_corridor_speed;
			this->msg_BA2TRPLongt.ba_trp_v_trgt_spd = lgtSpdLmt.m_target_speed;
			this->msg_BA2TRPLongt.ba_trp_d_related_dist_of_trgt_spd = lgtSpdLmt.m_related_distance;
			this->msg_BA2TRPLongt.ba_trp_d_yllw_brdr_dist = lgtYllwBrdr.m_corridor_distance;
			this->msg_BA2TRPLongt.ba_trp_v_yllw_brdr_spd = lgtYllwBrdr.m_corridor_speed;
			this->msg_BA2TRPLongt.ba_trp_d_red_brdr_dist = lgtRedBrdr.m_corridor_distance;
			this->msg_BA2TRPLongt.ba_trp_v_red_brdr_spd = lgtRedBrdr.m_corridor_speed;
			this->msg_BA2TRPLongt.ba_trp_indx_opt_action = this->BA_OptAction_Deb;

			// 20220901: remap action and corridor id to TRPLgt when lane change comfirmation
			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto)
			{
				this->msg_BA2TRPLat.ba_opt_action = BA_Action4TRPLgt;
			}

			if (this->BA_OptTargetObjID_Deb != 0)
			{
				Object trgtObj;
				if (!FindTgtObj(this->input_BA.RM_RoadModel, this->BA_OptTargetObjID_Deb, trgtObj))
				{
					DLOG(WARNING) << "The target object does not exist." << std::endl;
					return false;
				}
				this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = trgtObj.relative_position_x - ADC::common::k_BA_DistBtwVehFrt2BackAxis; // minus dist btw vehicle front to back axis
				this->msg_BA2TRPLongt.ba_trp_d_tgt_lat_dist_iv = trgtObj.relative_position_y;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_longt_spd_iv = trgtObj.absolute_velocity_x;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_lat_spd_iv = trgtObj.absolute_velocity_y;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_rlt_spd_iv = trgtObj.absolute_velocity_x - this->input_BA.VSE_LongitudinalVelocity;
				this->msg_BA2TRPLongt.ba_trp_is_tgt_prsnt_iv = true;
				this->msg_BA2TRPLongt.ba_trp_t_tgt_dur_tm_iv = 5; // temp assignment
				this->msg_BA2TRPLongt.ba_trp_indx_tgt_id_iv = trgtObj.object_id;
				this->msg_BA2TRPLongt.ba_trp_a_tgt_longt_accel_iv = trgtObj.absolute_acceleration_x;
				this->msg_BA2TRPLongt.ba_trp_a_tgt_lat_accel_iv = trgtObj.absolute_acceleration_y;
				this->msg_BA2TRPLongt.ba_trp_indx_tgt_typ_iv = trgtObj.classification;
				this->msg_BA2TRPLongt.ba_trp_is_tgt_new_iv = false;		  // temp assignment
				this->msg_BA2TRPLongt.ba_trp_is_tgt_lne_chg_flag_iv = false; // temp assignment
				this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = true;
				this->msg_BA2TRPLongt.ba_trp_indx_blue_crrdr_tgt_id = this->BA_OptTargetObjID_Deb;
			}
			else
			{
				this->msg_BA2TRPLongt.ba_trp_d_tgt_longt_dist_iv = 1000.0f;
				this->msg_BA2TRPLongt.ba_trp_d_tgt_lat_dist_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_longt_spd_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_lat_spd_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_v_tgt_rlt_spd_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_is_tgt_prsnt_iv = false;
				this->msg_BA2TRPLongt.ba_trp_t_tgt_dur_tm_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_indx_tgt_id_iv = 0;
				this->msg_BA2TRPLongt.ba_trp_a_tgt_longt_accel_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_a_tgt_lat_accel_iv = 0.0f;
				this->msg_BA2TRPLongt.ba_trp_indx_tgt_typ_iv = 0;
				this->msg_BA2TRPLongt.ba_trp_is_tgt_new_iv = false;
				this->msg_BA2TRPLongt.ba_trp_is_tgt_lne_chg_flag_iv = false;
				this->msg_BA2TRPLongt.ba_trp_is_blue_crrdr_obj_exist = false;
				this->msg_BA2TRPLongt.ba_trp_indx_blue_crrdr_tgt_id = uint32_t(0);
			}

			return true;
		}

		bool BehaviorAnalysis::WrapUpMsgToSelf()
		{
			BA::Index optPassageID;
			optPassageID.m_X = this->BA_OptPassageID_Deb.m_X;
			optPassageID.m_Y = this->BA_OptPassageID_Deb.m_Y;
			auto it_passage = std::find_if(this->BA_FirstPassageList->begin(), this->BA_FirstPassageList->end(),
										   [optPassageID](const Passage &passage_)
										   {
											   return (passage_.passage_id.x == optPassageID.m_X && passage_.passage_id.y == optPassageID.m_Y);
										   });
			this->msg_BA2Obs.m_ba_passage_list_ob.clear();
			if (!this->BA_PassageListObs.empty())
			{
				this->msg_BA2Obs.m_ba_passage_list_ob.resize(this->BA_PassageListObs.size());
				copy(this->BA_PassageListObs.begin(), this->BA_PassageListObs.end(), this->msg_BA2Obs.m_ba_passage_list_ob.begin());
			}

			bool lnChgObsPassageExist = false;
			Passage lnChgObsPassage;
			if (!this->msg_BA2Obs.m_ba_passage_list_ob.empty())
			{
				lnChgObsPassage = this->msg_BA2Obs.m_ba_passage_list_ob.front();
				lnChgObsPassageExist = true;
			}

			if (it_passage != this->BA_FirstPassageList->end())
			{
				if (lnChgObsPassageExist == true)
				{
					this->msg_BA2Obs.m_ba_passage_genflg_lc = lnChgObsPassage.lc_passagegenflg;
					this->msg_BA2Obs.m_ba_interest_obj_id_list.resize(lnChgObsPassage.interest_obj_id_list.size());
					copy(lnChgObsPassage.interest_obj_id_list.begin(), lnChgObsPassage.interest_obj_id_list.end(), this->msg_BA2Obs.m_ba_interest_obj_id_list.begin());					
					this->msg_BA2Obs.m_ba_lc_ego_frt_obj_exist_flg = lnChgObsPassage.lc_ego_frt_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lc_ego_frt_obj_id = lnChgObsPassage.lc_ego_frt_obj_id;					
					this->msg_BA2Obs.m_ba_lc_tar_bck_obj_exist_flg = lnChgObsPassage.lc_tar_bck_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lc_tar_bck_obj_id = lnChgObsPassage.lc_tar_bck_obj_id;
					this->msg_BA2Obs.m_ba_lc_tar_frt_obj_exist_flg = lnChgObsPassage.lc_tar_frt_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lc_tar_frt_obj_id = lnChgObsPassage.lc_tar_frt_obj_id;
					this->msg_BA2Obs.m_ba_lcp_ego_bck_obj_exist_flg = lnChgObsPassage.lcp_ego_bck_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lcp_ego_bck_obj_id = lnChgObsPassage.lcp_ego_bck_obj_id;
					this->msg_BA2Obs.m_ba_lcp_ego_frt_obj_exist_flg = lnChgObsPassage.lcp_ego_frt_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lcp_ego_frt_obj_id = lnChgObsPassage.lcp_ego_frt_obj_id;
					this->msg_BA2Obs.m_ba_lcp_nr_bck_obj_exist_flg = lnChgObsPassage.lcp_nr_bck_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lcp_nr_bck_obj_id = lnChgObsPassage.lcp_nr_bck_obj_id;
					this->msg_BA2Obs.m_ba_lcp_nr_frt_obj_exist_flg = lnChgObsPassage.lcp_nr_frt_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lcp_nr_frt_obj_id = lnChgObsPassage.lcp_nr_frt_obj_id;
					this->msg_BA2Obs.m_ba_lcp_nr_obj_exist_flg = lnChgObsPassage.lcp_nr_obj_exist_flg;
					this->msg_BA2Obs.m_ba_lcp_nr_obj_id = lnChgObsPassage.lcp_nr_obj_id;
				}
			}
			else
			{
				this->msg_BA2Obs.m_ba_passage_genflg_lc = false;
				this->msg_BA2Obs.m_ba_lc_ego_frt_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lc_ego_frt_obj_id = 0;
				this->msg_BA2Obs.m_ba_lc_tar_bck_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lc_tar_bck_obj_id = 0;
				this->msg_BA2Obs.m_ba_lc_tar_frt_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lc_tar_frt_obj_id = 0;
				this->msg_BA2Obs.m_ba_lcp_ego_bck_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lcp_ego_bck_obj_id = 0;
				this->msg_BA2Obs.m_ba_lcp_ego_frt_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lcp_ego_frt_obj_id = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_bck_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_bck_obj_id = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_frt_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_frt_obj_id = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_obj_exist_flg = 0;
				this->msg_BA2Obs.m_ba_lcp_nr_obj_id = 0;
			}
			/* if (!this->BA_PassageListObs.empty())
			{
				this->msg_BA2Obs.m_ba_passage_list_ob.resize(this->BA_PassageListObs.size());
				copy(this->BA_PassageListObs.begin(), this->BA_PassageListObs.end(), this->msg_BA2Obs.m_ba_passage_list_ob.begin());
			} */

			SwitchBhvrNodeGlobalVec(this->msg_BA2Obs.m_ba_behavior_node_global_vector);

			this->msg_BA2Obs.m_ba_behavior_node_update_done = this->BA_BehaviorNodeUpdateDone;

			// SwitchLatAvlblBhvrVec(this->msg_BA2Obs.m_ba_lat_avlbl_behavior); // comment on 20220705

			this->msg_BA2Obs.m_ba_optimal_action = uint8_t(this->BA_OptAction_Deb);
			this->msg_BA2Obs.m_ba_optimal_corridor_id = this->BA_OptCorridorID_Deb;
			this->msg_BA2Obs.m_ba_optimal_target_obj_id = this->BA_OptTargetObjID_Deb;
			this->msg_BA2Obs.m_ba_step_num = this->BA_StepNum;

			SwitchBhvrNodeGlobalVec(this->msg_BA2Obs.m_ba_terminated_behavior_nodes_vector);
			DLOG(INFO) << "tbn vector size = " << this->msg_BA2Obs.m_ba_terminated_behavior_nodes_vector.size() << std::endl;

			if (this->BA_TerminatedBehaviorNodesVector->size() != this->BA_FirstPassageList->size())
			{
				// certain BNG is eliminated by collision res
				for (auto psg : *this->BA_FirstPassageList)
				{
					uint8_t psg_act = psg.action_type;
					bool res = false;
					for (auto tbn : *this->BA_TerminatedBehaviorNodesVector)
					{
						if (tbn.firstActionTyp() == psg_act)
						{
							res = true;
							break;
						}
					}
					if (res == false)
					{
						// elements in BA_FirstPassageList is not found in BA_TerminatedBehaviorNodesVector
						this->msg_BA2Obs.m_ba_lat_avlbl_behavior.emplace_back(psg_act);
					}
				}
			}
			DLOG(INFO) << "BA_TerminatedBehaviorNodesVector size = " << this->BA_TerminatedBehaviorNodesVector->size() << std::endl;
			DLOG(INFO) << "BA_FirstPassageList size = " << this->BA_FirstPassageList->size() << std::endl;
			DLOG(INFO) << "m_ba_lat_avlbl_behavior size = " << this->msg_BA2Obs.m_ba_lat_avlbl_behavior.size() << std::endl;

			this->msg_BA2Obs.m_ba_passage_ob_lst.clear(); //20220322 added
			for (auto passageob : this->BA_PassageObLst)
			{
				if (passageob.passage_id.x == 1)
				{
					this->msg_BA2Obs.m_ba_passage_ob_lst.emplace_back(passageob);
				}
			}
			this->BA_PassageObLst.clear();
			return true;
		}

		void BehaviorAnalysis::UpdateOrgLaneInfo()
		{
			if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange)
			{
				this->BA_OrgLaneStID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
				this->BA_OrgLaneEndID = this->input_BA.RM_RoadModel.car_position.ego_lane_id; // to be updated, the end lane id
			}
		}

		float BehaviorAnalysis::GetTimeGapInfo(const uint8_t &slctDistLvl)
		{
			switch (slctDistLvl)
			{
			case BA::SeldDistLvl_DistLvl1:
				return k_BA_TmGapLvl0;
			case BA::SeldDistLvl_DistLvl2:
				return k_BA_TmGapLvl1;
			case BA::SeldDistLvl_DistLvl3:
				return k_BA_TmGapLvl2;
			default:
				return k_BA_TmGapLvlx;
			}
		}

		bool BehaviorAnalysis::GetTrgtObjInfo(uint32_t &objID, float &objRelDist, float &objVel, Corridor &objBlgCrrd, bool &lineFllwCrrdrFound)
		{
			if (this->input_BA.RM_RoadModel.lane_list.empty())
			{
				DLOG(WARNING) << "lane list is empty!" << std::endl;
				return false;
			}
			else
			{
				// init tmp obj attributes
				uint8_t tmpObjID = 0;
				float tmpObjDst = 1000;
				float tmpObjVel = 0;
				Corridor tmpObjBlgCrrd;
				tmpObjBlgCrrd.corridor_id = uint32_t(0);
				// if lane number is bigger than 1, go to find ego lane
				int laneNum = this->input_BA.RM_RoadModel.lane_list.size();
				if (laneNum > 1)
				{
					for (int x = 0; x < laneNum; x++)
					{
						if (this->input_BA.RM_RoadModel.lane_list.at(x).start_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id)
						{
							// ego lane has been found, then go to check corridor list
							// check whether there exists corridor with line_following tag
							Lane laneObj = this->input_BA.RM_RoadModel.lane_list.at(x);
							if (laneObj.corridor_list.empty())
							{
								DLOG(WARNING) << "corridor list is empty!" << std::endl;
								return false;
							}
							else
							{
								int crrdNum = laneObj.corridor_list.size();
								// bool lineFllwCrrdrFound = false;
								for (int y = 0; y < crrdNum; y++)
								{
									if (laneObj.corridor_list.at(y).feature_tag == BA::FeatureTag_LineFllw)
									{
										tmpObjBlgCrrd = laneObj.corridor_list.at(y);
										std::vector<ObjectLocation> objLocList = laneObj.corridor_list.at(y).object_in_corridor_list;
										for (int z = 0; z < objLocList.size(); z++)
										{
											if (objLocList[z].reference_point2corridor_frenet.s < tmpObjDst && objLocList[z].reference_point2corridor_frenet.s > 0)
											{
												tmpObjID = objLocList[z].object_id;
												tmpObjVel = objLocList[z].reference_point2corridor_frenet.ds;
												tmpObjDst = objLocList[z].reference_point2corridor_frenet.s;
											}
										}
									}
									lineFllwCrrdrFound = true;
								}
								// check carFollow corridor
								if (lineFllwCrrdrFound == false)
								{
									for (int y = 0; y < crrdNum; y++)
									{
										if (laneObj.corridor_list.at(y).feature_tag == BA::FeatureTag_CarFllw)
										{
											tmpObjBlgCrrd = laneObj.corridor_list.at(y);
											std::vector<ObjectLocation> objLocList = laneObj.corridor_list.at(y).object_in_corridor_list;
											for (int z = 0; z < objLocList.size(); z++)
											{
												if (objLocList[z].reference_point2corridor_frenet.s < tmpObjDst && objLocList[z].reference_point2corridor_frenet.s > 0)
												{
													tmpObjID = objLocList[z].object_id;
													tmpObjVel = objLocList[z].reference_point2corridor_frenet.ds;
													tmpObjDst = objLocList[z].reference_point2corridor_frenet.s;
												}
											}
										}
									}
								}
							}
							break;
						}
					}
				}
				else
				{
					// if lane number is equal to 1
					// search in corridor list
					Lane laneObj = this->input_BA.RM_RoadModel.lane_list.at(0);
					if (laneObj.start_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id)
					{
						if (laneObj.corridor_list.empty())
						{
							DLOG(WARNING) << "corridor list is empty!" << std::endl;
							return false;
						}
						else
						{
							int crrdNum = laneObj.corridor_list.size();
							// bool lineFllwCrrdrFound = false;
							for (int y = 0; y < crrdNum; y++)
							{
								if (laneObj.corridor_list.at(y).feature_tag == BA::FeatureTag_LineFllw)
								{
									tmpObjBlgCrrd = laneObj.corridor_list.at(y);
									std::vector<ObjectLocation> objLocList = laneObj.corridor_list.at(y).object_in_corridor_list;
									for (int z = 0; z < objLocList.size(); z++)
									{
										if (objLocList[z].reference_point2corridor_frenet.s < tmpObjDst && objLocList[z].reference_point2corridor_frenet.s > 0)
										{
											tmpObjID = objLocList[z].object_id;
											tmpObjVel = objLocList[z].reference_point2corridor_frenet.ds;
											tmpObjDst = objLocList[z].reference_point2corridor_frenet.s;
										}
									}
								}
								lineFllwCrrdrFound = true;
							}
							// check carFollow corridor
							if (lineFllwCrrdrFound == false)
							{
								for (int y = 0; y < crrdNum; y++)
								{
									if (laneObj.corridor_list.at(y).feature_tag == BA::FeatureTag_CarFllw)
									{
										tmpObjBlgCrrd = laneObj.corridor_list.at(y);
										std::vector<ObjectLocation> objLocList = laneObj.corridor_list.at(y).object_in_corridor_list;
										for (int z = 0; z < objLocList.size(); z++)
										{
											if (objLocList[z].reference_point2corridor_frenet.s < tmpObjDst && objLocList[z].reference_point2corridor_frenet.s > 0)
											{
												tmpObjID = objLocList[z].object_id;
												tmpObjVel = objLocList[z].reference_point2corridor_frenet.ds;
												tmpObjDst = objLocList[z].reference_point2corridor_frenet.s;
											}
										}
									}
								}
							}
						}
					}
				}
				objID = tmpObjID;
				objRelDist = tmpObjDst;
				objVel = tmpObjVel;
				objBlgCrrd = tmpObjBlgCrrd;
				return true;
			}
		}

		bool BehaviorAnalysis::JudgeNoLineNoCarCase()
		{
			if (this->input_BA.RM_RoadModel.lane_list.size() == 1)
			{
				// virtual lane type is 13?
				if (this->input_BA.RM_RoadModel.lane_list.at(0).lane_type == uint8_t(13))
				{
					// the only lane is virtual lane
					if (this->input_BA.RM_RoadModel.lane_list.at(0).corridor_list.size() == 1)
					{
						if (this->input_BA.RM_RoadModel.lane_list.at(0).corridor_list.at(0).feature_tag == BA::FeatureTag_LineMiss)
						{
							// the only corridor is line_miss
							return true;
						}
					}
				}
			}
			return false;
		}

		uint32_t BehaviorAnalysis::CalcRollCounter()
		{
			static uint32_t counter = 0;
			if (counter == 10000)
			{
				counter = 0;
			}
			else
			{
				counter++;
			}
			return counter;
		}

		bool BehaviorAnalysis::GetILCDrct(BA::ILCDirection &drctOfILC)
		{
			// compare optimal corridor c0 with line-follow corridor c0,
			// if the former is greater than the latter, then ILC is on the left side
			// or ILC is on the right side
			bool isLnFllwCrrdrFound = false;
			bool isOptCrrdrFound = false;
			float lnFllwCrrdrC0 = 0.0f;
			float optCrrdrC0 = 0.0f;
			drctOfILC = BA::ILCDirection_Unknown; // initialization
			bool resFlg = false;
			// get ego lane
			for (int i = 0; i < input_BA.RM_RoadModel.lane_list.size(); i++)
			{
				if (input_BA.RM_RoadModel.car_position.ego_lane_id == input_BA.RM_RoadModel.lane_list.at(i).start_id)
				{
					bool isLnFllwCrrdrFound = false;
				    bool isOptCrrdrFound = false;
					// get interested corridor
					for (int j = 0; j < input_BA.RM_RoadModel.lane_list.at(i).corridor_list.size(); j++)
					{
						// find line-follow corridor
						if (input_BA.RM_RoadModel.lane_list.at(i).corridor_list.at(j).feature_tag == BA::FeatureTag_LineFllw)
						{
							isLnFllwCrrdrFound = true;
							lnFllwCrrdrC0 = input_BA.RM_RoadModel.lane_list.at(i).corridor_list.at(j).reference_line.c0;
						}
						if (input_BA.RM_RoadModel.lane_list.at(i).corridor_list.at(j).corridor_id == this->BA_OptCorridorID_Deb)
						{
							isOptCrrdrFound = true;
							optCrrdrC0 = input_BA.RM_RoadModel.lane_list.at(i).corridor_list.at(j).reference_line.c0;
						}
						if (isLnFllwCrrdrFound == true && isOptCrrdrFound == true)
						{
							break;
						}
					}
					if (isLnFllwCrrdrFound == true && isOptCrrdrFound == true)
					{
						if (std::abs(lnFllwCrrdrC0 - optCrrdrC0) > ADC::common::k_BA_ILCCrrdC0DiffThr)
						{
							if (lnFllwCrrdrC0 > optCrrdrC0)
							{
								drctOfILC = BA::ILCDirection_Right;
							}
							else
							{
								drctOfILC = BA::ILCDirection_Left;
							}
						}
						resFlg = true;
					}
				}
				if (resFlg == true)
				{
					break;
				}
			}
			if (drctOfILC == BA::ILCDirection_Unknown)
			{
				return false;
			}
			else
			{
				return true;
			}
		}

		bool BehaviorAnalysis::UpdateGlobalInfo(const bool &toUpdtGlbInfo, GlobalInfoProcess &globalInfoProc)
		{
			if (toUpdtGlbInfo == true)
			{
				static uint16_t cycleCntLoc = 0;
				static float ccDurTm = 0.0f;
				// calc the duration time during which EHR cycleCounter keeps the same
				if (cycleCntLoc == this->input_BA.EHR_EHRInfo.msg_header.cycle_counter)
				{
					ccDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
					if (ccDurTm >= 300)
					{
						ccDurTm = 1.0;
					}
				}
				else
				{
					ccDurTm = 0.0f;
					cycleCntLoc = this->input_BA.EHR_EHRInfo.msg_header.cycle_counter;
				}

				if (ccDurTm < ADC::common::k_BA_EHRMsgStopUpdtThr)
				{
					globalInfoProc.initGlobalInfo(this->input_BA.EHR_EHRInfo, this->input_BA.RM_RoadModel, this->input_BA.VSE_LongitudinalVelocity);
					// 20220918 protection logic added
					if (std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), this->input_BA.RM_RoadModel.car_position.ego_lane_id) == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
					{
						if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_NotAvlbl)
						{
							if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == uint8_t(0))
							{
								globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum = uint8_t(1);
								if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
								{
									globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct = BA::LaneChgDrct_Right;
								}
								else
								{
									globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct = BA::LaneChgDrct_Left;
								}
							}
						}
					}
					float avgVehFlowDense = CalcAvgVehFlowDense(globalInfoProc);
					DLOG(INFO) << "average vehicle flow dense = " << avgVehFlowDense << std::endl;
					this->BA_AvgFlowDenseHigh = false;
					if (avgVehFlowDense >= ADC::common::k_BA_MinVehFlowDense)
					{
						// vehicle flow dense is relatively great
						this->BA_AvgFlowDenseHigh = true;
					}
					DebounceFlowDense();
					DLOG(INFO) << "is vehicle flow dense high = " << this->BA_AvgFlowDenseHigh << std::endl;
					DLOG(INFO) << "is vehicle flow dense high deb = " << this->BA_AvgFlowDenseHigh_Deb << std::endl;
					globalInfoProc.processGlobalInfo(this->input_BA.VSE_LongitudinalVelocity, this->BA_AvgFlowDenseHigh_Deb, globalInfoProc.globalInfoContainer.m_RouteRltvPos);
					return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				// updating global cost is negative
				return false;
			}
		}

		bool BehaviorAnalysis::UpdateGlobalCost(const bool &toUpdtGlbCost, GlobalInfoProcess &globalInfoProc, const bool &allwLCAntiDrctCond)
		{
			bool isEffcLCIntGnrt = false;
			GlobalCostCalculation globalCostCalc;
			bool vldWayPntDtc = false;
			// if only virtual control waypoint exists, then there is no valid waypoint
			if (globalInfoProc.globalInfoContainer.m_GlobalInfoSrc != BA::GlobalInfoSrc_OnlyVirtCtrlWayPoint && globalInfoProc.globalInfoContainer.m_GlobalInfoSrc != BA::GlobalInfoSrc_NoVldWayPoint)
			{
				vldWayPntDtc = true;
			}
			DLOG(INFO) << "vldWayPntDtc = " << int(vldWayPntDtc) << std::endl;
			bool isSplitTrgtRdTyp = (/* globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd || */
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd);
			bool isLCIntCntRst = false;
			if (toUpdtGlbCost == false || isSplitTrgtRdTyp == true)
			{
				isLCIntCntRst = true;
			}
			// 20220622: select ego speed or front object speed in ego lane
			float egoVehSpd = this->input_BA.VSE_LongitudinalVelocity;
			float frtVehSpd = this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh;
			uint8_t egoLaneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			uint8_t lftNgbrLaneID = uint8_t(0);
			uint8_t rghtNgbrLaneID = uint8_t(0);
			uint8_t frtVehTyp = uint8_t(0);
			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
										[egoLaneID](Lane &lane_)
										{
											return (egoLaneID == lane_.start_id);
										});
			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
			{
				auto it_crrd = std::find_if(it_lane->corridor_list.begin(), it_lane->corridor_list.end(),
											[](Corridor &corridor_)
											{
												return (BA::FeatureTag_LineFllw == corridor_.feature_tag);
											});
				if (it_crrd != it_lane->corridor_list.end())
				{
					std::vector<ObjectLocation> tmpObjLocList;
					for (auto objLoc_ : it_crrd->object_in_corridor_list)
					{
						if (objLoc_.relative_center_x > 0)
						{
							tmpObjLocList.emplace_back(objLoc_);
						}
					}
			    	// sort tmpObjLocList by relative center_x
					std::sort(tmpObjLocList.begin(), tmpObjLocList.end(), Sort_by_distX);
					if (!tmpObjLocList.empty())
					{
						frtVehSpd = tmpObjLocList.front().reference_point2corridor_frenet.ds;
					}
				}

				// 20220822 added
				// get left neighbor lane id
				if (!it_lane->left_neighbor_lane.empty())
				{
					for (auto lftNgbrLn : it_lane->left_neighbor_lane)
					{
						if (lftNgbrLn.start_range < 0.0f)
						{
							lftNgbrLaneID = lftNgbrLn.lane_id;
							break;
						}
					}
				}
				// get right neighbor lane id
				if (!it_lane->right_neighbor_lane.empty())
				{
					for (auto rghtNgbrLn : it_lane->right_neighbor_lane)
					{
						if (rghtNgbrLn.start_range < 0.0f)
						{
							rghtNgbrLaneID = rghtNgbrLn.lane_id;
							break;
						}
					}
				}
				auto it_rghtlane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
												[rghtNgbrLaneID](Lane &lane_)
												{
													return (rghtNgbrLaneID == lane_.start_id);
												});
				if (it_rghtlane != this->input_BA.RM_RoadModel.lane_list.end() && rghtNgbrLaneID != uint8_t(0))
				{
					auto it_crrd = std::find_if(it_rghtlane->corridor_list.begin(), it_rghtlane->corridor_list.end(),
											[](Corridor &corridor_)
											{
												return (BA::FeatureTag_LineFllw == corridor_.feature_tag);
											});
					if (it_crrd != it_rghtlane->corridor_list.end())
					{
						std::vector<ObjectLocation> tmpObjLocList01;
						for (auto objLoc_ : it_crrd->object_in_corridor_list)
						{
							if (objLoc_.relative_center_x > 0)
							{
								tmpObjLocList01.emplace_back(objLoc_);
							}
						}
			    		// sort tmpObjLocList01 by relative center_x
						std::sort(tmpObjLocList01.begin(), tmpObjLocList01.end(), Sort_by_distX);
						if (!tmpObjLocList01.empty())
						{
							frtVehTyp = tmpObjLocList01.front().classification;
						}
					}
				}
			}
			IsLCIntGnrt(frtVehSpd,
						this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
						this->input_BA.RM_RoadModel.car_position,
						isLCIntCntRst,
						isEffcLCIntGnrt,
						this->input_BA.SM_DriveMode);
			DLOG(INFO) << "isEffcLCIntGnrt: " << int(isEffcLCIntGnrt) << std::endl;
			DLOG(INFO) << "isLCIntCntRst: " << int(isLCIntCntRst) << std::endl;
			static float rghtLnSpdCondTm = 0.0f;
			if (toUpdtGlbCost == true)
			{
				uint8_t lftNbghLnID = uint8_t(0);
				uint8_t rghtNbghLnID = uint8_t(0);
				// 20221114: 
				UpdateLftNghrDesLCPnt(allwLCAntiDrctCond, globalInfoProc);
				// ##########################################
				ProcessLnVehFlowSpdCond(lftNbghLnID, rghtNbghLnID, allwLCAntiDrctCond, globalInfoProc);
				DLOG(INFO) << "-- avg spd lane id = " << int(lftNbghLnID) << std::endl;
				DLOG(INFO) << "-- flow spd cond flag = " << int(this->BA_LeftLnSpdCond_Deb) << std::endl;
				DLOG(INFO) << "-- avg spd lane id = " << int(rghtNbghLnID) << std::endl;
				DLOG(INFO) << "-- flow spd cond flag = " << int(this->BA_RightLnSpdCond_Deb) << std::endl;
				// 20220822: add special limit condition 01
				bool spclLmtCond01 = false;
				// get total lane number
				int totalLaneNum = 0;
				if (!this->input_BA.EHR_EHRInfo.control_waypoints_list.empty())
				{
					for (auto extendLane : this->input_BA.EHR_EHRInfo.control_waypoints_list.front().control_wp_extendlane_list)
    				{
        				if (extendLane.is_emergency_lane == uint8_t(0) && extendLane.current_extendlane_offset < 1.0f)
        				{
            				totalLaneNum++;
        				}
    				}
				}
				// get left corner lane id
				uint8_t lftCrnrLnID = globalInfoProc.globalInfoContainer.m_LftCrnrLaneID;
				if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd && 
				    isEffcLCIntGnrt == true && 
					totalLaneNum >= 4 && 
					this->input_BA.SM_DrvrSetSpd >= 90.0f &&
					lftNgbrLaneID == lftCrnrLnID)
				{
					DLOG(INFO) << "get inside flag!" << std::endl;
					DLOG(INFO) << "front vehicle type = " << int(frtVehTyp) << std::endl;
					// sub condition 01: right lane speed condition is satisfied for 15 sec
					if (this->BA_RightLnSpdCond_Deb == true)
					{
						rghtLnSpdCondTm += BEHAVIORANALYZERCYCLE / 1000.0f;
						if (rghtLnSpdCondTm > 10.0f)
						{
							rghtLnSpdCondTm = 10.1f;
						}
					}
					else
					{
						rghtLnSpdCondTm = 0.0f;
					}
					if (rghtLnSpdCondTm < 10.0f)
					{
						spclLmtCond01 = true;
					}
					if (frtVehTyp == 2)
					{
						spclLmtCond01 = true;
					}
				}
				else
				{
					spclLmtCond01 = false;
					rghtLnSpdCondTm = 0.0f;
				}
				DLOG(INFO) << "special limit condition 01 = " << int(spclLmtCond01) << std::endl;
				DLOG(INFO) << "rghtLnSpdCondTm = " << rghtLnSpdCondTm << std::endl;

				// 20220829: add guide cost for lane change to leave right second lane
				// if right neighbor lane id equals to right corner lane id, total lane number is greater than 3, ctrl waypoint is virtFixDistRd
				bool isEgoOnRghtSndLnFlg = false;
				uint8_t rightCrnrLnID = globalInfoProc.globalInfoContainer.m_RghtCrnrLaneID;
				if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd && 
					totalLaneNum >= 4 && 
					this->input_BA.SM_DrvrSetSpd >= 90.0f &&
					(rghtNgbrLaneID == rightCrnrLnID || egoLaneID == rightCrnrLnID))
				{
					isEgoOnRghtSndLnFlg = true;
				}
				DLOG(INFO) << "isEgoOnRghtSndLnFlg = " << int(isEgoOnRghtSndLnFlg) << std::endl;

				bool tmpSpdCondFlg = false;
				// update BA_TerminatedBehaviorNodesVector
				for (int x = 0; x < this->BA_TerminatedBehaviorNodesVector->size(); x++)
				{
					float tmpAvgSpd = 0;
					// calc average speed of the BA_TerminatedBehaviorNode
					if (!this->BA_TerminatedBehaviorNodesVector->at(x).m_HVInfoGlobal.m_LaneLocation.empty())
					{
						uint8_t tmpLaneID = this->BA_TerminatedBehaviorNodesVector->at(x).m_HVInfoGlobal.m_LaneLocation.back();
						// 20220809: determine speed condition flag
						if (tmpLaneID == lftNbghLnID)
						{
							tmpSpdCondFlg = this->BA_LeftLnSpdCond_Deb;
						}
						else if (tmpLaneID == rghtNbghLnID)
						{
							if (spclLmtCond01 == false)
							{
								tmpSpdCondFlg = this->BA_RightLnSpdCond_Deb;
							}
							else
							{
								tmpSpdCondFlg = false;
							}
						}
						else
						{
							tmpSpdCondFlg = false;
						}
					}

					globalCostCalc.Init(globalInfoProc.globalInfoContainer,
										this->BA_TerminatedBehaviorNodesVector->at(x).m_HVInfoGlobal,
										this->input_BA.RM_RoadModel,
										this->BA_TerminatedBehaviorNodesVector->at(x).m_ActionTypArr);
					bool isUpdtHistVehInfo = true;
					globalCostCalc.UpdtHistVehInfo(this->input_BA.RM_RoadModel.car_position,
												   this->input_BA.VSE_LongitudinalVelocity,
												   isUpdtHistVehInfo);
					if (this->BA_TerminatedBehaviorNodesVector->at(x).m_ActionTypArr.size() > 1)
					{
						DLOG(INFO) << "Evaluated Action: " << int(this->BA_TerminatedBehaviorNodesVector->at(x).m_ActionTypArr.at(1)) << std::endl;
					}
					else
					{
						DLOG(INFO) << "Action not found!" << std::endl;
					}
					// 20221017: 
					// if sm_lcavlblbhvr is right lane change auto or left lane change auto
					// then stop calculationg global costs
					if (this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_LeftLaneChange_Auto && this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_RightLaneChange_Auto)
					{
						globalCostCalc.CalculateGlobalCost(this->input_BA.EHR_EHRInfo,
														   this->input_BA.RM_RoadModel,
														   this->input_BA.RM_RoadModel.car_position,
														   this->input_BA.VSE_LongitudinalVelocity,
														   vldWayPntDtc,
														   isEffcLCIntGnrt,
														   this->input_BA.SM_DriveMode,
														   tmpSpdCondFlg,
														   isEgoOnRghtSndLnFlg);
					}

					// TODO: global and local cost weight assignment 20220129
					// copy m_GlobalCost value into m_BACalcGlobalCost for monitoring
					// m_TotalCost = 0.3 * LocalTotalCost + 0.7 * GlobalTotalCost
					this->BA_TerminatedBehaviorNodesVector->at(x).m_BACalcGlobalCost = globalCostCalc.m_GlobalCost;
					this->BA_TerminatedBehaviorNodesVector->at(x).m_TotalCost = (1 - ADC::common::k_BA_GlobalCostCoeff) * this->BA_TerminatedBehaviorNodesVector->at(x).m_TotalCost + ADC::common::k_BA_GlobalCostCoeff * globalCostCalc.m_GlobalCost;
				}
				// 20221104:
				// calc wait for lane change duration time
				if (isEffcLCIntGnrt == true)
				{
					this->BA_Wait4LCDurTm += BEHAVIORANALYZERCYCLE / 1000.0f; 
				}
				else
				{
					this->BA_Wait4LCDurTm = 0.0f;
				}
				return true;
			}
			else
			{
				// updating global cost is negative
				this->BA_Wait4LCDurTm = 0.0f;
				return false;
			}
		}

		void BehaviorAnalysis::SwitchBhvrNodeGlobalVec(std::vector<DM::BehaviorNodeGlobal> &newBhvrNodeGlobalVec)
		{
			newBhvrNodeGlobalVec.clear();
			DM::BehaviorNodeGlobal tmp;
			std::vector<uint8_t> tmpActionArr;
			std::vector<BA::BA_ActionTyp> curActionArr;
			for (int i = 0; i < this->BA_TerminatedBehaviorNodesVector->size(); i++)
			{
				tmp.m_node_id.x = this->BA_TerminatedBehaviorNodesVector->at(i).m_NodeID.m_X;
				tmp.m_node_id.y = this->BA_TerminatedBehaviorNodesVector->at(i).m_NodeID.m_Y;
				tmp.m_total_cost = this->BA_TerminatedBehaviorNodesVector->at(i).m_TotalCost;
				tmp.m_process_cost = this->BA_TerminatedBehaviorNodesVector->at(i).m_ProcessCost;
				tmp.m_ba_calc_cost = this->BA_TerminatedBehaviorNodesVector->at(i).m_BACalcCost;
				tmp.m_duration_time_arr = this->BA_TerminatedBehaviorNodesVector->at(i).m_DurationTimeArr;
				tmpActionArr.clear();
				curActionArr = this->BA_TerminatedBehaviorNodesVector->at(i).m_ActionTypArr;
				for (int j = 0; j < curActionArr.size(); j++)
				{
					tmpActionArr.emplace_back(uint8_t(curActionArr[j]));
				}
				tmp.m_action_typ_arr = tmpActionArr;
				tmp.m_corridor_id_arr = this->BA_TerminatedBehaviorNodesVector->at(i).m_CorridorIDArr;
				tmp.m_target_obj_id_arr = this->BA_TerminatedBehaviorNodesVector->at(i).m_TargetObjIDArr;
				tmp.m_hv_info_global.m_distance_along_s = this->BA_TerminatedBehaviorNodesVector->at(i).m_HVInfoGlobal.m_DistanceAlongS;
				tmp.m_hv_info_global.m_velocity = this->BA_TerminatedBehaviorNodesVector->at(i).m_HVInfoGlobal.m_Velocity;
				tmp.m_hv_info_global.m_lane_location = this->BA_TerminatedBehaviorNodesVector->at(i).m_HVInfoGlobal.m_LaneLocation;
				tmp.m_first_corridor_id = this->BA_TerminatedBehaviorNodesVector->at(i).m_FirstCorridorID;
				tmp.m_is_node_terminated = this->BA_TerminatedBehaviorNodesVector->at(i).m_IsNodeTerminated;
				tmp.m_total_duration_time = this->BA_TerminatedBehaviorNodesVector->at(i).m_TotalDurationTime;
				tmp.m_first_passage_id.x = this->BA_TerminatedBehaviorNodesVector->at(i).m_FirstPassageID.m_X;

				newBhvrNodeGlobalVec.emplace_back(tmp);
			}
		}

		void BehaviorAnalysis::SwitchLatAvlblBhvrVec(std::vector<uint8_t> &newLatAvlblBhvrVec)
		{
			newLatAvlblBhvrVec.clear();
			for (int k = 0; k < this->BA_LatAvlblBehavior.size(); k++)
			{
				newLatAvlblBhvrVec.emplace_back(uint8_t(this->BA_LatAvlblBehavior[k]));
			}
		}

		bool BehaviorAnalysis::Sort_by_distX(const ObjectLocation &objLoc1, const ObjectLocation &objLoc2)
		{
			return objLoc1.relative_center_x < objLoc2.relative_center_x;
		}

		bool BehaviorAnalysis::GetFrtObjToSM(std::vector<ObjectTrim> &objLstTrim, const bool &trgtObjFound, const bool &lineFllwCrrdrFound, const uint32_t &trgtObjID)
		{
			// init object id to 0
			int objNum = 6;
			for (int i = 0; i < objNum; i++)
			{
				objLstTrim.at(i).object_id = 0;
				objLstTrim.at(i).object_exist_probability = 0;
			}

			if (trgtObjFound == true)
			{
				// lane list is not empty
				if (trgtObjID != uint32_t(0))
				{
					// assign object 01 value
					if (!LoadObjTrim(trgtObjID, objLstTrim.at(0)) == true)
					{
						return false;
					}

					if (lineFllwCrrdrFound == true)
					{
						// try to find object 02
						uint8_t egoLaneID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
						auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
													[egoLaneID](Lane &lane_)
													{
														return (egoLaneID == lane_.start_id);
													});
						if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
						{
							auto it_crrd = std::find_if(it_lane->corridor_list.begin(), it_lane->corridor_list.end(),
														[](Corridor &corridor_)
														{
															return (BA::FeatureTag_LineFllw == corridor_.feature_tag);
														});
							if (it_crrd != it_lane->corridor_list.end())
							{
								std::vector<ObjectLocation> tmpObjLocList;
								for (auto objLoc_ : it_crrd->object_in_corridor_list)
								{
									if (objLoc_.relative_center_x > 0)
									{
										tmpObjLocList.emplace_back(objLoc_);
									}
								}
								// sort tmpObjLocList by relative center_x
								std::sort(tmpObjLocList.begin(), tmpObjLocList.end(), Sort_by_distX);
								if (tmpObjLocList.size() > 1)
								{
									uint32_t obj2ID = tmpObjLocList.at(1).object_id;
									if (!LoadObjTrim(obj2ID, objLstTrim.at(1)) == true)
									{
										DLOG(INFO) << "object 02 assignment failed!" << std::endl;
									}
								}
							}

							// try to find object 03 and 05
							if (!it_lane->left_neighbor_lane.empty())
							{
								uint8_t lftNeighborLnID = it_lane->left_neighbor_lane.front().lane_id;
								auto it_lftLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
															   [lftNeighborLnID](Lane &lftLane_)
															   {
																   return (lftNeighborLnID == lftLane_.start_id);
															   });

								if (it_lftLane != this->input_BA.RM_RoadModel.lane_list.end())
								{
									auto it_lftCrrd = std::find_if(it_lftLane->corridor_list.begin(), it_lftLane->corridor_list.end(),
																   [](Corridor &lftCorridor_)
																   {
																	   return (BA::FeatureTag_LineFllw == lftCorridor_.feature_tag);
																   });
									if (it_lftCrrd != it_lftLane->corridor_list.end())
									{
										std::vector<ObjectLocation> tmpObjLocList;
										for (auto objLoc_ : it_lftCrrd->object_in_corridor_list)
										{
											if (objLoc_.relative_center_x > 0)
											{
												tmpObjLocList.emplace_back(objLoc_);
											}
										}
										// sort tmpObjLocList by relative center_x
										std::sort(tmpObjLocList.begin(), tmpObjLocList.end(), Sort_by_distX);
										if (tmpObjLocList.size() == 1)
										{
											uint32_t obj5ID = tmpObjLocList.at(0).object_id;
											if (!LoadObjTrim(obj5ID, objLstTrim.at(4)) == true)
											{
												DLOG(INFO) << "object 05 assignment failed!" << std::endl;
												// return false;
											}
										}
										else if (tmpObjLocList.size() > 1)
										{
											uint32_t obj5ID = tmpObjLocList.at(0).object_id;
											if (!LoadObjTrim(obj5ID, objLstTrim.at(4)) == true)
											{
												DLOG(INFO) << "object 05 assignment failed!" << std::endl;
												// return false;
											}

											uint32_t obj3ID = tmpObjLocList.at(1).object_id;
											if (!LoadObjTrim(obj3ID, objLstTrim.at(2)) == true)
											{
												DLOG(INFO) << "object 03 assignment failed!" << std::endl;
												// return false;
											}
										}
									}
								}
							}

							// try to find object 04 and 06
							if (!it_lane->right_neighbor_lane.empty())
							{
								uint8_t rghtNeighborLnID = it_lane->right_neighbor_lane.front().lane_id;
								auto it_rghtLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
																[rghtNeighborLnID](Lane &rghtLane_)
																{
																	return (rghtNeighborLnID == rghtLane_.start_id);
																});

								if (it_rghtLane != this->input_BA.RM_RoadModel.lane_list.end())
								{
									auto it_rghtCrrd = std::find_if(it_rghtLane->corridor_list.begin(), it_rghtLane->corridor_list.end(),
																	[](Corridor &rghtCorridor_)
																	{
																		return (BA::FeatureTag_LineFllw == rghtCorridor_.feature_tag);
																	});
									if (it_rghtCrrd != it_rghtLane->corridor_list.end())
									{
										std::vector<ObjectLocation> tmpObjLocList;
										for (auto objLoc_ : it_rghtCrrd->object_in_corridor_list)
										{
											if (objLoc_.relative_center_x > 0)
											{
												tmpObjLocList.emplace_back(objLoc_);
											}
										}
										// sort tmpObjLocList by relative center_x
										std::sort(tmpObjLocList.begin(), tmpObjLocList.end(), Sort_by_distX);
										if (tmpObjLocList.size() == 1)
										{
											uint32_t obj6ID = tmpObjLocList.at(0).object_id;
											if (!LoadObjTrim(obj6ID, objLstTrim.at(5)) == true)
											{
												DLOG(INFO) << "object 06 assignment failed!" << std::endl;
												// return false;
											}
										}
										else if (tmpObjLocList.size() > 1)
										{
											uint32_t obj6ID = tmpObjLocList.at(0).object_id;
											if (!LoadObjTrim(obj6ID, objLstTrim.at(5)) == true)
											{
												DLOG(INFO) << "object 06 assignment failed!" << std::endl;
												// return false;
											}

											uint32_t obj4ID = tmpObjLocList.at(1).object_id;
											if (!LoadObjTrim(obj4ID, objLstTrim.at(3)) == true)
											{
												DLOG(INFO) << "object 04 assignment failed!" << std::endl;
												// return false;
											}
										}
									}
								}
							}
						}
					}
				}
			}
			return true;
		}

		bool BehaviorAnalysis::LoadObjTrim(const uint32_t &objID, ObjectTrim &objTrim)
		{
			auto it_obj = std::find_if(this->input_BA.RM_RoadModel.object_list.begin(), this->input_BA.RM_RoadModel.object_list.end(),
									   [objID](const Object &objectObj_)
									   {
										   return (objID == objectObj_.object_id);
									   });
			if (it_obj != this->input_BA.RM_RoadModel.object_list.end())
			{
				// object is found
				objTrim.object_id = objID;
				objTrim.object_exist_probability = it_obj->object_exist_probability;
				objTrim.classification = it_obj->classification;
				objTrim.relative_center_x = it_obj->relative_center_x;
				objTrim.relative_center_y = it_obj->relative_center_y;
				objTrim.relative_position_x = it_obj->relative_position_x;
				objTrim.relative_position_y = it_obj->relative_position_y;
				objTrim.absolute_velocity_x = it_obj->absolute_velocity_x;
				objTrim.absolute_velocity_y = it_obj->absolute_velocity_y;
				objTrim.absolute_acceleration_x = it_obj->absolute_acceleration_x;
				objTrim.absolute_acceleration_y = it_obj->absolute_acceleration_y;
				objTrim.absolute_yawrate = it_obj->absolute_yawrate;
				objTrim.size_length = it_obj->size_length;
				objTrim.size_width = it_obj->size_width;

				return true;
			}
			else
			{
				return false;
			}
		}

		bool BehaviorAnalysis::CalcIntrSpprAction(GlobalInfoProcess &globalInfoProc, 
												  const bool &crvCond4AutoLC, 
												  const BA::TargetRoadTyp &lastTrgtRdTyp, 
												  bool &allwLCAntiDrctCond)
		{
			this->BA_SupprActLst.clear();
			//1. if waypoint is JCT/IC/RmpDvdRd and distance btw ego car and waypoint is less than 1200m and
			//      lane change number is 0, namely ego car is on the target lane and
			//      no commanded lane change
			//   then suppressed action list should contain BOTH lane change and lane change prepare
			//2. if waypoint is JCT/IC/RmpDvdRd and distance btw ego car and waypoint is less than 1500m and
			//   then suppressed action list should include corresponding lane change and lane change prepare
			//3. if adjacent lane' transit lane type is lane end, and end offset is less than 500m
			//   then suppressed action list should include corresponding lane change and lane change prepare
			//4. if current target road type is VirtFixDistRd OR SubSpltRd, and its certain neighbor lane id is not within target lane list 
			//   then suppressed action list should include corresponding lane change and lane change prepare
			//   in case lane change to non-target lane
			//5. if speed condition for auto lane change is not satisfied %% replaced by curve condition 20220920 %%
			//   then suppressed action list should include corresponding lane change and lane change prepare
			//6. if ego veh is on the corner side, and distance to waypoint is more than 800m, and the total lane change number is no more than 1, and following target type is truck, and distance to last lane change point is more than 700m
			//   then permit lane change to anti-direction
			//7. if last target road type is RmpMrgCARd and duration time is less than 0.3sec
			//   then suppressed action list should include both lane change and lane change prepare --ignored
			//8. if target road type is RmpMrgCARd and distance to waypoint is less than 3m, and current action is not lane change
			//   then suppressed action list should include both lane change and lane change prepare
			//9. if right neighbor lane is emergency lane
			//   then suppressed action list should include right lane change and right lane change prepare
			//A. if left neighbor lane is emergency lane
			//   then suppressed action list should include left lane change and left lane change prepare
			//B. if last behavior is LF, and certain direction indicator is turned on
			//   then suppressed action list should include anti-direction lane change and anti-direction lane change prepare
			//C. if line type is solid
			//   then suppressed action list should include lane change prepare
			//D. if waypoint type is split and target lane list size is 1 and raw target lane list size is greater than 1 and distance to waypoint is less than 1.5km and route is right side
			//   then suppressed action list should include left lane change and left lane change prepare
			//E. if waypoint type is split and neighbor lane endoffset is less than distance to waypoint, 
			//   then suppressed action list should include relevant lane change and lane change prepare
			//F. if neighbor lane has sma info,
			//   then suppressed action list should include relevant lane change and lane change prepare

			bool isSplitTrgtRdTyp = (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd);
			bool isNotCmdLnChg = (this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_LeftLaneChange_C &&
								  this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_RightLaneChange_C &&
								  this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_LeftLaneChange_Auto &&
								  this->input_BA.SM_LCAvlblBehavior != BA::LCAvlblBhvr_RightLaneChange_Auto);
			bool isDistLessThanThrd01 = (globalInfoProc.globalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_SpprLCDistThrd01);
			bool isDistLessThanThrd02 = (globalInfoProc.globalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_SpprLCDistThrd02);
			bool isLCNumZero = (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == uint8_t(0));
			bool isLCLastCycle = false;
			// 20220427: if last cycle optAction is right lane change or left lane change, isLCLastCycle is true
			if (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange || this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange)
			{
				isLCLastCycle = true;
			}
			// 20220806: ego lane is not ending before split point
			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			auto it_ego = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
			                           [egoLnID](const WPExtendLaneStruct &wpLne_)
									   {
										   return (egoLnID == wpLne_.lane_id);
									   });
			float egoLnEndOffset = 0.0f;
			bool isRghtExtLnEmerg = false;
			if (it_ego != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
			{
				egoLnEndOffset = it_ego->current_extendlane_endoffset;
				DLOG(INFO) << "egoLnEndOffset = " << egoLnEndOffset << std::endl;
				// 20221010: get right neighbor lane
				uint8_t rghtExtLnID = uint8_t(0);
				if (!it_ego->right_extend_lane.empty())
				{
					for (auto rghtExtLn : it_ego->right_extend_lane)
					{
						if (rghtExtLn.adjacent_extendlane_offset <= 0.0f)
						{
							rghtExtLnID = rghtExtLn.adjacent_extendlane_id;
							break;
						}
					}
				}
				auto it_rghtExtLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
			                           			 [rghtExtLnID](const WPExtendLaneStruct &wpLne_)
									   			 {
										   			return (rghtExtLnID == wpLne_.lane_id);
									   			 });
				if (it_rghtExtLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					if (it_rghtExtLn->is_emergency_lane == uint8_t(1))
					{
						isRghtExtLnEmerg = true;
					}
				}
			}
			if (isSplitTrgtRdTyp == true &&
				isNotCmdLnChg == true &&
				isDistLessThanThrd01 == true &&
				isLCNumZero == true &&
				isLCLastCycle == false &&
				egoLnEndOffset >= globalInfoProc.globalInfoContainer.m_DistToWayPoint)
			{
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);  // added 20220503
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare); // added 20220503
				DLOG(INFO) << "cond01" << std::endl;
			}

			if (isSplitTrgtRdTyp == true &&
				isDistLessThanThrd02 == true &&
				isLCLastCycle == false &&
				egoLnEndOffset >= globalInfoProc.globalInfoContainer.m_DistToWayPoint)
			{
				if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
				{
					// 20220720: check whether anti-direction neighbor lane is target lane and ego lane is not target lane
					// if positive, no anti-direction lane change suppression
					auto it_egoLnTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), egoLnID);
					if (it_egoLnTrgtLn != globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
					{
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
						DLOG(INFO) << "cond02_1" << std::endl;
					}
					else
					{
						auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
													 [egoLnID](const WPExtendLaneStruct &wpExtLn_)
													 {
														 return (egoLnID == wpExtLn_.lane_id);
													 });
						if (it_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
						{
							if (!it_egoLn->left_extend_lane.empty())
							{
								uint8_t leftNgbrLnID = uint8_t(0);
								for (auto lftLn : it_egoLn->left_extend_lane)
								{
									if (lftLn.adjacent_extendlane_offset < 1.0f && lftLn.adjacent_extendlane_endoffset >= 1.0f)
									{
										leftNgbrLnID = lftLn.adjacent_extendlane_id;
										break;
									}
								}
								if (leftNgbrLnID != uint8_t(0))
								{
									auto it_ngbrLnTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), leftNgbrLnID);
									if (it_ngbrLnTrgtLn == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
									{
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
										DLOG(INFO) << "cond02_2" << std::endl;
									}
								}
							}
						}
					}
					/* this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);  // added 20220503 */
				}
				else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
				{
					// 20220720: check whether anti-direction neighbor lane is target lane and ego lane is not target lane
					// if positive, no anti-direction lane change suppression
					auto it_egoLnTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), egoLnID);
					if (it_egoLnTrgtLn != globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
					{
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
						DLOG(INFO) << "cond02_3" << std::endl;
					}
					else
					{
						auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
													 [egoLnID](const WPExtendLaneStruct &wpExtLn_)
													 {
														 return (egoLnID == wpExtLn_.lane_id);
													 });
						if (it_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
						{
							if (!it_egoLn->right_extend_lane.empty())
							{
								uint8_t rightNgbrLnID = uint8_t(0);
								for (auto rghtLn : it_egoLn->right_extend_lane)
								{
									if (rghtLn.adjacent_extendlane_offset < 1.0f && rghtLn.adjacent_extendlane_endoffset >= 1.0f)
									{
										rightNgbrLnID = rghtLn.adjacent_extendlane_id;
										break;
									}
								}
								if (rightNgbrLnID != uint8_t(0))
								{
									auto it_ngbrLnTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), rightNgbrLnID);
									if (it_ngbrLnTrgtLn == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
									{
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
										DLOG(INFO) << "cond02_4" << std::endl;
									}
								}
							}
						}
					}
					/* this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare); // added 20220503 */
				}
			}
			for (auto transitLane : this->input_BA.EHR_EHRInfo.transit_lanes_list)
			{
				uint8_t transitLaneID = transitLane.transit_lane_id;
				auto it_wpExtLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
							                  [transitLaneID](const WPExtendLaneStruct &wpExtLn_)
											  {
												return (transitLaneID == wpExtLn_.lane_id);
											  });
				bool isTransitLnEmergLn = false;
				if (it_wpExtLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					if (it_wpExtLn->is_emergency_lane == uint8_t(1))
					{
						isTransitLnEmergLn = true;
					}
				}
				// 20221010
				// if right neighbor lane is emergency lane and CWP target road type is virtual fixed distance road, then no suppression below
				// lane end transition type is 13?
				if (transitLane.transit_type == uint8_t(13) && 
				    isTransitLnEmergLn == false && 
					isLCLastCycle == false && 
					(isRghtExtLnEmerg == false || globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_VirtFixDstRd)
					)
				{
					if (transitLane.transit_connected_lane_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id)
					{
						// if the transitLane id belongs to target lane list,
						// then cancel suppression
						auto it_isTransLnInTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), transitLane.transit_lane_id);
						if (it_isTransLnInTrgtLn == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
						{
							if (transitLane.transit_lane_offset <= ADC::common::k_BA_LaneEndOffsetSpprThr)
							{
								// check whether transit lane id is in the right neighbor lane list of current lane
								// if yes, suppress right lane change
								// or, suppress left lane change
								auto it_wpExtLane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
							                               	    [egoLnID](const WPExtendLaneStruct &wpExtLn_)
																{
																	return (egoLnID == wpExtLn_.lane_id);
																});
								// uint8_t transitLaneID = transitLane.transit_lane_id;
								if (it_wpExtLane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									auto it_rghtNgbLn = std::find_if(it_wpExtLane->right_extend_lane.begin(), it_wpExtLane->right_extend_lane.end(),
								                            	    [transitLaneID](const AdjacentExtendLaneStruct &adjLn_)
																	{
																		return (transitLaneID == adjLn_.adjacent_extendlane_id);
																	});
									if (it_rghtNgbLn != it_wpExtLane->right_extend_lane.end())
									{
										// yes
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare); // added 20220503
										DLOG(INFO) << "cond03_1" << std::endl;
									}
									else
									{
										// no
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);  // added 20220503
										DLOG(INFO) << "cond03_2" << std::endl;
									}
								}
							}
						}
					}
				}
			}

			if ((globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd && !globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.empty()) || 
			    (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd && !globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.empty()) &&
				 isLCLastCycle == false)
			{
				if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
				{
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);  // added 20220503
					DLOG(INFO) << "cond04" << std::endl;
				}
				else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
				{
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare); // added 20220503
					DLOG(INFO) << "cond05" << std::endl;
				}
			}
			// added 20220613
			if (crvCond4AutoLC == false && isLCLastCycle == false)
			{
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
				DLOG(INFO) << "cond100" << std::endl;
			}

			// added 20220624
			static float tmpTimer = 10.0f;
			if (this->BA_TmpTrgtRdTypMem == BA::TargetRoadTyp_RmpMrgCARd &&
			    globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_RmpMrgCARd)
			{
				tmpTimer = 0.0f;
			}
			/* if (lastTrgtRdTyp == BA::TargetRoadTyp_RmpMrgCARd)
			{
				tmpTimer = tmpTimer + BEHAVIORANALYZERCYCLE/1000.0;
				DLOG(INFO) << "tmpTimer = " << tmpTimer << std::endl;
				if (tmpTimer > 10.0f)
				{
					tmpTimer = 10.0f;
				}
			} */
			// 20221110: modify tmpTimer accumulation logic
			if (tmpTimer < 10.0f)
			{
				tmpTimer = tmpTimer + BEHAVIORANALYZERCYCLE/1000.0;
			}
			else
			{
				tmpTimer = 10.0f;
			}
			DLOG(INFO) << "tmpTimer = " << tmpTimer << std::endl;
			if (tmpTimer <= ADC::common::k_BA_MinCoolDur4LCRmpMrgCARd &&
				globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == 1 && 
				isLCLastCycle == false)
			{
				// get ego lane type
				// if it is entry or acceleration lane
				// then cooling duration time is 0.6 sec
				auto it_egoLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
											   [egoLnID](const Lane &lane_)
											   {
												   return(egoLnID == lane_.start_id);
											   });
				uint8_t egoLnTyp = uint8_t(0);
				if (it_egoLane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					egoLnTyp = it_egoLane->lane_type;
				}
				// 9: acceleration lane 
				// 16: entry
				if (egoLnTyp == 9 || egoLnTyp == 16)
				{
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
					this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
					DLOG(INFO) << "cond08" << std::endl;
				}
			}

			// added 20220626
			if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd &&
			    globalInfoProc.globalInfoContainer.m_DistToWayPoint <= 3.0f &&
				isLCLastCycle == false)
			{
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
				DLOG(INFO) << "cond09" << std::endl;
			}

			// added 20220711
			auto it_wpExtLane = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
							                 [egoLnID](const WPExtendLaneStruct &wpExtLn_)
											 {
											 	return (egoLnID == wpExtLn_.lane_id);
											 });
			if (it_wpExtLane != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
			{
				// 20220928:
				bool isRghtLnEmerg = false;
				bool isLftLnEmerg = false;
				float rghtLnEndOffset = globalInfoProc.globalInfoContainer.m_DistToWayPoint;
				float lftLnEndOffset = globalInfoProc.globalInfoContainer.m_DistToWayPoint;
				bool isRghtLnExist = false;
				bool isLftLnExist = false;
				if (!it_wpExtLane->right_extend_lane.empty())
				{
					uint8_t rghtExtLnID = uint8_t(0);
					for (auto rghtExtLn : it_wpExtLane->right_extend_lane)
					{
						if (rghtExtLn.adjacent_extendlane_offset < 5.0f)
						{
							rghtExtLnID = rghtExtLn.adjacent_extendlane_id;
							break;
						}
					}
					DLOG(INFO) << "rghtExtLnID = " << int(rghtExtLnID) << std::endl;
					auto it_rghtWpExtLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
													   [rghtExtLnID](const WPExtendLaneStruct &rghtWpExtLn_)
													   {
														   return (rghtExtLnID == rghtWpExtLn_.lane_id);
													   });
					if (it_rghtWpExtLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
					{
						isRghtLnExist = true; // 20220930
				    	if (it_rghtWpExtLn->is_emergency_lane == uint8_t(1) && isLCLastCycle == false)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
							DLOG(INFO) << "cond10" << std::endl;
						}
						if (!this->input_BA.EHR_EHRInfo.reminder_waypoint.empty())
						{
							// 20221117: optimize contraint condition of suppressing lane change to non-target lanes
							// deal with CWP VirtFixDistRd specifically
							if (this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_target_road_type != BA::TargetRoadTyp_NotAvlbl &&
								(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd || 
								 (this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset < globalInfoProc.globalInfoContainer.m_DistToWayPoint &&
								  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_VirtFixDstRd)) &&
						    	this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset < ADC::common::k_BA_SpprLC2RmdWPExtLnDstThr &&
						    	!this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.empty())
					   		{
								auto it_rghtln = std::find(this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.begin(),
														   this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.end(),
														   rghtExtLnID);
								if (it_rghtln == this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.end() && isLCLastCycle == false)
								{
									// the lane should be excluded
									this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
									this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
									DLOG(INFO) << "cond11" << std::endl;
								}
							}
						}
						// 20220928: E
						if (it_rghtWpExtLn->is_emergency_lane == uint8_t(1))
						{
							isRghtLnEmerg = true;
						}
						rghtLnEndOffset = it_rghtWpExtLn->current_extendlane_endoffset;
						// 20220929: F
						if (!it_rghtWpExtLn->sma_info.empty() && isLCLastCycle == false && ADC::common::k_BA_SmaNrrwLnInFlg == true)
						{
							for (auto smaInfo : it_rghtWpExtLn->sma_info)
							{
								if (smaInfo.type == 5)
								{
									if (smaInfo.start_offset < ADC::common::k_BA_SmaNrrwLnOffsetThrd01)
									{
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
										DLOG(INFO) << "cond101" << std::endl;
										break;
									}
								}
							}
						}
					}
					else
					{
						if (isLCLastCycle == false)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
							DLOG(INFO) << "cond19" << std::endl;
						}
					}
				}
				if (!it_wpExtLane->left_extend_lane.empty())
				{
					uint8_t lftExtLnID = uint8_t(0);
					for (auto lftExtLn : it_wpExtLane->left_extend_lane)
					{
						if (lftExtLn.adjacent_extendlane_offset < 1.0f && lftExtLn.adjacent_extendlane_endoffset >= 1.0f)
						{
							lftExtLnID = lftExtLn.adjacent_extendlane_id;
							break;
						}
					}
					DLOG(INFO) << "lftExtLnID = " << int(lftExtLnID) << std::endl;
					auto it_lftWpExtLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
													 [lftExtLnID](const WPExtendLaneStruct &lftWpExtLn_)
													 {
													     return (lftExtLnID == lftWpExtLn_.lane_id);
													 });
					if (it_lftWpExtLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
					{
						isLftLnExist = true; // 20220930
				    	if (it_lftWpExtLn->is_emergency_lane == uint8_t(1) && isLCLastCycle == false)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
							DLOG(INFO) << "cond12" << std::endl;
						}
						if (!this->input_BA.EHR_EHRInfo.reminder_waypoint.empty())
						{
							// 20221117: optimize contraint condition of suppressing lane change to non-target lanes
							// deal with CWP VirtFixDistRd specifically
							if (this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_target_road_type != BA::TargetRoadTyp_NotAvlbl &&
								(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd || 
								 (this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset < globalInfoProc.globalInfoContainer.m_DistToWayPoint &&
								  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_VirtFixDstRd)) &&
						    	this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset < ADC::common::k_BA_SpprLC2RmdWPExtLnDstThr &&
						    	!this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.empty())
							{
								auto it_lftln = std::find(this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.begin(),
														  this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.end(),
														  lftExtLnID);
								if (it_lftln == this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.end() && isLCLastCycle == false)
								{
									// the lane should be excluded
									this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
									this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
									DLOG(INFO) << "cond13" << std::endl;
								}
							}
						}
						// 20220928: E
						if (it_lftWpExtLn->is_emergency_lane == uint8_t(1))
						{
							isLftLnEmerg = true;
						}
						lftLnEndOffset = it_lftWpExtLn->current_extendlane_endoffset;
						// 20220929: F
						if (!it_lftWpExtLn->sma_info.empty() && isLCLastCycle == false && ADC::common::k_BA_SmaNrrwLnInFlg == true)
						{
							for (auto smaInfo : it_lftWpExtLn->sma_info)
							{
								if (smaInfo.type == 5)
								{
									if (smaInfo.start_offset < ADC::common::k_BA_SmaNrrwLnOffsetThrd01)
									{
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
										this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
										DLOG(INFO) << "cond102" << std::endl;
										break;
									}
								}
							}
						}
					}
					else
					{
						if (isLCLastCycle == false)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
							DLOG(INFO) << "cond23" << std::endl;
						}
					}
				}
				//20220928:E
				if (isSplitTrgtRdTyp == true &&
					isDistLessThanThrd02 == true &&
					isLCLastCycle == false)
				{
					if (isRghtLnEmerg == false && isLftLnEmerg == false && isRghtLnExist == true && isLftLnExist == true)
					{
						if (rghtLnEndOffset < globalInfoProc.globalInfoContainer.m_DistToWayPoint && 
					    	lftLnEndOffset >= globalInfoProc.globalInfoContainer.m_DistToWayPoint)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
							DLOG(INFO) << "cond52" << std::endl;
						}
						else if (rghtLnEndOffset >= globalInfoProc.globalInfoContainer.m_DistToWayPoint &&
								 lftLnEndOffset < globalInfoProc.globalInfoContainer.m_DistToWayPoint)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
							DLOG(INFO) << "cond53" << std::endl;
						}
					}
				}				
			}
			// 20220722: if left direction indicator is on, then suppress right lane change and right lane change prepare
			// right direction indicator is on, then suppress left lane change and left lane change prepare
			if (this->input_BA.SM_DirectionIndicatorSts == BA::DircnIndSts_LeftOn && isLCLastCycle == false)
			{
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
				DLOG(INFO) << "cond14" << std::endl;
			}
			if (this->input_BA.SM_DirectionIndicatorSts == BA::DircnIndSts_RightOn && isLCLastCycle == false)
			{
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
				this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
				DLOG(INFO) << "cond15" << std::endl;
			}
			// added 20220805
			// get current line type and front nearest line type
			/* #enum:LINETYPE
			uint8 LINETYPE_UNKNOWN = 0
			uint8 LINETYPE_SOLID = 1
			uint8 LINETYPE_DASHED = 2
			uint8 LINETYPE_DOUBLESOLIDDASHED = 3
			uint8 LINETYPE_DOUBLEDASHEDSOLID = 4
			uint8 LINETYPE_DOUBLESOLIDSOLID = 5
			uint8 LINETYPE_DOUBLEDASHEDDASHED = 6
			uint8 LINETYPE_BOTTS = 7
			uint8 LINETYPE_DECELERATION = 8 */
			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   [egoLnID](const Lane &lane_)
                     		           {
                                		   return (egoLnID == lane_.start_id);
                              		   });
 			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
  			{
    			// get first lane marker id of right side
    			uint32_t currRghtLnMrkID = uint32_t(0);
    			if (!it_lane->right_lane_marker_id_list.empty())
    			{
      				currRghtLnMrkID = it_lane->right_lane_marker_id_list.front();
					uint8_t currRghtLnTyp = uint8_t(0); // unknown
					if (GetLineTyp(currRghtLnMrkID, currRghtLnTyp) == true)
					{
						if (currRghtLnTyp == 1 || currRghtLnTyp == 3 || currRghtLnTyp == 5)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
							DLOG(INFO) << "cond31" << std::endl;
						}
					}
    			}
    			else
    			{
					DLOG(INFO) << "current right lane marker is empty!" << std::endl;
    			}
				// get first lane marker id of left side
    			uint32_t currLftLnMrkID = uint32_t(0);
    			if (!it_lane->left_lane_marker_id_list.empty())
    			{
      				currLftLnMrkID = it_lane->left_lane_marker_id_list.front();
					uint8_t currLftLnTyp = uint8_t(0); // unknown
					if (GetLineTyp(currLftLnMrkID, currLftLnTyp) == true)
					{
						if (currLftLnTyp == 1 || currLftLnTyp == 4 || currLftLnTyp == 5)
						{
							this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
							DLOG(INFO) << "cond30" << std::endl;
						}
					}
    			}
    			else
    			{
					DLOG(INFO) << "current left lane marker is empty!" << std::endl;
    			}

				// added 20220930
				// if rm neighbor lane end range is less than ego lane's and ego lane is in raw target lane list
				// then suppress correspoinding lane change
				bool isEgoLnInRwTgtLn = false;
				auto it_egoLnRwTrgtLn = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), egoLnID);
				if (it_egoLnRwTrgtLn != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
				{
					isEgoLnInRwTgtLn = true;
				}
				float egoLaneEndRng = it_lane->end_range;
				float lftNgbrLaneEndRng = it_lane->end_range;
				float rghtNgbrLaneEndRng = it_lane->end_range; 
				uint8_t lftNgbrLnID = uint8_t(0);
				uint8_t rghtNgbrLnID = uint8_t(0);
				for (auto ngbrLn : it_lane->left_neighbor_lane)
				{
					if (ngbrLn.start_range < 0.0f)
					{
						lftNgbrLnID = ngbrLn.lane_id;
						lftNgbrLaneEndRng = ngbrLn.end_range;
						auto it_lftNgbrLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   					 [lftNgbrLnID](const Lane &lane_)
                     		          	 				 {
                                		  	 			    return (lftNgbrLnID == lane_.start_id);
                              		   					 });
						if (it_lftNgbrLn != this->input_BA.RM_RoadModel.lane_list.end())
						{
							lftNgbrLaneEndRng = it_lftNgbrLn->end_range;
						}
						break;
					}
				}
				if (lftNgbrLnID != uint8_t(0))
				{
					// 20221110: include condition that target road type is not ramp merge CA road
					if (lftNgbrLaneEndRng < egoLaneEndRng && isLCLastCycle == false && isEgoLnInRwTgtLn == true && globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_RmpMrgCARd)
					{
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
						DLOG(INFO) << "cond201" << std::endl;
					}
				}
				for (auto ngbrLn : it_lane->right_neighbor_lane)
				{
					if (ngbrLn.start_range < 0.0f)
					{
						rghtNgbrLnID = ngbrLn.lane_id;
						rghtNgbrLaneEndRng = ngbrLn.end_range;
						auto it_rghtNgbrLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   					  [rghtNgbrLnID](const Lane &lane_)
                     		          	 				  {
                                		  	 				  return (rghtNgbrLnID == lane_.start_id);
                              		   					  });
						if (it_rghtNgbrLn != this->input_BA.RM_RoadModel.lane_list.end())
						{
							rghtNgbrLaneEndRng = it_rghtNgbrLn->end_range;
						}
						break;
					}
				}
				if (rghtNgbrLnID != uint8_t(0))
				{
					// 20221010: include condition that target road type is not ramp merge CA road
					if (rghtNgbrLaneEndRng < egoLaneEndRng && isLCLastCycle == false && isEgoLnInRwTgtLn == true && globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_RmpMrgCARd)
					{
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChange);
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_RightLaneChangePrepare);
						DLOG(INFO) << "cond202" << std::endl;
					} 
				}
			}
			// added 20220909: type D
			if (isSplitTrgtRdTyp == true &&
				globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right &&
				globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.size() == 1 &&
				globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.size() > 1 && 
				globalInfoProc.globalInfoContainer.m_DistToWayPoint < 1500.0f &&
				isLCLastCycle == false)
			{
				uint8_t targetLnID = globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.front();
				WPExtendLaneStruct targetLnObj;
				if (globalInfoProc.FindWPExtendObj(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst, targetLnID, targetLnObj) == true)
				{
					if (targetLnObj.current_extendlane_offset > 0.0f)
					{
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChange);
						this->BA_SupprActLst.emplace_back(BA::BA_ActionTyp_LeftLaneChangePrepare);
						DLOG(INFO) << "cond40" << std::endl;
					}
				}
			}
			
			// added 20220618
			bool lnChgNumCond = ( globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum <= ADC::common::k_BA_MinAntiDrctLCNum );
			bool dist2WPCond = (   globalInfoProc.globalInfoContainer.m_DistToWayPoint >= ADC::common::k_BA_AntiDrctLCDist2WPThr
			                   && (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd || 
							       globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
								   globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd) );
			bool dist2LastLCPntCond = false;
			bool fllwObjTypCond = false;
			bool lnPosCond = false;
			// bool allwLCAntiDrctCond = false;
			uint32_t fllwObjID = this->input_BA.BA_OptTargetObjID_LastCycle;
			DLOG(INFO) << "OPT Target id = " << int(fllwObjID) << std::endl;
			auto it_FllwObj = std::find_if(this->input_BA.RM_RoadModel.object_list.begin(), this->input_BA.RM_RoadModel.object_list.end(),
										   [fllwObjID](const Object &obj_)
										   {
											   return (fllwObjID == obj_.object_id);
										   });
			if (it_FllwObj != this->input_BA.RM_RoadModel.object_list.end())
			{
				// object type:
				// 0=Unknown;
				// 1=Car;
				// 2=Truck;
 				// 3=Bus;
				// 4=Bicycle;
				// 5=Pedestrian;
				if (it_FllwObj->classification == 2)
				{
					fllwObjTypCond = true;
				}
			}
			if (lnChgNumCond == true && dist2WPCond == true /* && fllwObjTypCond == true */)
			{
				auto it_GlbInfoLn = std::find_if(globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.begin(), globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end(),
												 [egoLnID](const BA::GlobalLaneInfo &glbLnInfo_)
												 {
													 return (egoLnID == glbLnInfo_.m_LaneID);
												 });
				if (it_GlbInfoLn != globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end())
				{
					float antiDrctLCDist2LastLCPntThr = ADC::common::k_BA_AntiDrctLCDist2LastLCPntThr_Cnsr;
					if (this->input_BA.SM_DriveMode == BA::DriveMd_Sport)
					{
						antiDrctLCDist2LastLCPntThr = ADC::common::k_BA_AntiDrctLCDist2LastLCPntThr_Aggr;
					}
					if ((globalInfoProc.globalInfoContainer.m_DistToWayPoint - it_GlbInfoLn->m_LastLaneChangePoint.front()) > antiDrctLCDist2LastLCPntThr)
					{
						dist2LastLCPntCond = true;
					}
				}
				auto it_Lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
								            [egoLnID](const Lane &lane_)
											{
												return (egoLnID == lane_.start_id);
											});
				if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					// 20221111: replace lane change direction with route relative position
					if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right &&
					    /* globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right && */
						this->input_BA.SM_DirectionIndicatorSts != BA::DircnIndSts_RightOn)
					{
						if (!it_Lane->right_neighbor_lane.empty())
						{
							// get the neighbor lane with the smallest and negative start range
							auto it_minStRngLn = std::min_element(it_Lane->right_neighbor_lane.begin(), it_Lane->right_neighbor_lane.end(), Cmp_by_stRng);
							uint8_t rghtNeighborLnID = it_minStRngLn->lane_id;
							float rghtNeighborLnStRng = it_minStRngLn->start_range;
							if (rghtNeighborLnStRng < ADC::common::k_BA_MinNgbrLnStRng)
							{
								// check whether right neighbor lane is emergency lane  
								auto it_rghtNghbLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
							    	        					  [rghtNeighborLnID](const Lane &lane_)
											                  	  {
																	  return (rghtNeighborLnID == lane_.start_id);
																  });
								if (it_rghtNghbLn != this->input_BA.RM_RoadModel.lane_list.end())
								{
									if (it_rghtNghbLn->lane_type == 5 || it_rghtNghbLn->lane_type == 21)
									{
										lnPosCond = true;
									}	
								}
							}
						}
						else
						{
							lnPosCond = true;
						}
					}
					else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left && 
					         /* globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left && */ 
							 this->input_BA.SM_DirectionIndicatorSts != BA::DircnIndSts_LeftOn)
					{
						if (!it_Lane->left_neighbor_lane.empty())
						{
							// get the neighbor lane with the smallest and negative start range
							auto it_minStRngLn = std::min_element(it_Lane->left_neighbor_lane.begin(), it_Lane->left_neighbor_lane.end(), Cmp_by_stRng);
							uint8_t leftNeighborLnID = it_minStRngLn->lane_id;
							float lftNeighborLnStRng = it_minStRngLn->start_range;
							if (lftNeighborLnStRng < ADC::common::k_BA_MinNgbrLnStRng)
							{
								// check whether left neighbor lane is emergency lane  
								auto it_lftNghbLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
							    	        					[leftNeighborLnID](const Lane &lane_)
											                  	{
																   return (leftNeighborLnID == lane_.start_id);
																});
								if (it_lftNghbLn != this->input_BA.RM_RoadModel.lane_list.end())
								{
									if (it_lftNghbLn->lane_type == 5 || it_lftNghbLn->lane_type == 21)
									{
										lnPosCond = true;
									}	
								}
							}
						}
						else
						{
							lnPosCond = true;
						}
					}		
				}
				if (lnPosCond == true && dist2LastLCPntCond == true)
				{
					allwLCAntiDrctCond = true;
					DLOG(INFO) << "cond20" << std::endl;
				}
			}
			DLOG(INFO) << "lnChgNumCond = " << int(lnChgNumCond) << std::endl;
			DLOG(INFO) << "dist2WPCond = " << int(dist2WPCond) << std::endl;
			DLOG(INFO) << "lnPosCond = " << int(lnPosCond) << std::endl;
			DLOG(INFO) << "fllwObjTypCond = " << int(fllwObjTypCond) << std::endl;
			DLOG(INFO) << "dist2LastLCPntCond = " << int(dist2LastLCPntCond) << std::endl;
			DLOG(INFO) << "allwLCAntiDrctCond = " << int(allwLCAntiDrctCond) << std::endl;

			// delete duplicated action
			std::sort(this->BA_SupprActLst.begin(), this->BA_SupprActLst.end());
			this->BA_SupprActLst.erase(unique(this->BA_SupprActLst.begin(), this->BA_SupprActLst.end()), this->BA_SupprActLst.end());
			// delete permitted action in specific case
			if (allwLCAntiDrctCond == true)
			{
				auto it_SpprAct = this->BA_SupprActLst.begin();
				for (; it_SpprAct != this->BA_SupprActLst.end(); )
				{
					if (/* globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right */
						globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
					{
						if (*it_SpprAct == BA::BA_ActionTyp_LeftLaneChange || *it_SpprAct == BA::BA_ActionTyp_LeftLaneChangePrepare)
						{
							it_SpprAct = this->BA_SupprActLst.erase(it_SpprAct);
						}
						else
						{
							it_SpprAct++;
						}
					}
					else if (/* globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left */
							globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
					{
						if (*it_SpprAct == BA::BA_ActionTyp_RightLaneChange || *it_SpprAct == BA::BA_ActionTyp_RightLaneChangePrepare)
						{
							it_SpprAct = this->BA_SupprActLst.erase(it_SpprAct);
						}
						else
						{
							it_SpprAct++;
						}
					}
					else
					{
						it_SpprAct++;
					}
				}
			}	
			return true;
		}

		bool BehaviorAnalysis::Sort_by_actionTyp(BA::behaviorNodeGlobal &bng01, BA::behaviorNodeGlobal &bng02)
		{
			return (int(bng01.firstActionTyp()) < int(bng02.firstActionTyp()));
		}

		void BehaviorAnalysis::CalcLnVehFlowSpd(const uint8_t &laneID,
												const RoadInfo &roadModel,
												const float &setSpd,
												float &lnVehFlowSpd,
												uint32_t &frtNrstObjID,
												float &frtNrstObjDst)
		{
			// only take vehicles in front of ego vehicle into consideration
			// the distance btw ego vehicle and other vehicles have to be within 200 m, or the accuracy is not high enough
			// the vehicle flow speed is the average of sampling speed in last 500ms tbd
			// vehicle flow speed is weighted by reciprocal of distance
			// if there is no vehicle in front of ego vehicle, then set vehicle flow speed to driver set speed
			// And vehicle flow speed should not be greater than driver set speed
			// float curVehFlowSpd = 0;

			auto it_lane = std::find_if(roadModel.lane_list.begin(), roadModel.lane_list.end(),
										[laneID](const Lane &lane_)
										{
											return (laneID == lane_.start_id);
										});
			if (it_lane != roadModel.lane_list.end())
			{
				BA::FeatureTag LineFollowTag = BA::FeatureTag_LineFllw;
				std::vector<ObjectLocation> tmpObjLocLst;
				auto it_corridor = std::find_if(it_lane->corridor_list.begin(), it_lane->corridor_list.end(),
												[LineFollowTag](const Corridor &corridor_)
												{
													return (LineFollowTag == corridor_.feature_tag);
												});
				if (it_corridor != it_lane->corridor_list.end())
				{
					for (auto objLoc_ : it_corridor->object_in_corridor_list)
					{
						if (objLoc_.relative_center_x > 0 && objLoc_.relative_center_x <= ADC::common::k_BA_VehFlowCalcMaxRng)
						{
							tmpObjLocLst.emplace_back(objLoc_);
							if (objLoc_.relative_center_x < frtNrstObjDst)
							{
								frtNrstObjDst = objLoc_.relative_center_x;
								frtNrstObjID = objLoc_.object_id;
							}
						}
					}
				}
				if (!tmpObjLocLst.empty())
				{
					float numerator = 0;
					float denominator = 0;
					for (auto objLoc : tmpObjLocLst)
					{
						if (std::abs(objLoc.relative_center_x) < 0.1)
						{
							objLoc.relative_center_x = 0.1;
						}
						numerator = numerator + 1.0 / std::abs(objLoc.relative_center_x) * std::abs(objLoc.reference_point2corridor_frenet.ds);
						denominator = denominator + 1.0 / std::abs(objLoc.relative_center_x);
					}
					lnVehFlowSpd = numerator / denominator;
					if (lnVehFlowSpd > setSpd)
					{
						lnVehFlowSpd = setSpd;
					}
				}
				else
				{
					lnVehFlowSpd = setSpd;
				}
			}
			else
			{
				// the lane is not found in roadModel
				lnVehFlowSpd = 0;
			}
		}

		void BehaviorAnalysis::CalcLnVehFlowSpd_alpha(const uint8_t &laneID,
													  const RoadInfo &roadModel,
													  const float &setSpd,
													  float &lnVehFlowSpd,
													  uint32_t &frtNrstObjID,
													  float &frtNrstObjDst)
		{
			// only take vehicles in front of ego vehicle into consideration
			// the distance btw ego vehicle and other vehicles have to be within 200 m, or the accuracy is not high enough
			// the vehicle flow speed is the average of sampling speed in last 500ms tbd
			// vehicle flow speed is weighted by reciprocal of distance
			// if there is no vehicle in front of ego vehicle, then set vehicle flow speed to driver set speed
			// And vehicle flow speed should not be greater than driver set speed
			// float curVehFlowSpd = 0;
			// alpha version
			// if the err btw front nearest vehicle speed and weighted speed of the rest vehicles is greater than 5 m/s
			// then vehicle flow speed is calculated without the front nearest vehicle 
			auto it_lane = std::find_if(roadModel.lane_list.begin(), roadModel.lane_list.end(),
										[laneID](const Lane &lane_)
										{
											return (laneID == lane_.start_id);
										});
			if (it_lane != roadModel.lane_list.end())
			{
				BA::FeatureTag LineFollowTag = BA::FeatureTag_LineFllw;
				std::vector<ObjectLocation> tmpObjLocLst;
				auto it_corridor = std::find_if(it_lane->corridor_list.begin(), it_lane->corridor_list.end(),
												[LineFollowTag](const Corridor &corridor_)
												{
													return (LineFollowTag == corridor_.feature_tag);
												});
				float frtNrstObjSpd = 0.0f;
				if (it_corridor != it_lane->corridor_list.end())
				{
					for (auto objLoc_ : it_corridor->object_in_corridor_list)
					{
						if (objLoc_.relative_center_x > 0 && objLoc_.relative_center_x <= ADC::common::k_BA_VehFlowCalcMaxRng)
						{
							tmpObjLocLst.emplace_back(objLoc_);
							if (objLoc_.relative_center_x < frtNrstObjDst)
							{
								frtNrstObjDst = objLoc_.relative_center_x;
								frtNrstObjID = objLoc_.object_id;
								frtNrstObjSpd = objLoc_.reference_point2corridor_frenet.ds;
							}
						}
					}
				}
				if (!tmpObjLocLst.empty())
				{
					float numerator = 0.0f;
					float denominator = 0.0f;
					float lnVehFlowSpd_ALL = 0.0f;
					float lnVehFlowSpd_ExclFrtNrtObj = 0.0f;
					if (tmpObjLocLst.size() > 1)
					{
						// all obj included
						for (auto objLoc : tmpObjLocLst)
						{
							if (std::abs(objLoc.relative_center_x) < 0.1)
							{
								objLoc.relative_center_x = 0.1;
							}
							numerator = numerator + 1.0 / std::abs(objLoc.relative_center_x) * std::abs(objLoc.reference_point2corridor_frenet.ds);
							denominator = denominator + 1.0 / std::abs(objLoc.relative_center_x);
						}
						lnVehFlowSpd_ALL = numerator / denominator;
						if (lnVehFlowSpd_ALL > setSpd)
						{
							lnVehFlowSpd_ALL = setSpd;
						}
						// front nearest obj excluded
						numerator = 0.0f;
					    denominator = 0.0f;
						for (auto objLoc : tmpObjLocLst)
						{
							if (objLoc.object_id == frtNrstObjID)
							{
								continue;
							}
							if (std::abs(objLoc.relative_center_x) < 0.1)
							{
								objLoc.relative_center_x = 0.1;
							}
							numerator = numerator + 1.0 / std::abs(objLoc.relative_center_x) * std::abs(objLoc.reference_point2corridor_frenet.ds);
							denominator = denominator + 1.0 / std::abs(objLoc.relative_center_x);
						}
						lnVehFlowSpd_ExclFrtNrtObj = numerator / denominator;
						if (lnVehFlowSpd_ExclFrtNrtObj > setSpd)
						{
							lnVehFlowSpd_ExclFrtNrtObj = setSpd;
						}
						if ((frtNrstObjSpd - lnVehFlowSpd_ExclFrtNrtObj) >= ADC::common::k_BA_SpdGap4AlphaVehFlowCalc)
						{
							lnVehFlowSpd = lnVehFlowSpd_ExclFrtNrtObj;
							DLOG(INFO) << "front nearest obj is excluded!" << std::endl;
							DLOG(INFO) << "frtNrstObjSpd = " << frtNrstObjSpd << std::endl;
							DLOG(INFO) << "lnVehFlowSpd_ExclFrtNrtObj = " << lnVehFlowSpd_ExclFrtNrtObj << std::endl;
						}
						else
						{
							lnVehFlowSpd = lnVehFlowSpd_ALL;
						}
					}
					else
					{
						lnVehFlowSpd = frtNrstObjSpd;
						if (lnVehFlowSpd > setSpd)
						{
							lnVehFlowSpd = setSpd;
						}
					}
				}
				else
				{
					lnVehFlowSpd = setSpd;
				}
			}
			else
			{
				// the lane is not found in roadModel
				lnVehFlowSpd = 0;
			}
		}

		void BehaviorAnalysis::IsLCIntGnrt(const float &vehSpd,
										   const float &setSpd,
										   const CarPosition &carPos,
										   const bool &isLCIntReset,
										   bool &isEffcLCIntGnrt,
										   const BA::DriveMd &drvStyle)
		{
			// if in last x0 sec, vehicle speed is always less than 90% of set speed, then set flag to true
			// if in last x1 sec, vehicle speed is always less than 80% of set speed, then set flag to true
			// if in last x2 sec, vehicle speed is always less than 70% of set speed, then set flag to true
			// if in last x3 sec, vehicle speed is always less than 60% of set speed, then set flag to true
			bool unstsfyDegreeFlg01 = false;
			bool unstsfyDegreeFlg02 = false;
			bool unstsfyDegreeFlg03 = false;
			bool unstsfyDegreeFlg04 = false;
			static int counter01 = 0;
			static int counter02 = 0;
			static int counter03 = 0;
			static int counter04 = 0;
			DLOG(INFO) << "counter01 = " << counter01 << std::endl;
			DLOG(INFO) << "counter02 = " << counter02 << std::endl;
			DLOG(INFO) << "counter03 = " << counter03 << std::endl;
			DLOG(INFO) << "counter04 = " << counter04 << std::endl;
			static uint8_t laneIDinLastCycle = 0;

			std::vector<float> unstsfyDgrDurTmLst;
			switch (drvStyle)
			{
			case BA::DriveMd_Comfortable:
				unstsfyDgrDurTmLst = ADC::common::k_BA_UnstsfyDgrDurTmLst01;
				break;
			case BA::DriveMd_Normal:
				unstsfyDgrDurTmLst = ADC::common::k_BA_UnstsfyDgrDurTmLst02;
				break;
			case BA::DriveMd_Sport:
				unstsfyDgrDurTmLst = ADC::common::k_BA_UnstsfyDgrDurTmLst03;
				break;
			default:
				unstsfyDgrDurTmLst = ADC::common::k_BA_UnstsfyDgrDurTmLst02;
			}

			enum UNSTSFYDGRFLGState
			{
				eUDFS_Unsatisfied = 0,
				eUDFS_Satisfied = 1
			};
			static UNSTSFYDGRFLGState flg01Sts = eUDFS_Unsatisfied;
			static UNSTSFYDGRFLGState flg02Sts = eUDFS_Unsatisfied;
			static UNSTSFYDGRFLGState flg03Sts = eUDFS_Unsatisfied;
			static UNSTSFYDGRFLGState flg04Sts = eUDFS_Unsatisfied;
			static int flg01CntPos = 0;
			static int flg01CntNeg = 0;
			static int flg02CntPos = 0;
			static int flg02CntNeg = 0;
			static int flg03CntPos = 0;
			static int flg03CntNeg = 0;
			static int flg04CntPos = 0;
			static int flg04CntNeg = 0;

			bool isLaneIDChg = false;
			if (laneIDinLastCycle != carPos.ego_lane_id)
			{
				isLaneIDChg = true;
			}
			// DLOG(INFO) << "isLaneIDChg = " << int(isLaneIDChg) << std::endl;
			// DLOG(INFO) << "setSpd = " << setSpd << std::endl;

			if (isLCIntReset == false)
			{
				if (isLaneIDChg == false)
				{
					if (vehSpd <= setSpd * ADC::common::k_BA_UnstsfyDgrPercLstActv.at(0))
					{
						counter01 = std::min(counter01 + 1, 10000);
					}
					else if (vehSpd > setSpd * ADC::common::k_BA_UnstsfyDgrPercLstInactv.at(0))
					{
						counter01 = 0;
						// DLOG(INFO) << "counter01 enter" << std::endl;
					}

					if (vehSpd <= setSpd * ADC::common::k_BA_UnstsfyDgrPercLstActv.at(1))
					{
						counter02 = std::min(counter02 + 1, 10000);
					}
					else if (vehSpd > setSpd * ADC::common::k_BA_UnstsfyDgrPercLstInactv.at(1))
					{
						counter02 = 0;
						// DLOG(INFO) << "counter02 enter" << std::endl;
					}

					if (vehSpd <= setSpd * ADC::common::k_BA_UnstsfyDgrPercLstActv.at(2))
					{
						counter03 = std::min(counter03 + 1, 10000);
					}
					else if (vehSpd > setSpd * ADC::common::k_BA_UnstsfyDgrPercLstInactv.at(2))
					{
						counter03 = 0;
						// DLOG(INFO) << "counter03 enter" << std::endl;
					}

					if (vehSpd <= setSpd * ADC::common::k_BA_UnstsfyDgrPercLstActv.at(3))
					{
						counter04 = std::min(counter04 + 1, 10000);
					}
					else if (vehSpd > setSpd * ADC::common::k_BA_UnstsfyDgrPercLstInactv.at(3))
					{
						counter04 = 0;
						// DLOG(INFO) << "counter04 enter" << std::endl;
					}
				}
				else
				{
					// if lane ID changed, then reset counters
					counter01 = 0;
					counter02 = 0;
					counter03 = 0;
					counter04 = 0;
					// DLOG(INFO) << "lane id reset enter" << std::endl;
				}
			}
			else
			{
				// if isResetFlg from outside is true, then reset counters
				counter01 = 0;
				counter02 = 0;
				counter03 = 0;
				counter04 = 0;
				// DLOG(INFO) << "reset outside enter" << std::endl;
			}

			if (counter01 >= unstsfyDgrDurTmLst.at(0) * 1000 / BEHAVIORANALYZERCYCLE)
			{
				unstsfyDegreeFlg01 = true;
			}
			else
			{
				unstsfyDegreeFlg01 = false;
			}
			DLOG(INFO) << "unstsfyDegreeFlg01 = " << int(unstsfyDegreeFlg01) << std::endl;
			if (counter02 >= unstsfyDgrDurTmLst.at(1) * 1000 / BEHAVIORANALYZERCYCLE)
			{
				unstsfyDegreeFlg02 = true;
			}
			else
			{
				unstsfyDegreeFlg02 = false;
			}
			DLOG(INFO) << "unstsfyDegreeFlg02 = " << int(unstsfyDegreeFlg02) << std::endl;
			if (counter03 >= unstsfyDgrDurTmLst.at(2) * 1000 / BEHAVIORANALYZERCYCLE)
			{
				unstsfyDegreeFlg03 = true;
			}
			else
			{
				unstsfyDegreeFlg03 = false;
			}
			if (counter04 >= unstsfyDgrDurTmLst.at(3) * 1000 / BEHAVIORANALYZERCYCLE)
			{
				unstsfyDegreeFlg04 = true;
			}
			else
			{
				unstsfyDegreeFlg04 = false;
			}

			// unsatisfaction degree flag debounce process
			switch (flg01Sts)
			{
			case eUDFS_Unsatisfied:
				if (unstsfyDegreeFlg01 == true)
				{
					flg01CntPos++;
				}
				else
				{
					flg01CntPos = 0;
				}

				if (flg01CntPos > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg01Sts = eUDFS_Satisfied;
					flg01CntPos = 0;
				}
				break;

			case eUDFS_Satisfied:
				if (unstsfyDegreeFlg01 == false)
				{
					flg01CntNeg++;
				}
				else
				{
					flg01CntNeg = 0;
				}

				if (flg01CntNeg > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg01Sts = eUDFS_Unsatisfied;
				}
				break;
			}
			DLOG(INFO) << "flg01Sts = " << int(flg01Sts) << std::endl;

			switch (flg02Sts)
			{
			case eUDFS_Unsatisfied:
				if (unstsfyDegreeFlg02 == true)
				{
					flg02CntPos++;
				}
				else
				{
					flg02CntPos = 0;
				}

				if (flg02CntPos > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg02Sts = eUDFS_Satisfied;
					flg02CntPos = 0;
				}
				break;

			case eUDFS_Satisfied:
				if (unstsfyDegreeFlg02 == false)
				{
					flg02CntNeg++;
				}
				else
				{
					flg02CntNeg = 0;
				}

				if (flg02CntNeg > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg02Sts = eUDFS_Unsatisfied;
				}
				break;
			}
			DLOG(INFO) << "flg02Sts = " << int(flg02Sts) << std::endl;

			switch (flg03Sts)
			{
			case eUDFS_Unsatisfied:
				if (unstsfyDegreeFlg03 == true)
				{
					flg03CntPos++;
				}
				else
				{
					flg03CntPos = 0;
				}

				if (flg03CntPos > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg03Sts = eUDFS_Satisfied;
					flg03CntPos = 0;
				}
				break;

			case eUDFS_Satisfied:
				if (unstsfyDegreeFlg03 == false)
				{
					flg03CntNeg++;
				}
				else
				{
					flg03CntNeg = 0;
				}

				if (flg03CntNeg > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg03Sts = eUDFS_Unsatisfied;
				}
				break;
			}

			switch (flg04Sts)
			{
			case eUDFS_Unsatisfied:
				if (unstsfyDegreeFlg04 == true)
				{
					flg04CntPos++;
				}
				else
				{
					flg04CntPos = 0;
				}

				if (flg04CntPos > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg04Sts = eUDFS_Satisfied;
					flg04CntPos = 0;
				}
				break;

			case eUDFS_Satisfied:
				if (unstsfyDegreeFlg04 == false)
				{
					flg04CntNeg++;
				}
				else
				{
					flg04CntNeg = 0;
				}

				if (flg04CntNeg > ADC::common::k_BA_UnstsfDgrFlgCntThrd / BEHAVIORANALYZERCYCLE)
				{
					flg04Sts = eUDFS_Unsatisfied;
				}
				break;
			}
			// determine whether to set isEffcLCIntGnrt to true
			if (flg01Sts == eUDFS_Satisfied ||
				flg02Sts == eUDFS_Satisfied ||
				flg03Sts == eUDFS_Satisfied ||
				flg04Sts == eUDFS_Satisfied)
			{
				isEffcLCIntGnrt = true;
			}

			laneIDinLastCycle = carPos.ego_lane_id;
		}

		bool BehaviorAnalysis::DebounceOptRes()
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 200ms/50ms, accumulate the internal count value (if previous corridor ID and target ID can not been found, switch to current inputs)
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			//
			// output the previous input values inside
			static int internCntVal = 0;
			static int internCntVal_trgtid = 0;
			static int internCntVal_crrdid = 0;
			static BA::BA_ActionTyp actionPrev = BA::BA_ActionTyp_None;
			static uint32_t corridorIDPrev = uint32_t(0);
			static uint32_t targetIDPrev = uint32_t(0);
			static BA::Index passageIDPrev{0, 0};
			if (actionPrev != this->BA_OptimalAction)
			{
				if (internCntVal < ADC::common::k_BA_OptResDebCnt)
				{
					internCntVal++;
				}
				else
				{
					internCntVal = 0;
					actionPrev = this->BA_OptimalAction;
					// corridorIDPrev = this->BA_OptimalCorridorID;
					// targetIDPrev = this->BA_OptimalTargetObjID;
					passageIDPrev = this->BA_OptimalPassageID;
				}
			}
			else
			{
				internCntVal = 0;
			}
			// 20221109:
			// debounce target id specifically
			if (targetIDPrev != this->BA_OptimalTargetObjID)
			{
				if (internCntVal_trgtid < ADC::common::k_BA_OptResDebCnt)
				{
					internCntVal_trgtid++;
				}
				else
				{
					internCntVal_trgtid = 0;
					targetIDPrev = this->BA_OptimalTargetObjID;
				}
			}
			else
			{
				internCntVal_trgtid = 0;
			}
			if (corridorIDPrev != this->BA_OptimalCorridorID)
			{
				if (internCntVal_crrdid < ADC::common::k_BA_OptResDebCnt)
				{
					internCntVal_crrdid++;
				}
				else
				{
					internCntVal_crrdid = 0;
					corridorIDPrev = this->BA_OptimalCorridorID;
				}
			}
			else
			{
				internCntVal_crrdid = 0;
			}
			DLOG(INFO) << "internCntVal = " << internCntVal << std::endl;
			DLOG(INFO) << "actionPrev = " << int(actionPrev) << std::endl;

			// if BA_FuncOnDurTm is less than or equal to 0.5 sec
			// then set corridorIDPrev to ego lane corridor id and actionPrev to lanefollow
			if (this->BA_FuncOnDurTm <= 0.5f)
			{
				uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
				auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
					                            [egoLnID](const Lane &lane_)
												{
													return (egoLnID == lane_.start_id);
												});
				if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					for (auto crrd_ : it_lane->corridor_list)
					{
						if (crrd_.feature_tag == BA::FeatureTag_LineFllw)
						{
							corridorIDPrev = crrd_.corridor_id;
							actionPrev = BA::BA_ActionTyp_LaneFollow;
							break;
						}
					}
				}
			}
			this->BA_OptAction_Deb = actionPrev;
			this->BA_OptCorridorID_Deb = corridorIDPrev;
			this->BA_OptTargetObjID_Deb = targetIDPrev;
			this->BA_OptPassageID_Deb = passageIDPrev;
			// if debounced corridor ID AND target ID can not been found, then set them as current inputs
			bool isDebCrrdFound = false;
			bool isDebTrgtFound = false;
			for (auto lane_ : this->input_BA.RM_RoadModel.lane_list)
			{
				for (auto corridor_ : lane_.corridor_list)
				{
					if (corridor_.corridor_id == this->BA_OptCorridorID_Deb)
					{
						isDebCrrdFound = true;
						break;
					}
				}
			}
			for (auto object_ : this->input_BA.RM_RoadModel.object_list)
			{
				if (object_.object_id == this->BA_OptTargetObjID_Deb)
				{
					isDebTrgtFound = true;
					break;
				}
			}
			if ((isDebTrgtFound == false && this->BA_OptTargetObjID_Deb != 0) || isDebCrrdFound == false || ADC::common::k_BA_OptResDebProcFlg == false)
			{
				// this->BA_OptAction_Deb = this->BA_OptimalAction; // 02000503: optimal action has to be debounced anyway
				this->BA_OptCorridorID_Deb = this->BA_OptimalCorridorID;
				this->BA_OptTargetObjID_Deb = this->BA_OptimalTargetObjID;
				this->BA_OptPassageID_Deb = this->BA_OptimalPassageID;
			}
			return true;
		}

		bool BehaviorAnalysis::ManageDircnIndReq(GlobalInfoProcess &globalInfoProc, const bool &drcnTurnOff, const BA::TargetRoadTyp &lastTrtRdTyp)
		{
			// if any direction indicator has been on, then there should be no direction indicator request
			// if SM_LCAvlBhvr is not equal to LCAvlblBhvr_LaneChange_Auto
			// 1- if BA_OptAction_Deb is equal to Left Lane Change or Right Lane Change, then request direction indicator on
			//    and request direction indicator off when LC finished or BA_OptAction is Lane Change Abolition
			// 2- if SM_LCAvlBhvr equals to LCAvlblBhvr_LaneChange_Auto, and ego car is on the target lane and has been close enough to the start point of deceleration lane, then request
			//    direction indicator on based on relative direction of navigation route. And when any of following conditions are satisfied, request direction indicator off:
			//    2.1 direction indicator request has continued for 5 sec and ego car has been on the lane center;
			//    2.1 BA_OptAction_Deb changed to other action types instead of Lane Follow
			//    20220425: condition changed to that ego car is on the raw target lane list, instead retricted to target lane list
			// 3- if BA_OptAction_Deb is equal to Left Lane Change Prepare or Right Lane Change Prepare, then also request indicator on 20220419
			// 4- else if vehicle is not on the relevant corner lane, and the dist btw the vehicle and last lane change point is less than 300m, then light the direction indicator until miss the last lane change point 20220420
			//         how to judge if vehicle is not on the relevant corner lane:  if the neighbor lane with negative start range is not an emergency lane, then yes 
			static BA::DircnIndReqSts internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
			static float internDrcnIndReqDurTm = 0;
			// request type
			// 0: not available
			// 1: by lane change
			// 2: by line follow
			// 3: by lane change prepare
			// 4: by lane change remind close to ramp
			// 5: by lane change remind close to merge road
			// 6: by driving remind close to split wp
			static int internDrcnIndReqTyp = 0;
			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			bool isEgoCarOnTrgtLn = (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == 0);
			bool isEgoCarOnRawTrgtLn = false;
			auto it_egoLn = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), this->input_BA.RM_RoadModel.car_position.ego_lane_id);
			if (it_egoLn != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
			{
				isEgoCarOnRawTrgtLn = true;
			}
			bool isSplitTrgtRdTyp = (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd);
			bool isEgoCarClsToDecelStPnt = false;
			bool isEgoCarClsToSpltPnt = false;
			BA::TransitionTyp transitLnTyp_B = BA::TransitionTyp_NotAvlbl; // 20220711 added
			BA::TransitionTyp transitLnTyp_C = BA::TransitionTyp_NotAvlbl; // 20220809 added

			// 20220825: add isEgoCarClsToSpltPnt and isDrcnIndFrcToOnand limit condition
			// if neighbor lane on another side of relative route direction is not in the raw target lane list is empty
			// or lane on the same side of relative route direction is empty or is emergency lane
			// then set isEgoCarClsToSpltPnt to false
			bool extraCond4ClsToSpltPnt = false;
			auto iter_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
										  [egoLnID](const WPExtendLaneStruct &wp_ln)
										  {
										     return (egoLnID == wp_ln.lane_id);
										  });
			if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
			{
				// get left neighbor lane
				if (iter_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					uint8_t lftLaneID = uint8_t(0);
					if (!iter_egoLn->left_extend_lane.empty())
					{
						for (auto lftExtLn : iter_egoLn->left_extend_lane)
						{
							if (lftExtLn.adjacent_extendlane_offset < 0.0f)
							{
								lftLaneID = lftExtLn.adjacent_extendlane_id;
								break;
							}
						}
						if (lftLaneID != uint8_t(0))
						{
							if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_NotAvlbl)
							{
								auto isLftExtLnInRwTgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), lftLaneID);
								if (isLftExtLnInRwTgtLnLst != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
							else
							{
								auto isLftExtLnInRwTgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.end(), lftLaneID);
								if (isLftExtLnInRwTgtLnLst != globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.end())
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
						}
						else
						{
							extraCond4ClsToSpltPnt = true;
						}
					}
					else
					{
						extraCond4ClsToSpltPnt = true;
					}
				}
				// get right neighbor lane
				if (iter_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					uint8_t rightLaneID = uint8_t(0);
					if (!iter_egoLn->right_extend_lane.empty())
					{
						for (auto rghtExtLn : iter_egoLn->right_extend_lane)
						{
							if (rghtExtLn.adjacent_extendlane_offset <= 0.0f)
							{
								rightLaneID = rghtExtLn.adjacent_extendlane_id;
								break;
							}
						}
						if (rightLaneID != uint8_t(0))
						{
							// whether it is emergency lane or not
							auto it_rightLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
														   [rightLaneID](const WPExtendLaneStruct &wpLn_)
														   {
															   return (rightLaneID == wpLn_.lane_id);
														   });
							if (it_rightLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
							{
								if (it_rightLn->is_emergency_lane == true)
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
						}
						else
						{
							extraCond4ClsToSpltPnt = true;
						}
					}
					else
					{
						extraCond4ClsToSpltPnt = true;
					}
				}
			}
			else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
			{
				// get right neighbor lane
				if (iter_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					uint8_t rghtLaneID = uint8_t(0);
					if (!iter_egoLn->right_extend_lane.empty())
					{
						for (auto rghtExtLn : iter_egoLn->right_extend_lane)
						{
							if (rghtExtLn.adjacent_extendlane_offset <= 0.0f)
							{
								rghtLaneID = rghtExtLn.adjacent_extendlane_id;
								break;
							}
						}
						if (rghtLaneID != uint8_t(0))
						{
							if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_NotAvlbl)
							{
								auto isRghtExtLnInRwTgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), rghtLaneID);
								if (isRghtExtLnInRwTgtLnLst != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
							else
							{
								auto isRghtExtLnInRwTgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.end(), rghtLaneID);
								if (isRghtExtLnInRwTgtLnLst != globalInfoProc.globalInfoContainer.m_TgtLaneLst_RWP.end())
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
						}
						else
						{
							extraCond4ClsToSpltPnt = true;
						}
					}
					else
					{
						extraCond4ClsToSpltPnt = true;
					}
				}
				// get left neighbor lane
				if (iter_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					uint8_t leftLaneID = uint8_t(0);
					if (!iter_egoLn->left_extend_lane.empty())
					{
						for (auto lftExtLn : iter_egoLn->left_extend_lane)
						{
							if (lftExtLn.adjacent_extendlane_offset <= 0.0f)
							{
								leftLaneID = lftExtLn.adjacent_extendlane_id;
								break;
							}
						}
						if (leftLaneID != uint8_t(0))
						{
							// whether it is emergency lane or not
							auto it_leftLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
														  [leftLaneID](const WPExtendLaneStruct &wpLn_)
														  {
														      return (leftLaneID == wpLn_.lane_id);
														  });
							if (it_leftLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
							{
								if (it_leftLn->is_emergency_lane == true)
								{
									extraCond4ClsToSpltPnt = true;
								}
							}
						}
						else
						{
							extraCond4ClsToSpltPnt = true;
						}
					}
					else
					{
						extraCond4ClsToSpltPnt = true;
					}
				}
			}
			DLOG(INFO) << "extraCond4ClsToSpltPnt = " << int(extraCond4ClsToSpltPnt) << std::endl;

			if (isEgoCarOnRawTrgtLn == true && isSplitTrgtRdTyp == true)
			{
				// based on route relative position, get adjacent extend lane list
				// the lane which is not a target lane and start offset is greater than 0 will be considered as the deceleration lane
				// if the dist btw ego car and start point of the deceleration lane belongs to (0, 40], then set isEgoCarClsToDecelStPnt to true
				// 20220601: get dist btw ego car and start point of the deceleration lane by transit lane list
				float distBtwEgoCar2DecelLnOffset_B = 1000.0f;
				float distBtwEgoCar2DecelLnOffset_C = 1000.0f;

				float distBtwEgoCar2DecelLnOffset_A = 1000.0f;
				WPExtendLaneStruct tgtLnStruct;
				WPExtendLaneStruct adjLnStruct;
				if (globalInfoProc.FindWPExtendObj(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst, this->input_BA.RM_RoadModel.car_position.ego_lane_id, tgtLnStruct) == true)
				{
					std::vector<AdjacentExtendLaneStruct> adjExtLaneLst;
					if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
					{
						adjExtLaneLst = tgtLnStruct.left_extend_lane;
					}
					else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
					{
						adjExtLaneLst = tgtLnStruct.right_extend_lane;
					}
					if (!adjExtLaneLst.empty())
					{
						DLOG(INFO) << "adjExtLaneLst is not empty" << std::endl;
						for (auto adjExtLane_ : adjExtLaneLst)
						{
							auto it_adjExtLn = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), adjExtLane_.adjacent_extendlane_id);
							if (it_adjExtLn == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
							{
								// not belong to target lane list
								bool isAdjLnNotDrvbl = false;
								if (globalInfoProc.FindWPExtendObj(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst, adjExtLane_.adjacent_extendlane_id, adjLnStruct) == true)
								{
									if (adjLnStruct.is_emergency_lane == uint8_t(1))
									{
										isAdjLnNotDrvbl = true;
									}
								}
								if (adjExtLane_.adjacent_extendlane_offset > 0 && isAdjLnNotDrvbl == false)
								{
									distBtwEgoCar2DecelLnOffset_A = adjExtLane_.adjacent_extendlane_offset;
									break;
								}
							}
						}
					}
				}
                DLOG(INFO) << "distBtwEgoCar2DecelLnOffset_A = " << distBtwEgoCar2DecelLnOffset_A << std::endl;
				// 20220904:
				// if current lane number is 1
				// assign A to B
				int totalLneNum = 0;
    			for (auto extendLane : globalInfoProc.globalInfoContainer.m_WpExtendLaneLst)
    			{
        			if (extendLane.is_emergency_lane == uint8_t(0) && extendLane.current_extendlane_offset < 1.0f)
        			{
            			totalLneNum++;
        			}
    			}
				if (totalLneNum == 1 && 
					globalInfoProc.globalInfoContainer.m_DistToWayPoint <= ADC::common::k_BA_DrcnIndAheadOnDistThrd)
				{
					distBtwEgoCar2DecelLnOffset_B = distBtwEgoCar2DecelLnOffset_A;
					if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
					{
						transitLnTyp_B = BA::TransitionTyp_LaneAddRght;
					}
					else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
					{
						transitLnTyp_B = BA::TransitionTyp_LaneAddLft;
					}
					DLOG(INFO) << "assign A TO B" << std::endl;
				}
				else
				{
					for (auto transitLn : this->input_BA.EHR_EHRInfo.transit_lanes_list)
					{
						if (transitLn.transit_lane_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id \
					 		&& transitLn.transit_connected_lane_id == uint8_t(0) \
					 		&& (transitLn.transit_type == BA::TransitionTyp_LaneAddLft || transitLn.transit_type == BA::TransitionTyp_LaneAddRght))
						{
							distBtwEgoCar2DecelLnOffset_B = transitLn.transit_lane_offset;
							if (transitLn.transit_type == BA::TransitionTyp_LaneAddLft)
							{
								transitLnTyp_B = BA::TransitionTyp_LaneAddLft;
							}
							else
							{
								transitLnTyp_B = BA::TransitionTyp_LaneAddRght;
							}
							break;
						}
					}
				}

				// 20220809: add right exit begin and left exit begin
				if (transitLnTyp_B == BA::TransitionTyp_NotAvlbl)
				{
					for (auto transitLn : this->input_BA.EHR_EHRInfo.transit_lanes_list)
					{
						if (transitLn.transit_lane_id == this->input_BA.RM_RoadModel.car_position.ego_lane_id \
					 		&& transitLn.transit_connected_lane_id == uint8_t(0) \
					 		&& (transitLn.transit_type == BA::TransitionTyp_LftExitBegin || transitLn.transit_type == BA::TransitionTyp_RghtExitBegin))
						{
							distBtwEgoCar2DecelLnOffset_C = transitLn.transit_lane_offset;
							if (transitLn.transit_type == BA::TransitionTyp_LftExitBegin)
							{
								transitLnTyp_C = BA::TransitionTyp_LftExitBegin;
							}
							else
							{
								transitLnTyp_C = BA::TransitionTyp_RghtExitBegin;
							}
							break;
						}
					}
				}
				
				DLOG(INFO) << "distBtwEgoCar2DecelLnOffset_B = " << distBtwEgoCar2DecelLnOffset_B << std::endl;
				DLOG(INFO) << "distBtwEgoCar2DecelLnOffset_C = " << distBtwEgoCar2DecelLnOffset_C << std::endl;
				if (distBtwEgoCar2DecelLnOffset_B > 0 && distBtwEgoCar2DecelLnOffset_B <= ADC::common::k_BA_DrcnIndAheadOnDistThrd)
				{
					isEgoCarClsToDecelStPnt = true;
				}
				if (distBtwEgoCar2DecelLnOffset_C > 0 && distBtwEgoCar2DecelLnOffset_C <= ADC::common::k_BA_DrcnIndAheadOnDistThrd02 && extraCond4ClsToSpltPnt == false)
				{
					isEgoCarClsToSpltPnt = true;
				}
			}
			DLOG(INFO) << "~~~isEgoCarClsToDecelStPnt = " << int(isEgoCarClsToDecelStPnt) << std::endl;
			DLOG(INFO) << "~~~isEgoCarClsToSpltPnt = " << int(isEgoCarClsToSpltPnt) << std::endl;
			bool isRghtDrcnIndFrcToOn = false;
			bool isLftDrcnIndFrcToOn = false;
			if (isEgoCarOnTrgtLn == false && isSplitTrgtRdTyp == true)
			{
				// get relevant neighbor lane type
				auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
					                            [egoLnID](const Lane &lane_)
												{
													return (egoLnID == lane_.start_id);
												});
				if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					// get ego lane's neighbor lane
					uint8_t neighborLnID = uint8_t(0);
					float neighborLnStRng = 0.0f;
					if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
					{
						if (!it_lane->right_neighbor_lane.empty())
						{
							// get the neighbor lane with the smallest and negative start range
							auto it_minStRngLn = std::min_element(it_lane->right_neighbor_lane.begin(), it_lane->right_neighbor_lane.end(), Cmp_by_stRng);
							neighborLnID = it_minStRngLn->lane_id;
							neighborLnStRng = it_minStRngLn->start_range;
							if (neighborLnStRng < 0)
							{
								auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
															 [neighborLnID](const WPExtendLaneStruct &wpExtendLn_)
															 {
																return (neighborLnID == wpExtendLn_.lane_id);
															 });
								if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									if (it_extLn->is_emergency_lane == uint8_t(0))
									{
										// get dist btw last lane change point and ego vehicle
										auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
					                            					[egoLnID](const WPExtendLaneStruct &lane_)
																	{
																		return (egoLnID == lane_.lane_id);
																	});
										float dstBtwEgoCarLastLcPnt = globalInfoProc.globalInfoContainer.m_DistToWayPoint - it_egoLn->lane_latest_lc_point_offset_to_wp;
										DLOG(INFO) << "dstBtwEgoCarLastLcPnt right = " << dstBtwEgoCarLastLcPnt << std::endl;
										if (dstBtwEgoCarLastLcPnt > 0 && dstBtwEgoCarLastLcPnt < ADC::common::k_BA_DrcnIndFrcdOnAheadDst && extraCond4ClsToSpltPnt == false)
										{
											isRghtDrcnIndFrcToOn = true;
										}
									}
								}
							}
						}
					}
					else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
					{
						if (!it_lane->left_neighbor_lane.empty())
						{
							// get the neighbor lane with the smallest and negative start range
							auto it_minStRngLn = std::min_element(it_lane->left_neighbor_lane.begin(), it_lane->left_neighbor_lane.end(), Cmp_by_stRng);
							neighborLnID = it_minStRngLn->lane_id;
							neighborLnStRng = it_minStRngLn->start_range;
							if (neighborLnStRng < 0)
							{
								auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
															 [neighborLnID](const WPExtendLaneStruct &wpExtendLn_)
															 {
																return (neighborLnID == wpExtendLn_.lane_id);
															 });
								if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									if (it_extLn->is_emergency_lane == uint8_t(0))
									{
										// get dist btw last lane change point and ego vehicle
										auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
					                           						[egoLnID](const WPExtendLaneStruct &lane_)
																	{
																		return (egoLnID == lane_.lane_id);
																	});
										float dstBtwEgoCarLastLcPnt = globalInfoProc.globalInfoContainer.m_DistToWayPoint - it_egoLn->lane_latest_lc_point_offset_to_wp;
										DLOG(INFO) << "dstBtwEgoCarLastLcPnt left = " << dstBtwEgoCarLastLcPnt << std::endl;
										if (dstBtwEgoCarLastLcPnt > 0 && dstBtwEgoCarLastLcPnt < ADC::common::k_BA_DrcnIndFrcdOnAheadDst && extraCond4ClsToSpltPnt == false)
										{
											isLftDrcnIndFrcToOn = true;
										}
									}
								}
							}
						}
					}
				}
			}
			DLOG(INFO) << "~isRghtDrcnIndFrcToOn = " << int(isRghtDrcnIndFrcToOn) << std::endl;
			DLOG(INFO) << "~isLftDrcnIndFrcToOn = " << int(isLftDrcnIndFrcToOn) << std::endl;

			// 20220727: direction indicator forced on in advance when two lane merges 
			bool isRghtDrcnIndOnMrgRd = false;
			bool isLftDrcnIndOnMrgRd = false;
			static float drctIndOnTimer = 0.0f;
			if (lastTrtRdTyp == BA::TargetRoadTyp_RmpMrgCARd && drctIndOnTimer < 1800.0f)
			{
				drctIndOnTimer = drctIndOnTimer + BEHAVIORANALYZERCYCLE/1000.0;
			}
			else
			{
				drctIndOnTimer = 0.0f;
			}
			DLOG(INFO) << "drctIndOnTimer = " << drctIndOnTimer << std::endl;
			if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd &&
			    this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChangePrepare &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChangePrepare &&
				ADC::common::k_BA_DrctIndOnInAdvFlg == true)
			{
				// ramp merge controll access road
				if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == 0 &&
				    globalInfoProc.globalInfoContainer.m_DistToWayPoint <= 25.0f &&
					globalInfoProc.globalInfoContainer.m_DistToWayPoint > 0.0f)
				{
					// 20220730: 
					// if next control waypoint is JCT or IC or SpltRd
					// and ego lane id is contained in next control waypoint target lane list
					// then no direction indicator on request
					// 20220903:
					// if next control waypoint is JCT or IC or SpltRd and
					// dist btw current waypoint and next waypoint is less than 2800m,
					// then no direction indicator on request
					bool noReqBefMrgCARd = false;
					if (this->input_BA.EHR_EHRInfo.control_waypoints_list.size() > 1)
					{
						if (this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_JCTRd ||
							this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_ICRd ||
							this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_target_road_type == BA::TargetRoadTyp_SpltRd)
						{
							auto it_res = std::find(this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_lane_id_list.begin(),
													this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_lane_id_list.end(),
													this->input_BA.RM_RoadModel.car_position.ego_lane_id);
						    if (it_res != this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_lane_id_list.end())
							{
								noReqBefMrgCARd = true;
							}
							float distBtwWayPnt = this->input_BA.EHR_EHRInfo.control_waypoints_list.at(1).control_waypoint_offset - globalInfoProc.globalInfoContainer.m_DistToWayPoint;
							if (distBtwWayPnt > 0 && distBtwWayPnt < 2800.0f)
							{
								noReqBefMrgCARd = true;
							}
						}
					}
					DLOG(INFO) << "noReqBefMrgCARd = " << int(noReqBefMrgCARd) << std::endl;
					if (noReqBefMrgCARd == false)
					{
						if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
						{
							isLftDrcnIndOnMrgRd = true;
							DLOG(INFO) << "before merge CA" << std::endl;
						}
						/* else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
						{
							isRghtDrcnIndOnMrgRd = true;
							DLOG(INFO) << "before merge CA" << std::endl;
						} */
					}
				}
				// ramp connecting road
				uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
				auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
											 [egoLnID](const WPExtendLaneStruct &wpln_)
											 {
												 return (egoLnID == wpln_.lane_id);
											 });
				if (it_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
				{
					float egoLnEndOffset = it_egoLn->current_extendlane_endoffset;
					// get ego lane' neighbor lane
					uint8_t minNeighborLnID = uint8_t(0);
					float minNeighborLnOffset = 100.0f;
					float minNeighborLnEndOffset = 0.0f;
					if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
					{
						if (!it_egoLn->left_extend_lane.empty())
						{
							minNeighborLnOffset = it_egoLn->left_extend_lane.front().adjacent_extendlane_offset;
							minNeighborLnEndOffset = it_egoLn->left_extend_lane.front().adjacent_extendlane_endoffset;
							for (auto neighborLn : it_egoLn->left_extend_lane)
							{
								if (minNeighborLnOffset > neighborLn.adjacent_extendlane_offset)
								{
									minNeighborLnOffset = neighborLn.adjacent_extendlane_offset;
									minNeighborLnID = neighborLn.adjacent_extendlane_id;
									minNeighborLnEndOffset = neighborLn.adjacent_extendlane_endoffset;
								}
							}
							DLOG(INFO) << "minNeighborLnOffset = " << minNeighborLnOffset << std::endl;
							DLOG(INFO) << "minNeighborLnEndOffset = " << minNeighborLnEndOffset << std::endl;
							DLOG(INFO) << "ehr egoLnEndOffset = " << egoLnEndOffset << std::endl;
							DLOG(INFO) << "neighbor lane id = " << int(minNeighborLnID) << std::endl;
							// 20220825: remove equal symbol in the third condition below
							if (minNeighborLnOffset < 25.0f && minNeighborLnOffset > -30.0f && minNeighborLnEndOffset > egoLnEndOffset)
							{
								isLftDrcnIndOnMrgRd = true;
								DLOG(INFO) << "connecting ramp" << std::endl;
							}
						}
					}
					/* else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
					{
						if (!it_egoLn->right_extend_lane.empty())
						{
							minNeighborLnOffset = it_egoLn->right_extend_lane.front().adjacent_extendlane_offset;
							minNeighborLnEndOffset = it_egoLn->right_extend_lane.front().adjacent_extendlane_endoffset;
							for (auto neighborLn : it_egoLn->right_extend_lane)
							{
								if (minNeighborLnOffset > neighborLn.adjacent_extendlane_offset)
								{
									minNeighborLnOffset = neighborLn.adjacent_extendlane_offset;
									minNeighborLnID = neighborLn.adjacent_extendlane_id;
									minNeighborLnEndOffset = neighborLn.adjacent_extendlane_endoffset;
								}
							}
							if (minNeighborLnOffset < 25.0f && minNeighborLnOffset > -30.0f && minNeighborLnEndOffset >= egoLnEndOffset)
							{
								isRghtDrcnIndOnMrgRd = true;
							}
						}
					} */	
				}
			}
			if (lastTrtRdTyp == BA::TargetRoadTyp_RmpMrgCARd && 
				globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum > uint8_t(0) &&
				drctIndOnTimer <= 10.0f &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChangePrepare &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChangePrepare &&
				ADC::common::k_BA_DrctIndOnInAdvFlg == true)
			{
				// 20220831:
				// if next control waypoint is JCT or IC or SpltRd
				// and ego lane id is contained in current control waypoint raw target lane list
				// then no direction indicator on request
				// 20220903:
				// if next control waypoint is JCT or IC or SpltRd or remind waypoint is subSplt
				// then no direction indicator on request, as the distance to waypoint is less than 3km
				bool noReqAftMrgCARd = false;
				if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
					globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
					globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd ||
					globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd)
				{
					/* auto it_res = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(),
											globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(),
											this->input_BA.RM_RoadModel.car_position.ego_lane_id);
					if (it_res != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
					{
						noReqAftMrgCARd = true;
					} */
					noReqAftMrgCARd = true;
				}
				// if left lane change is not listed in first passage list
				// then no direction indicator on request
				BA::BA_ActionTyp trgtActTyp = BA::BA_ActionTyp_LeftLaneChange;
				auto iter_passage = std::find_if(this->BA_FirstPassageList->begin(), this->BA_FirstPassageList->end(),
										   		[trgtActTyp](const Passage &passage_)
										   		{
											  		return (passage_.action_type == trgtActTyp);
										   		});
				if (iter_passage == this->BA_FirstPassageList->end())
				{
					noReqAftMrgCARd = true;
				}
				if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left && noReqAftMrgCARd == false)
				{
					isLftDrcnIndOnMrgRd = true;
					DLOG(INFO) << "after merge CA" << std::endl;
				}
				// 20220805: no right indicator request on after merge CARd
				/* else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
				{
					isRghtDrcnIndOnMrgRd = true;
				} */
			}		
			DLOG(INFO) << "isLftDrcnIndOnMrgRd = " << int(isLftDrcnIndOnMrgRd) << std::endl;
			DLOG(INFO) << "isRghtDrcnIndOnMrgRd = " << int(isRghtDrcnIndOnMrgRd) << std::endl;

			// 20220626: if direction indicator is forced on, then maybe lower set speed accordingly
			// 20220717: adjust set speed based on current road speed limit
			// decrease lane change number from 2 to 1
			if (isLftDrcnIndFrcToOn == true || isRghtDrcnIndFrcToOn == true)
			{
				if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == 1)
				{
					if (this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt01)
					{
						this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_LastLCPntSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
					}
					if (this->input_BA.SM_DrvrSetSpd < ADC::common::k_BA_MrgRdSpdLmt01 &&
					    this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt02)
					{
						float spdLmtVal = std::max(this->input_BA.SM_DrvrSetSpd * 0.85f, 68.0f);
						this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (spdLmtVal + 0.35f)) / c_BA_ms2kmh / 1.05f;
					}
				}
				else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgNum == 2)
				{
					// maximum set speed is 100kph if lane change number is 2
					if (this->input_BA.SM_DrvrSetSpd > ADC::common::k_BA_MrgRdSpdLmt01)
					{
						this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_MrgRdSpdLmt01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
					}
				}
			}

			// 20220910
			// reset static variable when lcavlbl change
			static BA::LCAvlblBhvr lcAvlblBhvr_lastcycle = BA::LCAvlblBhvr_None;
			if (lcAvlblBhvr_lastcycle != this->input_BA.SM_LCAvlblBehavior)
			{
				internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
				internDrcnIndReqDurTm = 0.0f;
				internDrcnIndReqTyp = 0;
			} 
			lcAvlblBhvr_lastcycle = this->input_BA.SM_LCAvlblBehavior;

			if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_SemiAuto ||
				(this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_None && this->input_BA.SM_LatActvFunc == BA::LatActvFuncSts_NRP))
			{
				switch (internDrcnIndReqSts)
				{
				case BA::DircnIndReqSts_BothOff:
					if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && isEgoCarClsToDecelStPnt == true)
					{
						// 20220711: direction indicator request sts is determined by transit type instead of route relative position
						if (transitLnTyp_B == BA::TransitionTyp_LaneAddRght)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						else if (transitLnTyp_B == BA::TransitionTyp_LaneAddLft)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						internDrcnIndReqTyp = 2;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && isEgoCarClsToSpltPnt == true)
					{
						// 20220810: direction indicator request sts is determined by transit type instead of route relative position
						if (transitLnTyp_C == BA::TransitionTyp_RghtExitBegin)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						else if (transitLnTyp_C == BA::TransitionTyp_LftExitBegin)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						internDrcnIndReqTyp = 6;
					}
					break;	
				case BA::DircnIndReqSts_LeftOn:
					internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
					// if direction indicator was shut down by driver, then stop direction indicator request
					if (/* this->BA_DirectionIndicatorInfo.m_left_direction_indicator_duration == 0.0f */
					    drcnTurnOff == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if ((internDrcnIndReqTyp == 2 || internDrcnIndReqTyp == 6) && internDrcnIndReqDurTm >= ADC::common::k_BA_LineFllwDrcnIndOnDurThr &&
							 this->BA_VehInLnCenter == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqDurTm >= ADC::common::k_BA_MaxDrcnIndOnDurTm && internDrcnIndReqTyp != 4)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 6 && isEgoCarClsToSpltPnt == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					break;
				case BA::DircnIndReqSts_RightOn:
					internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
					// if direction indicator was shut down by driver, then stop direction indicator request
					if (/* this->BA_DirectionIndicatorInfo.m_left_direction_indicator_duration == 0.0f */
					    drcnTurnOff == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if ((internDrcnIndReqTyp == 2 || internDrcnIndReqTyp == 6) && internDrcnIndReqDurTm >= ADC::common::k_BA_LineFllwDrcnIndOnDurThr &&
							 this->BA_VehInLnCenter == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqDurTm >= ADC::common::k_BA_MaxDrcnIndOnDurTm && internDrcnIndReqTyp != 4)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 6 && isEgoCarClsToSpltPnt == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					break;
				}
			}
			if (this->input_BA.SM_LatActvFunc != BA::LatActvFuncSts_NRP)
			{
				internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
				internDrcnIndReqDurTm = 0;
				internDrcnIndReqTyp = 0;
			}
			else if (this->input_BA.SM_LCAvlblBehavior == BA::LCAvlblBhvr_LaneChange_Auto)
			{
				switch (internDrcnIndReqSts)
				{
				case BA::DircnIndReqSts_BothOff:
					if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 1;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 1;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && isEgoCarClsToDecelStPnt == true)
					{
						// 20220711: direction indicator request sts is determined by transit type instead of route relative position
						if (transitLnTyp_B == BA::TransitionTyp_LaneAddRght)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						else if (transitLnTyp_B == BA::TransitionTyp_LaneAddLft)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						internDrcnIndReqTyp = 2;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangePrepare)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 3;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangePrepare)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 3;
					}
					else if (isLftDrcnIndFrcToOn == true && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 4;
					}
					else if (isRghtDrcnIndFrcToOn == true && this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 4;
					}
					else if (isLftDrcnIndOnMrgRd == true &&
							 this->msg_BA2SM.ba_opt_action != BA::BA_ActionTyp_LeftLaneChangeAbolition &&
							 this->msg_BA2SM.ba_opt_action != BA::BA_ActionTyp_RightLaneChangeAbolition)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 5;
					}
					else if (isRghtDrcnIndOnMrgRd == true &&
							 this->msg_BA2SM.ba_opt_action != BA::BA_ActionTyp_LeftLaneChangeAbolition &&
							 this->msg_BA2SM.ba_opt_action != BA::BA_ActionTyp_RightLaneChangeAbolition)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
						internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						internDrcnIndReqTyp = 5;
					}
					else if (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow && isEgoCarClsToSpltPnt == true)
					{
						// 20220810: direction indicator request sts is determined by transit type instead of route relative position
						if (transitLnTyp_C == BA::TransitionTyp_RghtExitBegin)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_RightOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						else if (transitLnTyp_C == BA::TransitionTyp_LftExitBegin)
						{
							internDrcnIndReqSts = BA::DircnIndReqSts_LeftOn;
							internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
						}
						internDrcnIndReqTyp = 6;
					}
					break;
				case BA::DircnIndReqSts_LeftOn:
					internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
					// if direction indicator was shut down by driver, then stop direction indicator request
					if (/* this->BA_DirectionIndicatorInfo.m_left_direction_indicator_duration == 0.0f */
					    drcnTurnOff == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					// left on request and right on request could not be connected step by step
					else if (internDrcnIndReqTyp == 1 && this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if ((internDrcnIndReqTyp == 2 || internDrcnIndReqTyp == 6) && internDrcnIndReqDurTm >= ADC::common::k_BA_LineFllwDrcnIndOnDurThr &&
							 this->BA_VehInLnCenter == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 1 && (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow || this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangeAbolition))
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 3 && (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangePrepare || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange))
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqDurTm >= ADC::common::k_BA_MaxDrcnIndOnDurTm && internDrcnIndReqTyp != 4)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 4 && isLftDrcnIndFrcToOn == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 5 && isLftDrcnIndOnMrgRd == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 5 && this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_LeftLaneChangeAbolition)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 6 && isEgoCarClsToSpltPnt == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					break;
				case BA::DircnIndReqSts_RightOn:
					internDrcnIndReqDurTm += BEHAVIORANALYZERCYCLE / 1000.0;
					// if direction indicator was shut down by driver, then stop direction indicator request
					if (/* this->BA_DirectionIndicatorInfo.m_left_direction_indicator_duration == 0.0f */
					    drcnTurnOff == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					// left on request and right on request could not be connected step by step
					else if (internDrcnIndReqTyp == 1 && this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if ((internDrcnIndReqTyp == 2 || internDrcnIndReqTyp == 6) && internDrcnIndReqDurTm >= ADC::common::k_BA_LineFllwDrcnIndOnDurThr &&
							 this->BA_VehInLnCenter == true)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 1 && (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChangeAbolition))
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 3 && (this->BA_OptAction_Deb == BA::BA_ActionTyp_LaneFollow || this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChangePrepare || this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange))
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqDurTm >= ADC::common::k_BA_MaxDrcnIndOnDurTm && internDrcnIndReqTyp != 4)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 4 && isRghtDrcnIndFrcToOn == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 5 && isRghtDrcnIndOnMrgRd == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 5 && this->msg_BA2SM.ba_opt_action == BA::BA_ActionTyp_RightLaneChangeAbolition)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					else if (internDrcnIndReqTyp == 6 && isEgoCarClsToSpltPnt == false)
					{
						internDrcnIndReqSts = BA::DircnIndReqSts_BothOff;
						internDrcnIndReqDurTm = 0.0;
						internDrcnIndReqTyp = 0;
					}
					break;
				}
				DLOG(INFO) << "///internDrcnIndReqTyp = " << int(internDrcnIndReqTyp) << std::endl;
				DLOG(INFO) << "///internDrcnIndReqSts = " << int(internDrcnIndReqSts) << std::endl;
				DLOG(INFO) << "///internDrcnIndReqDurTm = " << internDrcnIndReqDurTm << std::endl;
			}
			this->BA_DircnIndReq = internDrcnIndReqSts;
			return true;
		}

		uint64_t BehaviorAnalysis::GetFusionTimeStamp()
		{
			return this->input_BA.RM_RoadModel.fusion_timestamp;
		}

		void BehaviorAnalysis::FilterSpdLmt(const GlobalInfoProcess &globalInfoProc)
		{
			// 20220420
			// 1- if vehicle is closing to the target ramp, and vehicle is not on the corner lane, then filter the speed limit info
			//    if vehicle is on the raw target lanes or its neighbor lane is emergency lane or its neighbor lane belongs to raw target lanes, then vehicle is considered on the corner lane
			bool rampOutInFrtFlg = (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
									globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd || 
									globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd);
			if (rampOutInFrtFlg == true && globalInfoProc.globalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_MaxDstFltrSpdLmtInfo)
			{
				bool vehOnCrnrLn = false;
				uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
				auto it_egoLnInTrgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), egoLnID);
				if (it_egoLnInTrgtLnLst != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
				{
					vehOnCrnrLn = true;
				}
				else
				{
					auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
					                            [egoLnID](const Lane &lane_)
												{
													return (egoLnID == lane_.start_id);
												});
					if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
					{
						// get ego lane's neighbor lane
						uint8_t neighborLnID = uint8_t(0);
						if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
						{
							if (!it_lane->right_neighbor_lane.empty())
							{
								// get the neighbor lane with the smallest start range
								auto it_minStRngLn = std::min_element(it_lane->right_neighbor_lane.begin(), it_lane->right_neighbor_lane.end(), Cmp_by_stRng);
								neighborLnID = it_minStRngLn->lane_id;
								auto it_ngbLnInTrgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), neighborLnID);
								if (it_ngbLnInTrgtLnLst != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
								{
									// the neighbor lane belongs to raw target lanes
									vehOnCrnrLn = true;
								}
								// check whether neighbor lane is emergency lane
								// 20220420 by wpExtendLane
								auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
															 [neighborLnID](const WPExtendLaneStruct &wpExtendLn_)
															 {
																 return (neighborLnID == wpExtendLn_.lane_id);
															 });
								if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									if (it_extLn->is_emergency_lane == uint8_t(1))
									{
										vehOnCrnrLn = true;
									}
								}
							}
						}
						else if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Left)
						{
							if (!it_lane->left_neighbor_lane.empty())
							{
								// get the neighbor lane with the smallest start range
								auto it_minStRngLn = std::min_element(it_lane->left_neighbor_lane.begin(), it_lane->left_neighbor_lane.end(), Cmp_by_stRng);
								neighborLnID = it_minStRngLn->lane_id;
								auto it_ngbLnInTrgtLnLst = std::find(globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end(), neighborLnID);
								if (it_ngbLnInTrgtLnLst != globalInfoProc.globalInfoContainer.m_RawTgtLaneLst_CWP.end())
								{
									// the neighbor lane belongs to raw target lanes
									vehOnCrnrLn = true;
								}
								// check whether neighbor lane is emergency lane
								// 20220420 by wpExtendLane
								auto it_extLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
															 [neighborLnID](const WPExtendLaneStruct &wpExtendLn_)
															 {
																 return (neighborLnID == wpExtendLn_.lane_id);
															 });
								if (it_extLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									if (it_extLn->is_emergency_lane == uint8_t(1))
									{
										vehOnCrnrLn = true;
									}
								}
							}
						}
					}
				}
				// 
				if (vehOnCrnrLn == true)
				{
					this->BA_SpdLmtInfoFltred = false;
				}
				else
				{
					this->BA_SpdLmtInfoFltred = true;
				}
			}
			else
			{
				this->BA_SpdLmtInfoFltred = false;
			}
		}

		bool BehaviorAnalysis::Cmp_by_stRng(const NeighborLane &nLn01, const NeighborLane &nLn02)
		{
			return (nLn01.start_range < nLn02.start_range);
		}

		void BehaviorAnalysis::DynChgLCInterval(const GlobalInfoProcess &globalInfoProc, const BA::TargetRoadTyp &lastTrtRdTyp)
		{
			// 20220429
			// 1- if the distance btw ego vehicle and split type control waypoint is less than 600m
			//    then change lane change internal level to 1
			// 2- if ego vehicle just passes RmpMrgCARd, and timer value is less than 30 sec
			//    then change lane change internal level to 1
			// default level is 0
			bool isSplitTrgtRdTyp = (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
									 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd);
			static float dynIntrTimer = 0.0f;
			if (lastTrtRdTyp == BA::TargetRoadTyp_RmpMrgCARd && dynIntrTimer < 1800.0f)
			{
				dynIntrTimer = dynIntrTimer + BEHAVIORANALYZERCYCLE/1000.0;
			}
			else
			{
				dynIntrTimer = 0.0f;
			}

			if (isSplitTrgtRdTyp == true && globalInfoProc.globalInfoContainer.m_DistToWayPoint <= ADC::common::k_BA_LCIntervalDstThr)
			{
				this->BA_DynChgLCintervalLvl = uint8_t(1);
			}
			else if (lastTrtRdTyp == BA::TargetRoadTyp_RmpMrgCARd && dynIntrTimer < ADC::common::k_BA_DynIntervalChgInRmpMrgCARd)
			{
				this->BA_DynChgLCintervalLvl = uint8_t(1);
			}
			else
			{
				this->BA_DynChgLCintervalLvl = uint8_t(0);
			}
		}

		bool BehaviorAnalysis::DebounceTOR()
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internTORCntVal = 0;
			static uint8_t TORStsPrev = uint8_t(0);
			if (TORStsPrev != this->BA_TORTyp)
			{
				if (internTORCntVal < ADC::common::k_BA_TORStsDebCnt)
				{
					internTORCntVal++;
				}
				else
				{
					internTORCntVal = 0;
					TORStsPrev = this->BA_TORTyp;
				}
			}
			else
			{
				internTORCntVal = 0;
			}
			this->BA_TORTyp_Deb = TORStsPrev;
			return true;
		}

		bool BehaviorAnalysis::GetOptLaneID()
		{
			uint32_t optCrrdID = this->BA_OptCorridorID_Deb;
			this->BA_OptLaneID_Deb = uint8_t(0);
			for (auto lane : this->input_BA.RM_RoadModel.lane_list)
			{
				auto it_crrd = std::find_if(lane.corridor_list.begin(), lane.corridor_list.end(),
								           [optCrrdID](const Corridor &corridor_)
										   {
										       return (corridor_.corridor_id == optCrrdID);
										   });
				if (it_crrd != lane.corridor_list.end())
				{
					this->BA_OptLaneID_Deb = lane.start_id;
					return true;
				}
			}
			return false;
		}

		bool BehaviorAnalysis::JudgeDrcnIndTurnOff()
		{
			bool res = false;
			if (this->input_BA.SM_DirectionIndicatorSts != this->input_BA.BA_DirectionIndicatorSts_LastCycle && this->input_BA.BA_DirectionIndicatorSts_LastCycle != BA::DircnIndSts_Off)
			{
				res = true;		
			}
			this->input_BA.BA_DirectionIndicatorSts_LastCycle = this->input_BA.SM_DirectionIndicatorSts;
			return res;
		}

		bool BehaviorAnalysis::JudgeAutoLCSpdCond()
		{
			static BA::CondSts autoLCSpdCond = CondSts_NotSatisfied;
			switch (autoLCSpdCond)
			{
			case BA::CondSts_NotSatisfied:
				if (this->input_BA.VSE_LongitudinalVelocity > ADC::common::k_BA_AutoLCSpdHighThr/c_BA_ms2kmh)
				{
					autoLCSpdCond = CondSts_Satisfied;
				}
				break;
			case BA::CondSts_Satisfied:
				if (this->input_BA.VSE_LongitudinalVelocity < ADC::common::k_BA_AutoLCSpdLowThr/c_BA_ms2kmh)
				{
					autoLCSpdCond = CondSts_NotSatisfied;
				}
				break;
			}
			if (autoLCSpdCond == CondSts_Satisfied)
			{
				return true;
			}
			else	
			{
				return false;
			}
		}

		void BehaviorAnalysis::ShineHighlght(uint8_t &toShineHghlght, const uint32_t &ILCTrgtID)
		{
			// if ILCTrgtID is 0, then return
			// else get speed and relative position of ILC target
			// 1- ego speed minus object speed is greater than 1 m/s AND
			// 2- the minimum longitudinal distance btw ego vehicle and ILC object is less than 0 m/s
			// set toShineHghlght to 1 
			if (ILCTrgtID == uint8_t(0))
			{
				return;
			}
			else
			{
				auto it_ILCObj = std::find_if(this->input_BA.RM_RoadModel.object_list.begin(), this->input_BA.RM_RoadModel.object_list.end(),
				                              [ILCTrgtID](const Object &ILCObj_)
											  {
												  return (ILCTrgtID == ILCObj_.object_id);
											  });
			    if (it_ILCObj != this->input_BA.RM_RoadModel.object_list.end())
				{
					float objSpd = it_ILCObj->absolute_velocity_x;
					float objRelDst = it_ILCObj->relative_center_x;
					bool spdCond = (this->input_BA.VSE_LongitudinalVelocity - objSpd) > ADC::common::k_BA_MinShineHghtlghtSpdGap;
					bool spaceCond = (objRelDst - it_ILCObj->size_length / 2.0f - ADC::common::k_BA_LCGEgoLenth / 2.0f) <= ADC::common::k_BA_MinShineHghtlghtTH * this->input_BA.VSE_LongitudinalVelocity;
					if (spdCond == true && spaceCond == true)
					{
						toShineHghlght = uint8_t(1);
					}
				}
				else
				{
					toShineHghlght = uint8_t(0);
				}
			}
		}

		void BehaviorAnalysis::AdjSetSpd(const GlobalInfoProcess &globalInfoProc)
		{
			// if ego vehicle is on the most right side and distance to subSpltRd OR subMrgRd is less than 200/100m or right neighbor lane is not included in target lane list
			//    if set speed is greater than 98 kph, then adjust set speed to 90kph
			//    if set speed is less or equal to 98 kph and greater or equal to 78kph, then adjust set speed to max(0.9 * set speed, 72kph)
			// if ego lane is ending after 200m，and debounced action is not LCP or LC
			// then adjust set speed to neighbor lane's vehicle flow speed
			//    take the vehicles in lon range [-10, 20]m into consideration
			//    if there is no vehicle
			//    then no change to set speed
			//    else if lower set speed based on speed of front and back vehicles
			//    
			// 20220916
			// if there is a merge in connect ramp lane and certain conditions are also satisfied
			// then limit set speed to 50kph at most 
			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
                                        [egoLnID](const WPExtendLaneStruct &wpLn_)
                                        {
                                            return (wpLn_.lane_id == egoLnID);
                                        });
            if (it_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
            {
				bool tmpFlg = false;
				bool tmpFlg01 = false;
				// 20220815: get dist to split type control waypoint or reminder waypoint
				// if dist is less than 400m, then no active deceleration
				bool isCls2SplitTrgtRdTyp = ((globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
										 	  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
									 	 	  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
									 	      globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd) &&
											  globalInfoProc.globalInfoContainer.m_DistToWayPoint <= 400.0f) ||
										 	(!this->input_BA.EHR_EHRInfo.reminder_waypoint.empty() && 
											 this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubSpltRd &&
											 this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset <= 400.0f);
			
                if (!it_egoLn->right_extend_lane.empty() && isCls2SplitTrgtRdTyp == false)
                {
					uint8_t rghtNgbLnID = uint8_t(0);
					for (auto rghtNgb : it_egoLn->right_extend_lane)
					{
						if (rghtNgb.adjacent_extendlane_offset < 1.0f)
						{
							rghtNgbLnID = rghtNgb.adjacent_extendlane_id;
							break;
						}
					}
					if (rghtNgbLnID != uint8_t(0)) //20220723 added
					{
						auto it_rghtNgbLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
					    	                             [rghtNgbLnID](const WPExtendLaneStruct &wpLn_)
														 {
															 return(wpLn_.lane_id == rghtNgbLnID);
														 });
						DLOG(INFO) << "&right neighbor lane id = " << int(rghtNgbLnID) << std::endl;
						if (it_rghtNgbLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
						{
							if (it_rghtNgbLn->is_emergency_lane == 1)
							{
								tmpFlg = true;
							}
							else
							{
								auto it_rghtNgbEndPntLnFound = std::find(globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.begin(), globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end(), rghtNgbLnID);
                            	if (it_rghtNgbEndPntLnFound == globalInfoProc.globalInfoContainer.m_TgtLaneLst_CWP.end())
								{
									DLOG(INFO) << "right neighbor lane is not target lane!" << std::endl;
									// 20220723: right neighbor lane's end offset is less than 300m
									if (it_rghtNgbLn->current_extendlane_endoffset < 300.0f && it_egoLn->current_extendlane_endoffset > 305.0f)
									{
										tmpFlg01 = true;
									}
								}
							}
						}
					}
					else
					{
						tmpFlg = true;
					}
				}
				else
				{
					tmpFlg = true;
				}
				DLOG(INFO) << "&tmpFlg = " << int(tmpFlg) << std::endl;
				DLOG(INFO) << "&tmpFlg01 = " << int(tmpFlg01) << std::endl;
				if (!this->input_BA.EHR_EHRInfo.reminder_waypoint.empty())
				{
					if (tmpFlg == true &&
						this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubMrgRd &&
						!this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_lane_id_list.empty() )
					{
						DLOG(INFO) << "dist to submrgRd = " << this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset << std::endl;
						float dstToRmdWayPnt = this->input_BA.EHR_EHRInfo.reminder_waypoint.front().reminder_waypoint_offset;
						if (this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt01 && dstToRmdWayPnt <= ADC::common::k_BA_MrgRdSpdLmtDstThr01)
						{
							this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_LastLCPntSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
						}
						if (this->input_BA.SM_DrvrSetSpd < ADC::common::k_BA_MrgRdSpdLmt01 &&
						    this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt02 &&
							dstToRmdWayPnt <= ADC::common::k_BA_MrgRdSpdLmtDstThr02)
						{
							float spdLmtVal = std::max(this->input_BA.SM_DrvrSetSpd * 0.9f, 72.0f);
							this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (spdLmtVal + 0.35f)) / c_BA_ms2kmh / 1.05f;
						}
					}
				}
				if (tmpFlg01 == true)
				{
					if (this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt01)
					{
						this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_LastLCPntSetSpd01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
					}
					if (this->input_BA.SM_DrvrSetSpd < ADC::common::k_BA_MrgRdSpdLmt01 &&
						this->input_BA.SM_DrvrSetSpd >= ADC::common::k_BA_MrgRdSpdLmt02)
					{
						float spdLmtVal = std::max(this->input_BA.SM_DrvrSetSpd * 0.9f, 72.0f);
						this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (spdLmtVal + 0.35f)) / c_BA_ms2kmh / 1.05f;
					}
				}
			}
			//20220721
			// whether ego lane is ending after 200m
			// lane type 9: acceleration lane
			auto it_egoLne = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
                                         [egoLnID](const Lane &ln_)
                                         {
                                             return (ln_.start_id == egoLnID);
                                         });
			if (it_egoLne != this->input_BA.RM_RoadModel.lane_list.end() && 
				this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_LeftLaneChangePrepare &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChange &&
				this->BA_OptAction_Deb != BA::BA_ActionTyp_RightLaneChangePrepare)
			{
				// 9: acceleration lane 
				// 20: connectingramp
				// 16: entry
				if (it_egoLne->end_range <= 200 && 
				    it_egoLne->is_end_out_of_range == false && 
					(it_egoLne->lane_type == 9 || it_egoLne->lane_type == 20 || it_egoLne->lane_type == 16))
				{
					/* if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
					{ */
						uint8_t lftNgbrLnID = uint8_t(0);
						for (auto ngbrLn : it_egoLne->left_neighbor_lane)
						{
							if (ngbrLn.start_range < 0.0f)
							{
								lftNgbrLnID = ngbrLn.lane_id;
								break;
							}
						}
						if (lftNgbrLnID != uint8_t(0))
						{
							// get iterator of neighbor lane
							auto it_lftNgbrLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
                                        					[lftNgbrLnID](const Lane &ln_)
                                        					{
                                            					return (ln_.start_id == lftNgbrLnID);
                                        					});
							if (it_lftNgbrLn != this->input_BA.RM_RoadModel.lane_list.end())
							{
								BA::FeatureTag trgtFeatureTg = BA::FeatureTag_LineFllw;
								auto it_crrd = std::find_if(it_lftNgbrLn->corridor_list.begin(), it_lftNgbrLn->corridor_list.end(),
															[trgtFeatureTg](const Corridor &crrd_)
															{
																return (crrd_.feature_tag == trgtFeatureTg);
															});
								if (it_crrd != it_lftNgbrLn->corridor_list.end())
								{
									std::vector<ObjectLocation> objLocLst;
									for (auto objLoc : it_crrd->object_in_corridor_list)
									{
										if (objLoc.relative_center_x < 20.0f && objLoc.relative_center_x > -10.0f)
										{
											objLocLst.emplace_back(objLoc);
										}
									}
									if (!objLocLst.empty())
									{
										float spdSum = 0.0f;
										float avgSpd = 0.0f;
										for (auto objLocInLst : objLocLst)
										{
											spdSum = spdSum + objLocInLst.reference_point2corridor_frenet.ds;
										}
										avgSpd = spdSum / objLocLst.size();
										// set speed should be greater than 50kph
										// gap btw set speed and vehicle speed should be less than 3m/s
										// gap btw set speed and average vehicle speed should be 2m/s
										float tmpSetSpd = std::max(std::max(this->input_BA.VSE_LongitudinalVelocity - 2.0f, avgSpd - 2.0f), 46.0f/3.6f);
										if (tmpSetSpd < this->msg_BA2TRPLongt.ba_trp_v_sv_setspd)
										{
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = tmpSetSpd;
											DLOG(INFO) << "VSE_LongitudinalVelocity = " << this->input_BA.VSE_LongitudinalVelocity << std::endl;
											DLOG(INFO) << "avgSpd = " << avgSpd << std::endl;
											DLOG(INFO) << " Modify set speed to trplongt from << " << this->input_BA.SM_DrvrSetSpd / 3.6f << " to " << tmpSetSpd << std::endl; 
										}
									}
								}
							}
						}
					/* } */
				}

				// 20220916 on connecting ramp
				if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpMrgCARd &&
					it_egoLne->lane_type == 20 &&
					it_egoLne->end_range < 300.0f)
				{
					// get adjacent lane's offset and endoffset
					auto it_egoLn_ehr = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
													[egoLnID](const WPExtendLaneStruct &wpLn_)
													{
														return (egoLnID == wpLn_.lane_id);
													});
					if (it_egoLn_ehr != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
					{
						if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left)
						{
							DLOG(INFO) << "ego lane current_extendlane_endoffset = " << it_egoLn_ehr->current_extendlane_endoffset << std::endl;
							uint8_t lftFrtNrstLnID = uint8_t(0);
							float lftFrtNrstLnOffset = 1000.0f;
							float lftFrtNrstLnEndOffset = 1000.0f;
							float lftFrtNrstLnRealEndOffset = 1000.0f;
							if (!it_egoLn_ehr->left_extend_lane.empty())
							{
								lftFrtNrstLnID = it_egoLn_ehr->left_extend_lane.front().adjacent_extendlane_id;
								lftFrtNrstLnOffset = it_egoLn_ehr->left_extend_lane.front().adjacent_extendlane_offset;
								lftFrtNrstLnEndOffset = it_egoLn_ehr->left_extend_lane.front().adjacent_extendlane_endoffset;
								for (auto leftExtLn : it_egoLn_ehr->left_extend_lane)
								{
									if (leftExtLn.adjacent_extendlane_offset < lftFrtNrstLnOffset)
									{
										lftFrtNrstLnID = leftExtLn.adjacent_extendlane_id;
										lftFrtNrstLnOffset = leftExtLn.adjacent_extendlane_offset;
										lftFrtNrstLnEndOffset = leftExtLn.adjacent_extendlane_endoffset;
									}
								}
								// 20221024: get left front nearest lane's real endoffset
								auto it_lftNgbrLn_ehr = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
																	[lftFrtNrstLnID](const WPExtendLaneStruct &wpLn_)
																	{
																		return (lftFrtNrstLnID == wpLn_.lane_id);
																	});
								if (it_lftNgbrLn_ehr != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									lftFrtNrstLnRealEndOffset = it_lftNgbrLn_ehr->current_extendlane_endoffset;
								}
								DLOG(INFO) << "lftFrtNrstLnID = " << int(lftFrtNrstLnID) << std::endl;
								DLOG(INFO) << "lftFrtNrstLnOffset = " << lftFrtNrstLnOffset << std::endl;
								DLOG(INFO) << "lftFrtNrstLnEndOffset = " << lftFrtNrstLnEndOffset << std::endl;
								DLOG(INFO) << "lftFrtNrstLnRealEndOffset = " << lftFrtNrstLnRealEndOffset << std::endl;
								auto it_lftLne = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
                                         					 [lftFrtNrstLnID](const Lane &ln_)
                                         					 {
                                             					return (ln_.start_id == lftFrtNrstLnID);
                                         					 });
								if (it_lftLne != this->input_BA.RM_RoadModel.lane_list.end())
								{
									DLOG(INFO) << "k_BA_Dst2FrtNrstOffsetThr = " << int(k_BA_Dst2FrtNrstOffsetThr) << std::endl;
									DLOG(INFO) << "k_BA_MinFrtNrstLnLgthInCnctRmp = " << int(k_BA_MinFrtNrstLnLgthInCnctRmp) << std::endl;
									DLOG(INFO) << "it_lftLne->lane_type = " << int(it_lftLne->lane_type) << std::endl;
									if (it_lftLne->lane_type == 20 && lftFrtNrstLnOffset < ADC::common::k_BA_Dst2FrtNrstOffsetThr && (lftFrtNrstLnEndOffset - lftFrtNrstLnOffset) <= ADC::common::k_BA_MinFrtNrstLnLgthInCnctRmp && 
									    (this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange && this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange) &&
										(it_egoLn_ehr->current_extendlane_endoffset < lftFrtNrstLnRealEndOffset))
									{
										if (ADC::common::k_BA_CnnctRmpSetSpdThr < this->input_BA.SM_DrvrSetSpd)
										{
											DLOG(INFO) << "left IN 00" << std::endl;
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_CnnctRmpSetSpdThr + 0.35f)) / c_BA_ms2kmh / 1.05f;
										}
									}
									else if (it_lftLne->lane_type == 20 && lftFrtNrstLnOffset < ADC::common::k_BA_Dst2FrtNrstOffsetThr01 && (lftFrtNrstLnEndOffset - lftFrtNrstLnOffset) <= ADC::common::k_BA_MinFrtNrstLnLgthInCnctRmp01 && 
									    	(this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange && this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange) &&
											(it_egoLn_ehr->current_extendlane_endoffset < lftFrtNrstLnRealEndOffset))
									{
										if (ADC::common::k_BA_CnnctRmpSetSpdThr01 < this->input_BA.SM_DrvrSetSpd)
										{
											DLOG(INFO) << "left IN 01" << std::endl;
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_CnnctRmpSetSpdThr01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
										}
									}
								}
							}
						}
						else if (globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right)
						{
							uint8_t rghtFrtNrstLnID = uint8_t(0);
							float rghtFrtNrstLnOffset = 1000.0f;
							float rghtFrtNrstLnEndOffset = 1000.0f;
							float rghtFrtNrstLnRealEndOffset = 1000.0f;
							if (!it_egoLn_ehr->right_extend_lane.empty())
							{
								rghtFrtNrstLnID = it_egoLn_ehr->right_extend_lane.front().adjacent_extendlane_id;
								rghtFrtNrstLnOffset = it_egoLn_ehr->right_extend_lane.front().adjacent_extendlane_offset;
								rghtFrtNrstLnEndOffset = it_egoLn_ehr->right_extend_lane.front().adjacent_extendlane_endoffset;
								for (auto rightExtLn : it_egoLn_ehr->right_extend_lane)
								{
									if (rightExtLn.adjacent_extendlane_offset < rghtFrtNrstLnOffset)
									{
										rghtFrtNrstLnID = rightExtLn.adjacent_extendlane_id;
										rghtFrtNrstLnOffset = rightExtLn.adjacent_extendlane_offset;
										rghtFrtNrstLnEndOffset = rightExtLn.adjacent_extendlane_endoffset;
									}
								}
								// 20221024: get right front nearest lane's real endoffset
								auto it_rghtNgbrLn_ehr = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
																	 [rghtFrtNrstLnID](const WPExtendLaneStruct &wpLn_)
																	 {
																		 return (rghtFrtNrstLnID == wpLn_.lane_id);
																	 });
								if (it_rghtNgbrLn_ehr != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
								{
									rghtFrtNrstLnRealEndOffset = it_rghtNgbrLn_ehr->current_extendlane_endoffset;
								}
								auto it_rghtLne = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
                                         					 [rghtFrtNrstLnID](const Lane &ln_)
                                         					 {
                                             					return (ln_.start_id == rghtFrtNrstLnID);
                                         					 });
								if (it_rghtLne != this->input_BA.RM_RoadModel.lane_list.end())
								{
									if (it_rghtLne->lane_type == 20 && rghtFrtNrstLnOffset < ADC::common::k_BA_Dst2FrtNrstOffsetThr && (rghtFrtNrstLnEndOffset - rghtFrtNrstLnOffset) <= ADC::common::k_BA_MinFrtNrstLnLgthInCnctRmp &&
									    (this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange && this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange) &&
										(it_egoLn_ehr->current_extendlane_endoffset < rghtFrtNrstLnRealEndOffset))
									{
										if (ADC::common::k_BA_CnnctRmpSetSpdThr < this->input_BA.SM_DrvrSetSpd)
										{
											DLOG(INFO) << "right IN 00" << std::endl;
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_CnnctRmpSetSpdThr + 0.35f)) / c_BA_ms2kmh / 1.05f;
										}
									}
									else if (it_rghtLne->lane_type == 20 && rghtFrtNrstLnOffset < ADC::common::k_BA_Dst2FrtNrstOffsetThr01 && (rghtFrtNrstLnEndOffset - rghtFrtNrstLnOffset) <= ADC::common::k_BA_MinFrtNrstLnLgthInCnctRmp01 &&
									    	(this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_RightLaneChange && this->input_BA.BA_OptAction_LastCycle != BA::BA_ActionTyp_LeftLaneChange) &&
											(it_egoLn_ehr->current_extendlane_endoffset < rghtFrtNrstLnRealEndOffset))
									{
										if (ADC::common::k_BA_CnnctRmpSetSpdThr01 < this->input_BA.SM_DrvrSetSpd)
										{
											DLOG(INFO) << "right IN 01" << std::endl;
											this->msg_BA2TRPLongt.ba_trp_v_sv_setspd = std::max(0.0f, (ADC::common::k_BA_CnnctRmpSetSpdThr01 + 0.35f)) / c_BA_ms2kmh / 1.05f;
										}
									}
								}
							}
						}
					}
				}
			}
		}

		int BehaviorAnalysis::CalcVehFlowDense(const uint8_t &laneID)
		{
			// calculate vehicle number in range of [-50, 150]m
			int vehNum = 0;
			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
			                            [laneID](const Lane &lane_)
										{
											return (lane_.start_id == laneID);
										});
			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
			{
				BA::FeatureTag featureTag = BA::FeatureTag_LineFllw;
				auto it_crrd = std::find_if(it_lane->corridor_list.begin(), it_lane->corridor_list.end(),
				                            [featureTag](const Corridor &corridor_)
											{
												return (corridor_.feature_tag == featureTag);
											});
				if (it_crrd != it_lane->corridor_list.end())
				{
					for (auto objInCrrd : it_crrd->object_in_corridor_list)
					{
						if (objInCrrd.relative_center_x > -50.0f && objInCrrd.relative_center_x < 150.0f)
						{
							vehNum++;
						}
					}
				}
			}
			return vehNum;
		}

		float BehaviorAnalysis::CalcAvgVehFlowDense(const GlobalInfoProcess &globalInfoProc)
		{
			float avgVehFlowDense = 0.0f;
			if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
			    globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
				globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd)
			{
				std::vector<uint8_t> evlLnLst;
				if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
				{
					// evaluate three lane at most including ego lane
					bool ngbrLn01Ext = false;
					bool ngbrLn02Ext = false;
					uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
					evlLnLst.emplace_back(egoLnID);
					auto it_Lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
								            [egoLnID](const Lane &lane_)
											{
												return (egoLnID == lane_.start_id);
											});
					if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
					{
						if (!it_Lane->right_neighbor_lane.empty())
						{
							// get the neighbor lane with the smallest and negative start range
							auto it_minStRngLn = std::min_element(it_Lane->right_neighbor_lane.begin(), it_Lane->right_neighbor_lane.end(), Cmp_by_stRng);
							uint8_t rghtNeighborLnID = it_minStRngLn->lane_id;
							float rghtNeighborLnStRng = it_minStRngLn->start_range;
							if (rghtNeighborLnStRng < ADC::common::k_BA_MinNgbrLnStRng)
							{
								// check whether right neighbor lane is emergency lane  
								auto it_rghtNghbLn = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
							    	        					 [rghtNeighborLnID](const Lane &lane_)
											                  	 {
																    return (rghtNeighborLnID == lane_.start_id);
																 });
								if (it_rghtNghbLn != this->input_BA.RM_RoadModel.lane_list.end())
								{
									if (it_rghtNghbLn->lane_type != 5 && it_rghtNghbLn->lane_type != 21)
									{
										ngbrLn01Ext = true;
										evlLnLst.emplace_back(rghtNeighborLnID);
										if (!it_rghtNghbLn->right_neighbor_lane.empty())
										{
											// get the neighbor lane with the smallest and negative start range
											auto it_minStRngRghtLn = std::min_element(it_rghtNghbLn->right_neighbor_lane.begin(), it_rghtNghbLn->right_neighbor_lane.end(), Cmp_by_stRng);
											uint8_t rrNeighborLnID = it_minStRngRghtLn->lane_id;
											float rrNeighborLnStRng = it_minStRngRghtLn->start_range;
											if (rrNeighborLnStRng < ADC::common::k_BA_MinNgbrLnStRng)
											{
												auto it_RghtLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
								            							   		[rrNeighborLnID](const Lane &lane_)
																		   		{
																					return (rrNeighborLnID == lane_.start_id);
																		   		});
												if (it_RghtLane != this->input_BA.RM_RoadModel.lane_list.end())
												{
													if (it_RghtLane->lane_type != 5 && it_RghtLane->lane_type != 21)
													{
														ngbrLn02Ext = true;
														evlLnLst.emplace_back(rrNeighborLnID);
													}
												}
											}
										}	
									}
								}
							}
						}
					}
					if (evlLnLst.size() == 2)
					{
						float egoLnVehFlowDense = CalcVehFlowDense(evlLnLst[0]);
						float rghtLnVehFlowDense = CalcVehFlowDense(evlLnLst[1]);
						DLOG(INFO) << "evaluated lane number is 2" << std::endl;
						DLOG(INFO) << "evaluated lane id 0 = " << int(evlLnLst[0]) << std::endl;
						DLOG(INFO) << "evaluated lane id 1 = " << int(evlLnLst[1]) << std::endl;
						DLOG(INFO) << "egoLnVehFlowDense = " << egoLnVehFlowDense << std::endl;
						DLOG(INFO) << "rghtLnVehFlowDense = " << rghtLnVehFlowDense << std::endl;
						avgVehFlowDense = (egoLnVehFlowDense + rghtLnVehFlowDense) / 2.0;
					}
					else if (evlLnLst.size() == 3)
					{
						float egoLnVehFlowDense = CalcVehFlowDense(evlLnLst[0]);
						float rghtLnVehFlowDense = CalcVehFlowDense(evlLnLst[1]);
						float rrLnVehFlowDense = CalcVehFlowDense(evlLnLst[2]);
						DLOG(INFO) << "evaluated lane number is 3" << std::endl;
						DLOG(INFO) << "evaluated lane id 0 = " << int(evlLnLst[0]) << std::endl;
						DLOG(INFO) << "evaluated lane id 1 = " << int(evlLnLst[1]) << std::endl;
						DLOG(INFO) << "evaluated lane id 2 = " << int(evlLnLst[2]) << std::endl;
						DLOG(INFO) << "egoLnVehFlowDense = " << egoLnVehFlowDense << std::endl;
						DLOG(INFO) << "rghtLnVehFlowDense = " << rghtLnVehFlowDense << std::endl;
						DLOG(INFO) << "rrLnVehFlowDense = " << rrLnVehFlowDense << std::endl;
						avgVehFlowDense = (egoLnVehFlowDense + rghtLnVehFlowDense + rrLnVehFlowDense) / 3.0;
					}
				}

				return avgVehFlowDense;
			}
			else
			{
				return avgVehFlowDense;
			}
		}

		bool BehaviorAnalysis::CalcSmplPntTimeStamp()
		{
    		// this->m_CycleTimer += BEHAVIORANALYZERCYCLE / 1000.0;
    		static float cycleTimer = 0.0;
    		if (cycleTimer >= ADC::common::k_BA_HistVehInfoSmplCycle)
    		{
        		// it is time to operate the HistHostVehInfo
        		cycleTimer = 0.0; // reset cycle timer
    	    	return true;
    		}
    		cycleTimer += BEHAVIORANALYZERCYCLE / 1000.0;
    		return false;
		}

		bool BehaviorAnalysis::CalcDstToRLnBndry(float &dstToRghtLnBndry, const uint8_t &laneID)
		{
 			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   [laneID](const Lane &lane_)
                     		           {
                                		   return (laneID == lane_.start_id);
                              		   });
 			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
  			{
    			// get first lane marker id of right side
    			uint32_t trgtRghtLnMrkID = uint32_t(0);
    			if (!it_lane->right_lane_marker_id_list.empty())
    			{
      				trgtRghtLnMrkID = it_lane->right_lane_marker_id_list.front();
    			}
    			else
    			{
					DLOG(INFO) << "right lane marker is empty!" << std::endl;
      				return false;
    			}
    			// get first front lane boundary of the lane marker object
    			auto it_laneMarker = std::find_if(this->input_BA.RM_RoadModel.lane_marker_list.begin(), this->input_BA.RM_RoadModel.lane_marker_list.end(),
            			                         [trgtRghtLnMrkID](const LaneMarker &laneMarker_)
                    			                 {
                        	    		            return (trgtRghtLnMrkID == laneMarker_.id);
                            	        	     });
    			if (it_laneMarker != this->input_BA.RM_RoadModel.lane_marker_list.end())
    			{
      				LaneBoundary trgtLnBndry;
      				if (!it_laneMarker->front_lane_boundary_list.empty())
      				{
        				trgtLnBndry = it_laneMarker->front_lane_boundary_list.front();
        				LineGeometry trgtLineGeo = trgtLnBndry.line_geometry;
        				dstToRghtLnBndry = trgtLineGeo.c0;
						return true;
					}
					else
					{
						DLOG(INFO) << "front lane boundary is empty!" << std::endl;
						return false;
					}
				}
      			else
      			{
					  DLOG(INFO) << "lane marker is not found!" << std::endl;
        			return false;
      			}  
  			}
  			else
  			{
				  DLOG(INFO) << "lane is not found!" << std::endl;
    			return false;
 			}
		}

		bool BehaviorAnalysis::CalcDstToLLnBndry(float &dstToLftLnBndry, const uint8_t &laneID)
		{
 			auto it_lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
            		                   [laneID](const Lane &lane_)
                     		           {
                                		   return (laneID == lane_.start_id);
                              		   });
 			if (it_lane != this->input_BA.RM_RoadModel.lane_list.end())
  			{
    			// get first lane marker id of left side
    			uint32_t trgtLftLnMrkID = uint32_t(0);
    			if (!it_lane->left_lane_marker_id_list.empty())
    			{
      				trgtLftLnMrkID = it_lane->left_lane_marker_id_list.front();
    			}
    			else
    			{
      				return false;
    			}
    			// get first front lane boundary of the lane marker object
    			auto it_laneMarker = std::find_if(this->input_BA.RM_RoadModel.lane_marker_list.begin(), this->input_BA.RM_RoadModel.lane_marker_list.end(),
            			                         [trgtLftLnMrkID](const LaneMarker &laneMarker_)
                    			                 {
                        	    		            return (trgtLftLnMrkID == laneMarker_.id);
                            	        	     });
    			if (it_laneMarker != this->input_BA.RM_RoadModel.lane_marker_list.end())
    			{
      				LaneBoundary trgtLnBndry;
      				if (!it_laneMarker->front_lane_boundary_list.empty())
      				{
        				trgtLnBndry = it_laneMarker->front_lane_boundary_list.front();
        				LineGeometry trgtLineGeo = trgtLnBndry.line_geometry;
        				dstToLftLnBndry = trgtLineGeo.c0;
					}
					return true;
				}
      			else
      			{
        			return false;
      			}  
  			}
  			else
  			{
    			return false;
 			}
		}

		void BehaviorAnalysis::UpdatePDDbgInfo(const PD::PDDebug &pd_Debug)
		{
			this->msg_BA2PDDebug = pd_Debug;
		}

		bool BehaviorAnalysis::DebounceDrcnIndReq()
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internIndReqCntVal = 0;
			static BA::DircnIndReqSts indReqPrev = BA::DircnIndReqSts_BothOff;
			if (indReqPrev != this->BA_DircnIndReq)
			{
				if (internIndReqCntVal < ADC::common::k_BA_DrcnIndReqDebCnt)
				{
					internIndReqCntVal++;
				}
				else
				{
					internIndReqCntVal = 0;
					indReqPrev = this->BA_DircnIndReq;
				}
			}
			else
			{
				internIndReqCntVal = 0;
			}
			this->BA_DircnIndReq_Deb = indReqPrev;
			/* if (this->BA_DircnIndReq != BA::DircnIndReqSts_BothOff)
			{
				this->BA_DircnIndReq_Deb = this->BA_DircnIndReq;
			} */
			return true;
		}

		bool BehaviorAnalysis::GetLineTyp(const uint32_t &laneMrkID, uint8_t &lineTyp)
		{
			auto it_laneMarker = std::find_if(this->input_BA.RM_RoadModel.lane_marker_list.begin(), this->input_BA.RM_RoadModel.lane_marker_list.end(),
            			                     [laneMrkID](const LaneMarker &laneMarker_)
                    			             {
                        	    		        return (laneMrkID == laneMarker_.id);
                            	        	 });
    		if (it_laneMarker != this->input_BA.RM_RoadModel.lane_marker_list.end())
    		{
				if (!it_laneMarker->line_type_attribute_list.empty())
				{
					// get current line type
					for (auto lineTypAtt : it_laneMarker->line_type_attribute_list)
					{
						if (lineTypAtt.start_offset <= 0.0f && lineTypAtt.end_offset > 0.0f)
						{
							lineTyp = lineTypAtt.linetype;
							return true;
						}
					}
				}
			}
			return false;
		}

		bool BehaviorAnalysis::DebounceFlowDense()
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internFlowDenseCntVal = 0;
			static bool flowDenseHighPrev = false;
			if (flowDenseHighPrev != this->BA_AvgFlowDenseHigh)
			{
				if (internFlowDenseCntVal < ADC::common::k_BA_FlowDenseDebCnt)
				{
					internFlowDenseCntVal++;
				}
				else
				{
					internFlowDenseCntVal = 0;
					flowDenseHighPrev = this->BA_AvgFlowDenseHigh;
				}
			}
			else
			{
				internFlowDenseCntVal = 0;
			}
			this->BA_AvgFlowDenseHigh_Deb = flowDenseHighPrev;
			return true;
		}

		bool BehaviorAnalysis::ProcessLnVehFlowSpdCond(uint8_t &lftNbghLnID, 
													   uint8_t &rghtNbghLnID, 
													   const bool &chgSpdThrFlg, 
													   const GlobalInfoProcess &globalInfoProc)
		{
			// get each lane id
			static uint8_t egoLnID_LastCycle = uint8_t(0);
			uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			// uint8_t lftNbghLnID = uint8_t(0);
			// uint8_t rghtNbghLnID = uint8_t(0);
			auto it_Lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
								        [egoLnID](const Lane &lane_)
										{
											return (egoLnID == lane_.start_id);
										});
			if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
			{
				if (!it_Lane->left_neighbor_lane.empty())
				{
					auto it_minLftStRngLn = std::min_element(it_Lane->left_neighbor_lane.begin(), it_Lane->left_neighbor_lane.end(), Cmp_by_stRng);
					lftNbghLnID = it_minLftStRngLn->lane_id;
				}
				if (!it_Lane->right_neighbor_lane.empty())
				{
					auto it_minRghtStRngLn = std::min_element(it_Lane->right_neighbor_lane.begin(), it_Lane->right_neighbor_lane.end(), Cmp_by_stRng);
					rghtNbghLnID = it_minRghtStRngLn->lane_id;
				}
			}
			// get each lane's vehicle flow speed
			float egoLnAvgSpd = 0.0f;
			float lftNghrLnAvgSpd = 0.0f;
			float rghtNghrLnAvgSpd = 0.0f;
			uint32_t egoFrtNrstObjID = uint32_t(0);
			uint32_t lftFrtNrstObjID = uint32_t(0);
			uint32_t rghtFrtNrstObjID = uint32_t(0);
			float egoFrtNrstObjDst = 1000.0f;
			float lftFrtNrstObjDst = 1000.0f;
			float rghtFrtNrstObjDst = 1000.0f;
			if (ADC::common::k_BA_AlphaVehFlowCalcAppl == true)
			{
				CalcLnVehFlowSpd_alpha(egoLnID,
							 		   this->input_BA.RM_RoadModel,
							 		   this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
							 		   egoLnAvgSpd,
							 		   egoFrtNrstObjID,
							 		   egoFrtNrstObjDst);
				CalcLnVehFlowSpd_alpha(lftNbghLnID,
							 		   this->input_BA.RM_RoadModel,
							 		   this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
							 		   lftNghrLnAvgSpd,
							 		   lftFrtNrstObjID,
							 		   lftFrtNrstObjDst);
				CalcLnVehFlowSpd_alpha(rghtNbghLnID,
							 		   this->input_BA.RM_RoadModel,
							 		   this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
							 		   rghtNghrLnAvgSpd,
							 		   rghtFrtNrstObjID,
							 		   rghtFrtNrstObjDst);
			}
			else
			{
				CalcLnVehFlowSpd(egoLnID,
				    			 this->input_BA.RM_RoadModel,
							     this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
							 	 egoLnAvgSpd,
							     egoFrtNrstObjID,
							     egoFrtNrstObjDst);
				CalcLnVehFlowSpd(lftNbghLnID,
				 				 this->input_BA.RM_RoadModel,
					 			 this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
						 		 lftNghrLnAvgSpd,
						 		 lftFrtNrstObjID,
							 	 lftFrtNrstObjDst);
				CalcLnVehFlowSpd(rghtNbghLnID,
				 				 this->input_BA.RM_RoadModel,
					 			 this->input_BA.SM_DrvrSetSpd / c_BA_ms2kmh,
						 		 rghtNghrLnAvgSpd,
							 	 rghtFrtNrstObjID,
							  	 rghtFrtNrstObjDst);
			}

			DLOG(INFO) << "lftNghrLnAvgSpd = " << lftNghrLnAvgSpd << std::endl;
			DLOG(INFO) << "rghtNghrLnAvgSpd = " << rghtNghrLnAvgSpd << std::endl;
			DLOG(INFO) << "egoLnAvgSpd = " << egoLnAvgSpd << std::endl;
			// calc current speed condition flag
			// float effcTrffcCostSpdPerc = ADC::common::k_BA_EffcTrffcCostSpdPerc;
			float effcTrffcCostSpdPerc_lft = ADC::common::k_BA_EffcTrffcCostSpdPerc;
			float effcTrffcCostSpdPerc_rght = ADC::common::k_BA_EffcTrffcCostSpdPerc;
			bool antiLCCond01 = false;
			bool antiLCCond02 = false;
			if (chgSpdThrFlg == true)
			{
				// 20220915: 
				// if dist btw front closest target on left neighbor lane and ego front target is greater than 70m
				// then change spdperc to 0.2
				// else 
				// change spdperc to 0.3
				// 20221112:
				// add another prediction about whether overtaking front slow car is available
				// get neighbor lane's desired lane change point and calc gap s1
				// first calc arriving time of above desired lane change point based on average speed of ego speed and neighbor lanes's vehicle flow speed
				// then calc moving distance of front nearest vehicle of ego lane during above arriving time s2
				// if s1 > s2 + current gap btw ego car and front nearest vehicle of ego lane, then overtaking the nearest front vehicle condition 01 is satisfied
				// ###########################################################
				// get gap btw front nearest vehicle and second nearest vehicle
				// if the above gap is greater than 4 times average speed of above two vehicles, then overtaking the nearest front vehicle condition 02 is satisfied
				float lftNgbrLnDesLCPnt = globalInfoProc.globalInfoContainer.m_DistToWayPoint;
				float s01 = 0.0f;
				float s02 = 0.0f;
				float arrvTm01 = 0.0f;
				float avgArrvSpd = 1.0f;
				if (globalInfoProc.globalInfoContainer.m_RouteRltvPos == BA::RouteRltvPos_Right)
				{
					auto it_glbInfo = std::find_if(globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.begin(), globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end(),
											   [lftNbghLnID](const BA::GlobalLaneInfo &glbLnInfo_)
											   {
												   return (lftNbghLnID == glbLnInfo_.m_LaneID);
											   });
					if (it_glbInfo != globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end())
					{
						if (!it_glbInfo->m_DesiredLaneChangePoint.empty())
						{
							lftNgbrLnDesLCPnt = it_glbInfo->m_DesiredLaneChangePoint.front();
						}
						s01 = globalInfoProc.globalInfoContainer.m_DistToWayPoint - lftNgbrLnDesLCPnt;
						if (s01 > 0.0f)
						{
							avgArrvSpd = 0.6f * lftNghrLnAvgSpd + 0.4f * this->input_BA.VSE_LongitudinalVelocity;
							if (std::abs(avgArrvSpd) < 1.0f)
							{
								avgArrvSpd = 1.0f;
							}
							arrvTm01 = s01 / avgArrvSpd;
							s02 = arrvTm01 * egoLnAvgSpd + egoFrtNrstObjDst;
							if (s01 > s02)
							{
								antiLCCond01 = true;
							}
						}
						DLOG(INFO) << "lftNgbrLnDesLCPnt = " << lftNgbrLnDesLCPnt << std::endl;
						DLOG(INFO) << "s01 = " << s01 << std::endl;
						DLOG(INFO) << "s02 = " << s02 << std::endl;
						DLOG(INFO) << "arrvTm01 = " << arrvTm01 << std::endl;
						DLOG(INFO) << "avgArrvSpd = " << avgArrvSpd << std::endl;
						DLOG(INFO) << "egoFrtNrstObjDst = " << egoFrtNrstObjDst << std::endl;
						// get front second nearest object info
						float frtScndNrstObjDst = 1000.0f;
						uint8_t frtScndNrstObjID = uint8_t(0);
						if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
						{
							BA::FeatureTag LineFollowTag = BA::FeatureTag_LineFllw;
							std::vector<ObjectLocation> tmpObjLocList;
							auto it_corridor = std::find_if(it_Lane->corridor_list.begin(), it_Lane->corridor_list.end(),
															[LineFollowTag](const Corridor &corridor_)
															{
																return (LineFollowTag == corridor_.feature_tag);
															});
							if (it_corridor != it_Lane->corridor_list.end())
							{
								for (auto objLoc_ : it_corridor->object_in_corridor_list)
								{
									if (objLoc_.relative_center_x > egoFrtNrstObjDst && objLoc_.relative_center_x <= ADC::common::k_BA_VehFlowCalcMaxRng)
									{
										tmpObjLocList.emplace_back(objLoc_);
										if (objLoc_.relative_center_x < frtScndNrstObjDst)
										{
											frtScndNrstObjDst = objLoc_.relative_center_x;
											frtScndNrstObjID = objLoc_.object_id;
										}
									}
								}
							}
						}
						float gapBtwFrtTwoNrtObj = 0.0f;
						float minGapBtwFrtTwoNrtObjThr = 0.0f;
						if (frtScndNrstObjID != uint8_t(0))
						{
							gapBtwFrtTwoNrtObj = frtScndNrstObjDst - egoFrtNrstObjDst;
							minGapBtwFrtTwoNrtObjThr = egoLnAvgSpd * ADC::common::k_BA_MinTHBtwFrtNrstObj;
							if (gapBtwFrtTwoNrtObj > minGapBtwFrtTwoNrtObjThr)
							{
								antiLCCond02 = true;
							}
						}
						else
						{
							// front second nearest obj does not exist
							antiLCCond02 = true;
						}
						DLOG(INFO) << "frtScndNrstObjDst = " << frtScndNrstObjDst << std::endl;
					}
				}
				
				if (lftFrtNrstObjDst - egoFrtNrstObjDst > ADC::common::k_BA_AllwAntiLCObjDstThr)
				{
					effcTrffcCostSpdPerc_lft = ADC::common::k_BA_EffcTrffcCostSpdPerc01;
				}
				else
				{
					effcTrffcCostSpdPerc_lft = ADC::common::k_BA_EffcTrffcCostSpdPerc02;
				}
				if (rghtFrtNrstObjDst - egoFrtNrstObjDst > ADC::common::k_BA_AllwAntiLCObjDstThr)
				{
					effcTrffcCostSpdPerc_rght = ADC::common::k_BA_EffcTrffcCostSpdPerc01;
				}
				else
				{
					effcTrffcCostSpdPerc_rght = ADC::common::k_BA_EffcTrffcCostSpdPerc02;
				}
			}
			DLOG(INFO) << "~~effcTrffcCostSpdPerc left = " << effcTrffcCostSpdPerc_rght << std::endl;
			DLOG(INFO) << "~~effcTrffcCostSpdPerc right= " << effcTrffcCostSpdPerc_lft << std::endl;
			DLOG(INFO) << "chgSpdThrFlg = " << int(chgSpdThrFlg) << std::endl;
			DLOG(INFO) << "antiLCCond01 = " << int(antiLCCond01) << std::endl;
			DLOG(INFO) << "antiLCCond02 = " << int(antiLCCond02) << std::endl;
			if (chgSpdThrFlg == false || (chgSpdThrFlg == true && antiLCCond01 == true && antiLCCond02 == true))
			{
				this->BA_LeftLnSpdCond = (lftNghrLnAvgSpd >= (1.0 + effcTrffcCostSpdPerc_lft) * egoLnAvgSpd);
				this->BA_RightLnSpdCond = (rghtNghrLnAvgSpd >= (1.0 + effcTrffcCostSpdPerc_rght) * egoLnAvgSpd);
			}
			else
			{
				this->BA_LeftLnSpdCond = false;
				this->BA_RightLnSpdCond = false;
			}
			
			// debounce speed condition flag
			bool egoLnIDChg = false;
			if (egoLnID != egoLnID_LastCycle)
			{
				egoLnIDChg = true;
			}
			DLOG(INFO) << "egoLnIDChg = " << int(egoLnIDChg) << std::endl;
			DebounceSpdCond_L(this->BA_LeftLnSpdCond, this->BA_LeftLnSpdCond_Deb, egoLnIDChg);
			DebounceSpdCond_R(this->BA_RightLnSpdCond, this->BA_RightLnSpdCond_Deb, egoLnIDChg);
			// 20221116 added: reset debounced flow speed condition flag directly
			if (chgSpdThrFlg == true && (antiLCCond01 == false || antiLCCond02 == false))
			{
				this->BA_LeftLnSpdCond_Deb = false;
				this->BA_RightLnSpdCond_Deb = false;
			}
			DLOG(INFO) << "-- left flow spd cond flag = " << int(this->BA_LeftLnSpdCond) << std::endl;
			DLOG(INFO) << "-- right flow spd cond flag = " << int(this->BA_RightLnSpdCond) << std::endl;
			DLOG(INFO) << "-- left flow spd cond flag deb = " << int(this->BA_LeftLnSpdCond_Deb) << std::endl;
			DLOG(INFO) << "-- right flow spd cond flag deb = " << int(this->BA_RightLnSpdCond_Deb) << std::endl;
			egoLnID_LastCycle = egoLnID;
			return true;
		}

		bool BehaviorAnalysis::DebounceSpdCond_L(const bool &input, bool &output, bool &laneChgFlg)
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internSpdCondCntVal_L = 0;
			static bool spdCondPrev_L = false;
			// if ego lane id changed, then reset internal value
			if (laneChgFlg == true)
			{
				spdCondPrev_L = false;
			}
			// select debounce counter value by drive mode
			int spdCondDebCnt = ADC::common::k_BA_SpdCondDebCnt;
			if (this->input_BA.SM_DriveMode == BA::DriveMd_Sport)
			{
				spdCondDebCnt = ADC::common::k_BA_SpdCondDebCnt_Aggr;
			}
			if (spdCondPrev_L != input)
			{
				if (internSpdCondCntVal_L < spdCondDebCnt)
				{
					internSpdCondCntVal_L++;
				}
				else
				{
					internSpdCondCntVal_L = 0;
					spdCondPrev_L = input;
				}
			}
			else
			{
				internSpdCondCntVal_L = 0;
			}
			output = spdCondPrev_L;
			return true;
		}

		bool BehaviorAnalysis::DebounceSpdCond_R(const bool &input, bool &output, bool &laneChgFlg)
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internSpdCondCntVal_R = 0;
			static bool spdCondPrev_R = false;
			// if ego lane id changed, then reset internal value
			if (laneChgFlg == true)
			{
				spdCondPrev_R = false;
			}
			// select debounce counter value by drive mode
			int spdCondDebCnt = ADC::common::k_BA_SpdCondDebCnt;
			if (this->input_BA.SM_DriveMode == BA::DriveMd_Sport)
			{
				spdCondDebCnt = ADC::common::k_BA_SpdCondDebCnt_Aggr;
			}
			if (spdCondPrev_R != input)
			{
				if (internSpdCondCntVal_R < spdCondDebCnt)
				{
					internSpdCondCntVal_R++;
				}
				else
				{
					internSpdCondCntVal_R = 0;
					spdCondPrev_R = input;
				}
			}
			else
			{
				internSpdCondCntVal_R = 0;
			}
			output = spdCondPrev_R;
			return true;
		}

		uint8_t BehaviorAnalysis::CalcCycleCounter2SM()
		{
			static uint8_t cnt = 0;
			if (cnt == 255)
			{
				cnt = 0;
			}
			else
			{
				cnt++;
			}
			return cnt;
		} 

		/* uint8_t BehaviorAnalysis::CalcCycleCounter2SM_InclOthrNodes()
		{
			static uint8_t cnt = 0;
			static uint16_t cycleCntLoc_RM = 0;
			static float ccDurTm_RM = 0.0f;
			// calc the duration time during which RM cycleCounter keeps the same
			if (cycleCntLoc_RM == this->input_BA.RM_RoadModel.rm_header.cyclecounter)
			{
				ccDurTm_RM += BEHAVIORANALYZERCYCLE / 1000.0;
				if (ccDurTm_RM > 2.0f)
				{
					ccDurTm_RM = 2.0f;
				}
			}
			else
			{
				ccDurTm_RM = 0.0f;
				cycleCntLoc_RM = this->input_BA.RM_RoadModel.rm_header.cyclecounter;
			}

			if (ccDurTm_RM <= ADC::common::k_BA_RMMsgStopUpdtThr)
			{
				if (cnt == 255)
				{
					cnt = 0;
				}
				else
				{
					cnt++;
				}
				return cnt;
			}
			else
			{
				return 100; // RM msg has not been updated
			}
		} */

		bool BehaviorAnalysis::MonitorRMCycleCounter()
		{
			static uint16_t cycleCntLoc_RM = 0;
			static float ccDurTm_RM = 0.0f;
			// calc the duration time during which RM cycleCounter keeps the same
			if (cycleCntLoc_RM == this->input_BA.RM_RoadModel.rm_header.cyclecounter)
			{
				ccDurTm_RM += BEHAVIORANALYZERCYCLE / 1000.0;
				if (ccDurTm_RM > 2.0f)
				{
					ccDurTm_RM = 2.0f;
				}
				this->msg_BA2Obs.m_ba_optimal_corridor_id = this->input_BA.RM_RoadModel.rm_header.cyclecounter;
				this->msg_BA2Obs.m_ba_optimal_target_obj_id = this->input_BA.EHR_EHRInfo.msg_header.cycle_counter;
				this->msg_BA2Obs.m_ba_optimal_action = this->input_BA.SM_LatActvFunc;
			}
			else
			{
				ccDurTm_RM = 0.0f;
				cycleCntLoc_RM = this->input_BA.RM_RoadModel.rm_header.cyclecounter;
			}
			if (ccDurTm_RM <= ADC::common::k_BA_RMMsgStopUpdtThr)
			{
				return true;
			}
			else
			{
				return false; // RM msg has not been updated
			}		
		} 

		bool BehaviorAnalysis::JudgeAutoLCCrvCond()
		{
			static BA::CondSts autoLCCrvCond = CondSts_NotSatisfied;
			// get the curve at the dist of 2 multiplying velocity by equation
			// curv = 2 * c2 + 6 * c3 * dist
			float param_c2 = 0.0f;
			float param_c3 = 0.0f;
			uint8_t ego_lane_id = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
			auto iter_egoLane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
											[ego_lane_id](const Lane &lane_)
											{
												return (ego_lane_id == lane_.start_id);
											});
			if (iter_egoLane != this->input_BA.RM_RoadModel.lane_list.end())
			{
				// get ego lane c2 and c3 
				BA::FeatureTag trgtFeatureTag = BA::FeatureTag_LineFllw;
				auto iter_egoCrrd = std::find_if(iter_egoLane->corridor_list.begin(), iter_egoLane->corridor_list.end(),
												[trgtFeatureTag](const Corridor &corridor_)
												{
													return (trgtFeatureTag == corridor_.feature_tag);
												});
				if (iter_egoCrrd != iter_egoLane->corridor_list.end())
				{
					param_c2 = iter_egoCrrd->reference_line.c2;
					param_c3 = iter_egoCrrd->reference_line.c3;
				}
			}
			float curvVal = 2 * param_c2 + 6 * param_c3 * ADC::common::k_BA_PreviewTH4CurvCalc * this->input_BA.VSE_LongitudinalVelocity; 
			DLOG(INFO) << "curv value = " << curvVal << std::endl;
			switch (autoLCCrvCond)
			{
			case BA::CondSts_NotSatisfied:
				if (curvVal < ADC::common::k_BA_AutoLCCrvLowThr)
				{
					autoLCCrvCond = CondSts_Satisfied;
				}
				break;
			case BA::CondSts_Satisfied:
				if (curvVal > ADC::common::k_BA_AutoLCCrvHighThr)
				{
					autoLCCrvCond = CondSts_NotSatisfied;
				}
				break;
			}
			if (ADC::common::k_BA_AutoLCCrvInFlg == false)
			{
				autoLCCrvCond = CondSts_Satisfied;
			}
			if (autoLCCrvCond == CondSts_Satisfied)
			{
				return true;
			}
			else	
			{
				return false;
			}
		}

		void BehaviorAnalysis::AdjTrpLatSts(const GlobalInfoProcess &globalInfoProc, const bool &toUpdtGlbInfo)
		{
			// 20221028: change trp lat sts to emergency in spec case
			// if dist btw ego car and last lc change point in ego lane is less than 3.5 * v_ego
			// and ctrl wp is JCT/IC/RmpDvdRd or rmd wp is subSpltRd
			// and dist btw ego car and wp is less than 400m
			// amd last action is lc and the same with lane change direction determined by navigation route
			this->BA_TRPLatSts = BA::TRPLatSts_Comfort;
			if (toUpdtGlbInfo == true &&
				(globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_RmpDvdRd ||
				 globalInfoProc.globalInfoContainer.m_TgtRoadTyp_RWP == BA::TargetRoadTyp_SubSpltRd) &&
				(this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_LeftLaneChange && globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Left) 
				|| (this->input_BA.BA_OptAction_LastCycle == BA::BA_ActionTyp_RightLaneChange && globalInfoProc.globalInfoContainer.m_GlobalLaneChgDrct == BA::LaneChgDrct_Right))
			{
				if (globalInfoProc.globalInfoContainer.m_DistToWayPoint < 400.0f)
				{
					// get ego lane's last lane change point
					uint8_t ego_lane_id = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
					auto it_egoLn = std::find_if(globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.begin(), globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end(),
												 [ego_lane_id](const WPExtendLaneStruct wp_ln)
												 {
													 return (ego_lane_id == wp_ln.lane_id);
												 });
			        if (it_egoLn != globalInfoProc.globalInfoContainer.m_WpExtendLaneLst.end())
					{
						float minSwitch2EmergDst = 3.5f * this->input_BA.VSE_LongitudinalVelocity;
						minSwitch2EmergDst = std::max(minSwitch2EmergDst, 45.0f);
						float dstBtwEgo2LastLCPnt = globalInfoProc.globalInfoContainer.m_DistToWayPoint - it_egoLn->lane_latest_lc_point_offset_to_wp;
						if ((minSwitch2EmergDst > dstBtwEgo2LastLCPnt) && (dstBtwEgo2LastLCPnt > 0.0f))
						{
							this->BA_TRPLatSts = BA::TRPLatSts_Emergency;
						} 
					}
				}
			}
			// 20221102: 
			// if vse speed is less than 45kph
			// then set trplatsts to Emergency
			// if vse speed is greater than 55kph
			// then set trplatsts to Comfort
			static bool isEgoSpdInLowRegion = false;
			if (isEgoSpdInLowRegion == false && this->input_BA.VSE_LongitudinalVelocity < ADC::common::k_BA_LowSpdRegionLowThr/3.6f)
			{
				isEgoSpdInLowRegion = true;
			}
			if (isEgoSpdInLowRegion == true && this->input_BA.VSE_LongitudinalVelocity > ADC::common::k_BA_LowSpdRegionHighThr/3.6f)
			{
				isEgoSpdInLowRegion = false;
			}
			if ((this->BA_OptAction_Deb == BA::BA_ActionTyp_LeftLaneChange || this->BA_OptAction_Deb == BA::BA_ActionTyp_RightLaneChange) &&
			    isEgoSpdInLowRegion == true)
			{
				this->BA_TRPLatSts = BA::TRPLatSts_Emergency;
			}
		}

		bool BehaviorAnalysis::DebounceTRPLatSts()
		{
			// if the current input value is the same with that of previous input, then reset the internal count value
			// else
			//    if the internal count value is smaller than 300ms/50ms, accumulate the internal count value
			//    else replace the previous input values inside with current inputs, and reset the internal count value
			// output the previous input values inside
			static int internTRPLatStsCntVal = 0;
			static uint8_t TRPLatStsPrev = uint8_t(0);
			if (TRPLatStsPrev != this->BA_TRPLatSts)
			{
				if (internTRPLatStsCntVal < ADC::common::k_BA_TRPLatStsDebCnt)
				{
					internTRPLatStsCntVal++;
				}
				else
				{
					internTRPLatStsCntVal = 0;
					TRPLatStsPrev = this->BA_TRPLatSts;
				}
			}
			else
			{
				internTRPLatStsCntVal = 0;
			}
			this->BA_TRPLatSts_Deb = TRPLatStsPrev;
			return true;
		}

		void BehaviorAnalysis::UpdateLftNghrDesLCPnt(const bool &chgSpdThrFlg, GlobalInfoProcess &globalInfoProc)
		{
			static bool lftNgbrLnDesLCPntUpdt = false;
			static uint8_t lockedLnID = uint8_t(0);
			if (chgSpdThrFlg == true)
			{
				lftNgbrLnDesLCPntUpdt = true;
			}
			if (globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_ICRd && 
				globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_JCTRd &&
				globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_SpltRd)
			{
				lftNgbrLnDesLCPntUpdt = false;
			}
			else if ((globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd || 
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd ||
					  globalInfoProc.globalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd) &&
					 globalInfoProc.globalInfoContainer.m_DistToWayPoint < 1300.0f)
		    {
				lftNgbrLnDesLCPntUpdt = false;
			}
			if (chgSpdThrFlg == true)
			{
				uint8_t egoLnID = this->input_BA.RM_RoadModel.car_position.ego_lane_id;
				uint8_t lftNbghLnID = uint8_t(0);
				auto it_Lane = std::find_if(this->input_BA.RM_RoadModel.lane_list.begin(), this->input_BA.RM_RoadModel.lane_list.end(),
									        [egoLnID](const Lane &lane_)
											{
												return (egoLnID == lane_.start_id);
											});
				if (it_Lane != this->input_BA.RM_RoadModel.lane_list.end())
				{
					if (!it_Lane->left_neighbor_lane.empty())
					{
						auto it_minLftStRngLn = std::min_element(it_Lane->left_neighbor_lane.begin(), it_Lane->left_neighbor_lane.end(), Cmp_by_stRng);
						lftNbghLnID = it_minLftStRngLn->lane_id;
					}
				}
				lockedLnID = lftNbghLnID;
				if (lftNbghLnID != uint8_t(0))
				{
					// try to get global lane info from GlobalInfoArr
					auto iter_LftGlbNgbrLn = std::find_if(globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.begin(), globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end(),
														  [lftNbghLnID](const BA::GlobalLaneInfo &glbLn_)
														  {
															  return (lftNbghLnID == glbLn_.m_LaneID);
														  });
					if (iter_LftGlbNgbrLn != globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end())
					{
						if (!iter_LftGlbNgbrLn->m_DesiredLaneChangePoint.empty())
						{
							iter_LftGlbNgbrLn->m_DesiredLaneChangePoint.front() = ADC::common::k_BA_AntiLCPntThr;
							DLOG(INFO) << "left neighbor lane's desired lc point updated 01" << std::endl;
							DLOG(INFO) << "left neighbor lane's id = " << int(iter_LftGlbNgbrLn->m_LaneID) << std::endl;
							DLOG(INFO) << "left neighbor lane's des lc point = " << int(iter_LftGlbNgbrLn->m_DesiredLaneChangePoint.front()) << std::endl;
						}
					}
				}
			}
			else if (lftNgbrLnDesLCPntUpdt == true && chgSpdThrFlg == false)
			{
				if (lockedLnID != uint8_t(0))
				{
					// try to get global lane info from GlobalInfoArr
					uint8_t tmpLnID = lockedLnID;
					auto iter_GlbNgbrLn = std::find_if(globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.begin(), globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end(),
													  [tmpLnID](const BA::GlobalLaneInfo &glbLn_)
													  {
														  return (tmpLnID == glbLn_.m_LaneID);
													  });
					if (iter_GlbNgbrLn != globalInfoProc.globalInfoContainer.m_GlobalLaneInfoArr.end())
					{
						if (!iter_GlbNgbrLn->m_DesiredLaneChangePoint.empty())
						{
							iter_GlbNgbrLn->m_DesiredLaneChangePoint.front() = ADC::common::k_BA_AntiLCPntThr;
							DLOG(INFO) << "lane's desired lc point updated 02" << std::endl;
							DLOG(INFO) << "lane's id = " << int(iter_GlbNgbrLn->m_LaneID) << std::endl;
							DLOG(INFO) << "lane's des lc point = " << int(iter_GlbNgbrLn->m_DesiredLaneChangePoint.front()) << std::endl;
						}
					}
				}
			}
			else
			{
				lockedLnID = uint8_t(0);
			}
		}
		void BehaviorAnalysis::testWrapper()
		{

		}

	} // namespace planning
} // namespace ADC
